@@ -6,6 +6,17 @@ parts of fish.
 
 #include "config.h"
 
+
+#include <unistd.h>
+
+#ifdef HAVE_STROPTS_H
+#include <stropts.h>
+#endif
+
+#ifdef HAVE_SIGINFO_H
+#include <siginfo.h>
+#endif
+
 #include <stdlib.h>
 #include <termios.h>
 #include <wchar.h>
@@ -13,14 +24,21 @@ parts of fish.
 #include <stdio.h>
 #include <dirent.h>
 #include <sys/types.h>
+
+#ifdef HAVE_SYS_TERMIOS_H
+#include <sys/termios.h>
+#endif
+
+#ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
+#endif
+
 #include <sys/stat.h>
 #include <unistd.h>
 #include <wctype.h>
 #include <errno.h>
 #include <limits.h>
 #include <stdarg.h>		
-#include <signal.h>		
 #include <locale.h>
 #include <time.h>
 #include <sys/time.h>
@@ -45,10 +63,15 @@ parts of fish.
 #include <termio.h>
 #endif
 
+#if HAVE_TERM_H
 #include <term.h>
+#elif HAVE_NCURSES_TERM_H
+#include <ncurses/term.h>
+#endif
 
-
+#include "fallback.h"
 #include "util.h"
+
 #include "wutil.h"
 #include "common.h"
 #include "expand.h"
@@ -56,15 +79,10 @@ parts of fish.
 #include "wildcard.h"
 #include "parser.h"
 
-/**
-   Error message to show on string convertion error
-*/
-#define STR2WCS_MSG "fish: Invalid multibyte sequence \'"
-
-/**
-   The maximum number of minor errors to report. Further errors will be omitted.
-*/
-#define ERROR_MAX_COUNT 1
+#include "util.c"
+#include "halloc.c"
+#include "halloc_util.c"
+#include "fallback.c"
 
 /**
    The number of milliseconds to wait between polls when attempting to acquire 
@@ -74,19 +92,8 @@ parts of fish.
 
 struct termios shell_modes;      
 
-/**
-   Number of error encountered. This is reset after each command, and
-   used to limit the number of error messages on commands with many
-   string convertion problems.
-*/
-static int error_count=0;
-
-int error_max=1;
-
 wchar_t ellipsis_char;
 
-static int c1=0, c2=0, c3=0, c4=0, c5;
-
 char *profile=0;
 
 wchar_t *program_name;
@@ -98,69 +105,27 @@ int debug_level=1;
 */
 static struct winsize termsize;
 
-
 /**
-   Number of nested calls to the block function. Unblock when this reaches 0.
+   String buffer used by the wsetlocale function
 */
-static int block_count=0;
-
 static string_buffer_t *setlocale_buff=0;
 
-
-void common_destroy()
-{
-	if( setlocale_buff )
-	{
-		sb_destroy( setlocale_buff );
-		free( setlocale_buff );
-	}
-	
-	debug( 3, L"Calls: wcsdupcat %d, wcsdupcat2 %d, wcsndup %d, str2wcs %d, wcs2str %d", c1, c2, c3, c4, c5 );
-}
-
-
-
 wchar_t **list_to_char_arr( array_list_t *l )
 {
 	wchar_t ** res = malloc( sizeof(wchar_t *)*(al_get_count( l )+1) );
 	int i;
 	if( res == 0 )
 	{
-		die_mem();
+		DIE_MEM();
 	}
 	for( i=0; i<al_get_count( l ); i++ )
+	{		
 		res[i] = (wchar_t *)al_get(l,i);
+	}
 	res[i]='\0';
 	return res;	
 }
 
-void block()
-{
-	block_count++;
-	if( block_count == 1 )
-	{
-		sigset_t chldset; 
-		sigemptyset( &chldset );
-		sigaddset( &chldset, SIGCHLD );
-		sigprocmask(SIG_BLOCK, &chldset, 0);
-	}
-}
-
-
-void unblock()
-{
-	block_count--;
-	if( block_count == 0 )
-	{
-		sigset_t chldset; 
-		sigemptyset( &chldset );
-		sigaddset( &chldset, SIGCHLD );
-		sigprocmask(SIG_UNBLOCK, &chldset, 0);
-	}
-}
-
-
-
 int fgetws2( wchar_t **b, int *len, FILE *f )
 {
 	int i=0;
@@ -168,14 +133,6 @@ int fgetws2( wchar_t **b, int *len, FILE *f )
 	
 	wchar_t *buff = *b;
 
-	/*
-	  This is a kludge: We block SIGCHLD while reading, since I can't
-	  get getwc to perform reliably when signals are flying. Even when
-	  watching for EINTR errors, bytes are lost. 
-	*/
-
-	block();
-
 	while( 1 )
 	{
 		/* Reallocate the buffer if necessary */
@@ -185,7 +142,7 @@ int fgetws2( wchar_t **b, int *len, FILE *f )
 			buff = realloc( buff, sizeof(wchar_t)*new_len );
 			if( buff == 0 )
 			{
-				die_mem();
+				DIE_MEM();
 			}
 			else
 			{
@@ -214,8 +171,6 @@ int fgetws2( wchar_t **b, int *len, FILE *f )
 			case L'\n':
 			case L'\0':
 				buff[i]=L'\0';
-				unblock();
-
 				return i;				
 				/* Ignore carriage returns */
 			case L'\r':
@@ -228,7 +183,6 @@ int fgetws2( wchar_t **b, int *len, FILE *f )
 
 
 	}
-	unblock();
 
 }
 
@@ -253,64 +207,122 @@ void sort_list( array_list_t *comp )
 
 wchar_t *str2wcs( const char *in )
 {
-	c4++;
+	wchar_t *out;
+	size_t len = strlen(in);
 	
-	wchar_t *res;
-	
-	res = malloc( sizeof(wchar_t)*(strlen(in)+1) );
-	
-	if( !res )
+	out = malloc( sizeof(wchar_t)*(len+1) );
+
+	if( !out )
 	{
-		die_mem();
+		DIE_MEM();
 	}
-	
-	if( (size_t)-1 == mbstowcs( res, in, sizeof(wchar_t)*(strlen(in)) +1) )
-	{
-		error_count++;
-		if( error_count <=error_max )
-		{
-			fflush( stderr );			
-			write( 2,
-				   STR2WCS_MSG,
-				   strlen(STR2WCS_MSG) );
-			write( 2,
-				   in,
-				   strlen(in ));
-			write( 2, 
-				   "\'\n",
-				   2 );
-		}
-		
-		free(res);
-		return 0;
-	}	
-	
-	return res;
-	
+
+	return str2wcs_internal( in, out );
 }
 
-void error_reset()
+wchar_t *str2wcs_internal( const char *in, wchar_t *out )
 {
-	error_count=0;
+	size_t res=0;
+	int in_pos=0;
+	int out_pos = 0;
+	mbstate_t state;
+	size_t len;
+
+	CHECK( in, 0 );
+	CHECK( out, 0 );
+		
+	len = strlen(in);
+
+	memset( &state, 0, sizeof(state) );
+	
+	while( in[in_pos] )
+	{
+		res = mbrtowc( &out[out_pos], &in[in_pos], len-in_pos, &state );
+
+		switch( res )
+		{
+			case (size_t)(-2):
+			case (size_t)(-1):
+				{
+					out[out_pos] = ENCODE_DIRECT_BASE + (unsigned char)in[in_pos];
+					in_pos++;
+					memset( &state, 0, sizeof(state) );
+					break;
+				}
+				
+			case 0:
+			{
+				return out;
+			}
+		
+			default:
+			{
+				in_pos += res;
+				break;
+			}
+		}
+		out_pos++;
+	}
+	out[out_pos] = 0;
+	
+	return out;	
 }
 
 char *wcs2str( const wchar_t *in )
 {
-	c5++;
+	char *out;	
+	
+	out = malloc( MAX_UTF8_BYTES*wcslen(in)+1 );
 
-	char *res = malloc( MAX_UTF8_BYTES*wcslen(in)+1 );
-	if( res == 0 )
+	if( !out )
 	{
-		die_mem();
+		DIE_MEM();
 	}
 
-	wcstombs( res, 
-			  in,
-			  MAX_UTF8_BYTES*wcslen(in)+1 );
+	return wcs2str_internal( in, out );
+}
 
-//	res = realloc( res, strlen( res )+1 );
+char *wcs2str_internal( const wchar_t *in, char *out )
+{
+	size_t res=0;
+	int in_pos=0;
+	int out_pos = 0;
+	mbstate_t state;
 
-	return res;
+	CHECK( in, 0 );
+	CHECK( out, 0 );
+	
+	memset( &state, 0, sizeof(state) );
+	
+	while( in[in_pos] )
+	{
+		if( in[in_pos] == INTERNAL_SEPARATOR )
+		{
+		}
+		else if( ( in[in_pos] >= ENCODE_DIRECT_BASE) &&
+			( in[in_pos] < ENCODE_DIRECT_BASE+256) )
+		{
+			out[out_pos++] = in[in_pos]- ENCODE_DIRECT_BASE;
+		}
+		else
+		{
+			res = wcrtomb( &out[out_pos], in[in_pos], &state );
+			
+			if( res == (size_t)(-1) )
+			{
+				debug( 1, L"Wide character %d has no narrow representation", in[in_pos] );
+				memset( &state, 0, sizeof(state) );
+			}
+			else
+			{
+				out_pos += res;
+			}
+		}
+		in_pos++;
+	}
+	out[out_pos] = 0;
+	
+	return out;	
 }
 
 char **wcsv2strv( const wchar_t **in )
@@ -323,7 +335,7 @@ char **wcsv2strv( const wchar_t **in )
 	char **res = malloc( sizeof( char *)*(count+1));
 	if( res == 0 )
 	{
-		die_mem();		
+		DIE_MEM();		
 	}
 
 	for( i=0; i<count; i++ )
@@ -337,15 +349,11 @@ char **wcsv2strv( const wchar_t **in )
 
 wchar_t *wcsdupcat( const wchar_t *a, const wchar_t *b )
 {
-	c1++;
-
-	return wcsdupcat2( a, b, 0 );
+	return wcsdupcat2( a, b, (void *)0 );
 }
 
 wchar_t *wcsdupcat2( const wchar_t *a, ... )
 {
-	c2++;
-	
 	int len=wcslen(a);
 	int pos;
 	va_list va, va2;
@@ -362,7 +370,7 @@ wchar_t *wcsdupcat2( const wchar_t *a, ... )
 	wchar_t *res = malloc( sizeof(wchar_t)*(len +1 ));
 	if( res == 0 )
 	{
-		die_mem();
+		DIE_MEM();
 	}
 	
 	wcscpy( res, a );
@@ -388,7 +396,7 @@ wchar_t **strv2wcsv( const char **in )
 	wchar_t **res = malloc( sizeof( wchar_t *)*(count+1));
 	if( res == 0 )
 	{
-		die_mem();
+		DIE_MEM();
 	}
 
 	for( i=0; i<count; i++ )
@@ -400,290 +408,24 @@ wchar_t **strv2wcsv( const char **in )
 
 }
 
-
-#ifndef HAVE_WCSNDUP
-wchar_t *wcsndup( const wchar_t *in, int c )
-{
-	c3++;
-	
-	wchar_t *res = malloc( sizeof(wchar_t)*(c+1) );
-	if( res == 0 )
-	{
-		die_mem();
-	}
-	wcsncpy( res, in, c );
-	res[c] = L'\0';	
-	return res;	
-}
-#endif
-
-long convert_digit( wchar_t d, int base )
-{
-	long res=-1;
-	if( (d <= L'9') && (d >= L'0') )
-	{
-		res = d - L'0';
-	}
-	else if( (d <= L'z') && (d >= L'a') )
-	{
-		res = d + 10 - L'a';		
-	}
-	else if( (d <= L'Z') && (d >= L'A') )
-	{
-		res = d + 10 - L'A';		
-	}
-	if( res >= base )
-	{
-		res = -1;
-	}
-	
-	return res;
-}
-
-
-long wcstol(const wchar_t *nptr, 
-			wchar_t **endptr,
-			int base)
-{
-	long long res=0;
-	int is_set=0;
-	if( base > 36 )
-	{
-		errno = EINVAL;
-		return 0;
-	}
-
-	while( 1 )
-	{
-		long nxt = convert_digit( *nptr, base );
-		if( endptr != 0 )
-			*endptr = (wchar_t *)nptr;
-		if( nxt < 0 )
-		{
-			if( !is_set )
-			{
-				errno = EINVAL;
-			}
-			return res;			
-		}
-		res = (res*base)+nxt;
-		is_set = 1;
-		if( res > LONG_MAX )
-		{
-			errno = ERANGE;
-			return LONG_MAX;
-		}
-		if( res < LONG_MIN )
-		{
-			errno = ERANGE;
-			return LONG_MIN;
-		}
-		nptr++;
-	}
-}
-
-/*$OpenBSD: strlcat.c,v 1.11 2003/06/17 21:56:24 millert Exp $*/
-
-/*
- * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
- *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
- *
- * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
- * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
- * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
- * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
- * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
- * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
- * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
- */
-
-size_t
-wcslcat(wchar_t *dst, const wchar_t *src, size_t siz)
-{
-	
-	register wchar_t *d = dst;
-	register const wchar_t *s = src;
-	register size_t n = siz;	
-	size_t dlen;
-
-	/* Find the end of dst and adjust bytes left but don't go past end */
-	while (n-- != 0 && *d != '\0')
-		d++;
-	
-	dlen = d - dst;
-	n = siz - dlen;	
-
-	if (n == 0)
-		return(dlen + wcslen(s));
-
-	while (*s != '\0') 
-	{
-		if (n != 1) 
-		{
-			*d++ = *s;
-			n--;
-		}
-		s++;
-	}
-	*d = '\0';
-
-	return(dlen + (s - src));
-	/* count does not include NUL */
-}
-
-/*$OpenBSD: strlcpy.c,v 1.8 2003/06/17 21:56:24 millert Exp $*/
-
-/*
- * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
- *
- * Permission to use, copy, modify, and distribute this software for any
- * purpose with or without fee is hereby granted, provided that the above
- * copyright notice and this permission notice appear in all copies.
- *
- * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
- * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
- * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
- * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
- * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
- * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
- * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
- */
-
-size_t
-wcslcpy(wchar_t *dst, const wchar_t *src, size_t siz)
-{
-	register wchar_t *d = dst;
-	register const wchar_t *s = src;
-	register size_t n = siz;
-	
-	/* Copy as many bytes as will fit */
-	if (n != 0 && --n != 0) 
-	{		
-		do 
-		{			
-			if ((*d++ = *s++) == 0)
-				break;
-		}
-		while (--n != 0);
-	}
-
-	/* Not enough room in dst, add NUL and traverse rest of src */
-	if (n == 0) 
-	{		
-		if (siz != 0)
-			*d = '\0';
-		/* NUL-terminate dst */
-		while (*s++)
-			;
-	}
-	return(s - src - 1);
-	/* count does not include NUL */
-}
-
-#ifndef HAVE_WCSDUP
-wchar_t *wcsdup( const wchar_t *in )
-{
-	size_t len=wcslen(in);
-	wchar_t *out = malloc( sizeof( wchar_t)*(len+1));
-	if( out == 0 )
-	{
-		die_mem();
-	}
-
-	memcpy( out, in, sizeof( wchar_t)*(len+1));
-	return out;
-	
-}
-#endif
-
-#ifndef HAVE_WCSLEN
-size_t wcslen(const wchar_t *in)
-{
-	const wchar_t *end=in;
-	while( *end )
-		end++;
-	return end-in;
-}
-#endif
-
-
-#ifndef HAVE_WCSCASECMP
-int wcscasecmp( const wchar_t *a, const wchar_t *b )
-{
-	if( *a == 0 )
-	{
-		return (*b==0)?0:-1;
-	}
-	else if( *b == 0 )
-	{
-		return 1;
-	}
-	int diff = towlower(*a)-towlower(*b);
-	if( diff != 0 )
-		return diff;
-	else
-		return wcscasecmp( a+1,b+1);
-}
-#endif
-
-
-#ifndef HAVE_WCSNCASECMP
-int wcsncasecmp( const wchar_t *a, const wchar_t *b, int count )
-{
-	if( count == 0 )
-		return 0;
-	
-	if( *a == 0 )
-	{
-		return (*b==0)?0:-1;
-	}
-	else if( *b == 0 )
-	{
-		return 1;
-	}
-	int diff = towlower(*a)-towlower(*b);
-	if( diff != 0 )
-		return diff;
-	else
-		return wcsncasecmp( a+1,b+1, count-1);
-}
-#endif
-
-int wcsvarname( wchar_t *str )
+wchar_t *wcsvarname( wchar_t *str )
 {
 	while( *str )
 	{
 		if( (!iswalnum(*str)) && (*str != L'_' ) )
 		{
-			return 0;
+			return str;
 		}
 		str++;
 	}
-	return 1;
-	
-	
+	return 0;
 }
 
-#if !HAVE_WCWIDTH
-/**
-   Return the number of columns used by a character. 
-
-   In locales without a native wcwidth, Unicode is probably so broken
-   that it isn't worth trying to implement a real wcwidth. This
-   wcwidth assumes any printing character takes up one column.
-*/
-int wcwidth( wchar_t c )
+int wcsvarchr( wchar_t chr )
 {
-	if( c < 32 )
-		return 0;
-	if ( c == 127 )
-		return 0;
-	return 1;
+	return ( (iswalnum(chr)) || (chr == L'_' ));
 }
-#endif
+
 
 /** 
 	The glibc version of wcswidth seems to hang on some strings. fish uses this replacement.
@@ -704,30 +446,36 @@ int my_wcswidth( const wchar_t *c )
 	return res;
 }
 
-wchar_t *quote_end( const wchar_t *in )
+wchar_t *quote_end( const wchar_t *pos )
 {
-	switch( *in )
+	wchar_t c = *pos;
+	
+	while( 1 )
 	{
-		case '"':
+		pos++;
+		
+		if( !*pos )
+			return 0;
+		
+		if( *pos == L'\\')
 		{
-			while(1)
-			{
-				in = wcschr( in+1, L'"' );
-				if( !in )
-					return 0;
-				if( *(in-1) != L'\\' )
-					return (wchar_t *)in;
-			}
+			pos++;
+			if( !*pos )
+				return 0;
 		}
-		case '\'':
+		else
 		{
-			return wcschr( in+1, L'\'' );
+			if( *pos == c )
+			{
+				return (wchar_t *)pos;
+			}
 		}
 	}
 	return 0;
+	
 }
 
-
+				
 const wchar_t *wsetlocale(int category, const wchar_t *locale)
 {
 
@@ -747,8 +495,7 @@ const wchar_t *wsetlocale(int category, const wchar_t *locale)
 	
 	if( !setlocale_buff )
 	{
-		setlocale_buff = malloc( sizeof(string_buffer_t) );
-		sb_init( setlocale_buff);
+		setlocale_buff = sb_halloc( global_context);
 	}
 	
 	sb_clear( setlocale_buff );
@@ -762,6 +509,8 @@ int contains_str( const wchar_t *a, ... )
 	wchar_t *arg;
 	va_list va;
 	int res = 0;
+
+	CHECK( a, 0 );
 	
 	va_start( va, a );
 	while( (arg=va_arg(va, wchar_t *) )!= 0 ) 
@@ -790,51 +539,43 @@ int read_blocked(int fd, void *buf, size_t count)
 	return res;	
 }
 
-int writeb( tputs_arg_t b )
-{
-	write( 1, &b, 1 );
-//	putc( b, stdout );
-	return 0;
-}
-
-void die_mem()
-{
-	/*
-	  Do not translate this message, and do not send it through the
-	  usual channels. This increases the odds that the message gets
-	  through correctly, even if we are out of memory.
-	*/
-	fwprintf( stderr, L"Out of memory, shutting down fish.\n" );
-	exit(1);
-}
-
 void debug( int level, const wchar_t *msg, ... )
 {
 	va_list va;
 	string_buffer_t sb;
+	string_buffer_t sb2;
 	
+	CHECK( msg, );
+		
 	if( level > debug_level )
 		return;
 
 	sb_init( &sb );
-	va_start( va, msg );
-	
+	sb_init( &sb2 );
+
 	sb_printf( &sb, L"%ls: ", program_name );
+
+	va_start( va, msg );	
 	sb_vprintf( &sb, msg, va );
 	va_end( va );	
 
-	write_screen( (wchar_t *)sb.buff );
+	write_screen( (wchar_t *)sb.buff, &sb2 );
+	fwprintf( stderr, L"%ls", sb2.buff );	
 
 	sb_destroy( &sb );	
+	sb_destroy( &sb2 );	
 }
 
-void write_screen( const wchar_t *msg )
+void write_screen( const wchar_t *msg, string_buffer_t *buff )
 {
 	const wchar_t *start, *pos;
 	int line_width = 0;
 	int tok_width = 0;
 	int screen_width = common_get_width();
 	
+	CHECK( msg, );
+	CHECK( buff, );
+		
 	if( screen_width )
 	{
 		start = pos = msg;
@@ -849,7 +590,7 @@ void write_screen( const wchar_t *msg )
 			*/
 			while( *pos && ( !wcschr( L" \n\r\t", *pos ) ) )
 			{
-
+				
 				/*
 				  Check is token is wider than one line.
 				  If so we mark it as an overflow and break the token.
@@ -874,12 +615,12 @@ void write_screen( const wchar_t *msg )
 			else if( overflow )
 			{
 				/*
-				  In case of overflow, we print a newline, except if we alreade are at position 0
+				  In case of overflow, we print a newline, except if we already are at position 0
 				*/
 				wchar_t *token = wcsndup( start, pos-start );
 				if( line_width != 0 )
-					putwc( L'\n', stderr );			
-				fwprintf( stderr, L"%ls-\n", token );
+					sb_append_char( buff, L'\n' );
+				sb_printf( buff, L"%ls-\n", token );
 				free( token );
 				line_width=0;
 			}
@@ -891,107 +632,148 @@ void write_screen( const wchar_t *msg )
 				wchar_t *token = wcsndup( start, pos-start );
 				if( (line_width + (line_width!=0?1:0) + tok_width) > screen_width )
 				{
-					putwc( L'\n', stderr );
+					sb_append_char( buff, L'\n' );
 					line_width=0;
 				}
-				fwprintf( stderr, L"%ls%ls", line_width?L" ":L"", token );
+				sb_printf( buff, L"%ls%ls", line_width?L" ":L"", token );
 				free( token );
 				line_width += (line_width!=0?1:0) + tok_width;
 			}
+			
 			/*
 			  Break on end of string
 			*/
 			if( !*pos )
+			{
 				break;
-		
+			}
+			
 			start=pos;
 		}
 	}
 	else
 	{
-		fwprintf( stderr, L"%ls", msg );
+		sb_printf( buff, L"%ls", msg );
 	}
-	putwc( L'\n', stderr );
+	sb_append_char( buff, L'\n' );
 }
 
 wchar_t *escape( const wchar_t *in, 
 				 int escape_all )
 {
-	wchar_t *out = malloc( sizeof(wchar_t)*(wcslen(in)*4 + 1));
-	wchar_t *pos=out;
+	wchar_t *out;
+	wchar_t *pos;
+
+	if( !in )
+	{
+		debug( 0, L"%s called with null input", __func__ );
+		exit(1);
+	}
+	
+	out = malloc( sizeof(wchar_t)*(wcslen(in)*4 + 1));
+	pos = out;
+	
 	if( !out )
-		die_mem();
+		DIE_MEM();
 	
 	while( *in != 0 )
 	{
-		switch( *in )
+
+		if( ( *in >= ENCODE_DIRECT_BASE) &&
+			( *in < ENCODE_DIRECT_BASE+256) )
 		{
-			case L'\t':
-				*(pos++) = L'\\';
-				*(pos++) = L't';					
-				break;
-					
-			case L'\n':
-				*(pos++) = L'\\';
-				*(pos++) = L'n';					
-				break;
-					
-			case L'\b':
-				*(pos++) = L'\\';
-				*(pos++) = L'b';					
-				break;
-					
-			case L'\r':
-				*(pos++) = L'\\';
-				*(pos++) = L'r';					
-				break;
-					
-			case L'\e':
-				*(pos++) = L'\\';
-				*(pos++) = L'e';					
-				break;
-					
-			case L'\\':
-			case L'&':
-			case L'$':
-			case L' ':
-			case L'#':
-			case L'^':
-			case L'<':
-			case L'>':
-			case L'(':
-			case L')':
-			case L'[':
-			case L']':
-			case L'{':
-			case L'}':
-			case L'?':
-			case L'*':
-			case L'|':
-			case L';':
-			case L':':
-			case L'\'':
-			case L'"':
-			case L'%':
-			case L'~':
-				if( escape_all )
-					*pos++ = L'\\';
-				*pos++ = *in;
-				break;
-					
-			default:
-				if( *in < 32 )
-				{
-					int tmp = (*in)%16;
-					*pos++ = L'\\';
-					*pos++ = L'x';
-					*pos++ = ((*in>15)? L'1' : L'0');
-					*pos++ = tmp > 9? L'a'+(tmp-10):L'0'+tmp;
-				}
-				else
-					*pos++ = *in;
-				break;
+			int val = *in - ENCODE_DIRECT_BASE;
+			int tmp;
+			
+			*(pos++) = L'\\';
+			*(pos++) = L'X';
+			
+			tmp = val/16;			
+			*pos++ = tmp > 9? L'a'+(tmp-10):L'0'+tmp;
+			
+			tmp = val%16;			
+			*pos++ = tmp > 9? L'a'+(tmp-10):L'0'+tmp;
+			
 		}
+		else
+		{
+			
+			switch( *in )
+			{
+				case L'\t':
+					*(pos++) = L'\\';
+					*(pos++) = L't';					
+					break;
+					
+				case L'\n':
+					*(pos++) = L'\\';
+					*(pos++) = L'n';					
+					break;
+					
+				case L'\b':
+					*(pos++) = L'\\';
+					*(pos++) = L'b';					
+					break;
+					
+				case L'\r':
+					*(pos++) = L'\\';
+					*(pos++) = L'r';					
+					break;
+					
+				case L'\e':
+					*(pos++) = L'\\';
+					*(pos++) = L'e';					
+					break;
+					
+				case L'\\':
+				case L'&':
+				case L'$':
+				case L' ':
+				case L'#':
+				case L'^':
+				case L'<':
+				case L'>':
+				case L'(':
+				case L')':
+				case L'[':
+				case L']':
+				case L'{':
+				case L'}':
+				case L'?':
+				case L'*':
+				case L'|':
+				case L';':
+				case L':':
+				case L'\'':
+				case L'"':
+				case L'%':
+				case L'~':
+				{
+					if( escape_all )
+						*pos++ = L'\\';
+					*pos++ = *in;
+					break;
+				}
+				
+				default:
+				{
+					if( *in < 32 )
+					{
+						int tmp = (*in)%16;
+						*pos++ = L'\\';
+						*pos++ = L'x';
+						*pos++ = ((*in>15)? L'1' : L'0');
+						*pos++ = tmp > 9? L'a'+(tmp-10):L'0'+tmp;
+					}
+					else
+					{
+						*pos++ = *in;
+					}
+					break;
+				}
+			}
+		}
+		
 		in++;
 	}
 	*pos = 0;
@@ -1003,13 +785,19 @@ wchar_t *unescape( const wchar_t * orig, int unescape_special )
 {
 	
 	int mode = 0; 
-	int in_pos, out_pos, len = wcslen( orig );
+	int in_pos, out_pos, len;
 	int c;
 	int bracket_count=0;
 	wchar_t prev=0;	
-	wchar_t *in = wcsdup(orig);
+	wchar_t *in;
+
+	CHECK( orig, 0 );
+		
+	len = wcslen( orig );
+	in = wcsdup( orig );
+
 	if( !in )
-		die_mem();
+		DIE_MEM();
 	
 	for( in_pos=0, out_pos=0; 
 		 in_pos<len; 
@@ -1028,104 +816,206 @@ wchar_t *unescape( const wchar_t * orig, int unescape_special )
 				{
 					switch( in[++in_pos] )
 					{
+						
+						/*
+						  A null character after a backslash is an
+						  error, return null
+						*/
 						case L'\0':
 						{
 							free(in);
 							return 0;
 						}
 						
-						case L'n':
+						/*
+						  Numeric escape sequences. No prefix means
+						  octal escape, otherwise hexadecimal.
+						*/
+						
+						case L'0':
+						case L'1':
+						case L'2':
+						case L'3':
+						case L'4':
+						case L'5':
+						case L'6':
+						case L'7':
+						case L'u':
+						case L'U':
+						case L'x':
+						case L'X':
 						{
-							in[out_pos]=L'\n';
+							int i;
+							long long res=0;
+							int chars=2;
+							int base=16;
+							
+							int byte = 0;
+							wchar_t max_val = ASCII_MAX;
+							
+							switch( in[in_pos] )
+							{
+								case L'u':
+								{
+									chars=4;
+									max_val = UCS2_MAX;
+									break;
+								}
+								
+								case L'U':
+								{
+									chars=8;
+									max_val = WCHAR_MAX;
+									break;
+								}
+								
+								case L'x':
+								{
+									break;
+								}
+								
+								case L'X':
+								{
+									byte=1;
+									max_val = BYTE_MAX;
+									break;
+								}
+								
+								default:
+								{
+									base=8;
+									chars=3;
+									in_pos--;
+									break;
+								}								
+							}
+					
+							for( i=0; i<chars; i++ )
+							{
+								int d = convert_digit( in[++in_pos],base);
+								
+								if( d < 0 )
+								{
+									in_pos--;
+									break;
+								}
+								
+								res=(res*base)|d;
+							}
+
+							if( (res <= max_val) )
+							{
+								in[out_pos] = (byte?ENCODE_DIRECT_BASE:0)+res;
+							}
+							else
+							{	
+								free(in);	
+								return 0;
+							}
+							
+							break;
+						}
+
+						/*
+						  \a means bell (alert)
+						*/
+						case L'a':
+						{
+							in[out_pos]=L'\a';
 							break;
 						}
 						
-						case L'r':
-						{
-							in[out_pos]=L'\r';
-							break;
-						}
-
-						case L't':
-						{
-							in[out_pos]=L'\t';
-							break;
-						}
-
+						/*
+						  \b means backspace
+						*/
 						case L'b':
 						{
 							in[out_pos]=L'\b';
 							break;
 						}
 						
+						/*
+						  \cX means control sequence X
+						*/
+						case L'c':
+						{
+							in_pos++;
+							if( in[in_pos] >= L'a' &&
+								in[in_pos] <= (L'a'+32) )
+							{
+								in[out_pos]=in[in_pos]-L'a'+1;
+							}
+							else if( in[in_pos] >= L'A' &&
+									 in[in_pos] <= (L'A'+32) )
+							{
+								in[out_pos]=in[in_pos]-L'A'+1;
+							}
+							else
+							{
+								free(in);	
+								return 0;
+							}
+							break;
+							
+						}
+						
+						/*
+						  \e means escape
+						*/
 						case L'e':
 						{
-							in[out_pos]=L'\e';
+							in[out_pos]=L'\x1b';
 							break;
 						}
 						
-						case L'u':
-						case L'U':
-						case L'x':
-						case L'o':
+						/*
+						  \f means form feed
+						*/
+						case L'f':
 						{
-							int i;
-							wchar_t res=0;
-							int chars=2;
-							int base=16;
-					
-							switch( in[in_pos] )
-							{
-								case L'u':
-								{
-									base=16;
-									chars=4;
-									break;
-								}
-								
-								case L'U':
-								{
-									base=16;
-									chars=8;
-									break;
-								}
-								
-								case L'x':
-								{
-									base=16;
-									chars=2;
-									break;
-								}
-								
-								case L'o':
-								{
-									base=8;
-									chars=3;
-									break;
-								}
-								
-							}
-					
-							for( i=0; i<chars; i++ )
-							{
-								int d = convert_digit( in[++in_pos],base);
-								if( d < 0 )
-								{
-									in_pos--;
-									break;
-								}
-						
-								res=(res*base)|d;
-						
-							}
-					
-							in[out_pos] = res;
-					
+							in[out_pos]=L'\f';
 							break;
 						}
 
+						/*
+						  \n means newline
+						*/
+						case L'n':
+						{
+							in[out_pos]=L'\n';
+							break;
+						}
+						
+						/*
+						  \r means carriage return
+						*/
+						case L'r':
+						{
+							in[out_pos]=L'\r';
+							break;
+						}
+						
+						/*
+						  \t means tab
+						 */
+						case L't':
+						{
+							in[out_pos]=L'\t';
+							break;
+						}
+
+						/*
+						  \v means vetrical tab
+						*/
+						case L'v':
+						{
+							in[out_pos]=L'\v';
+							break;
+						}
+						
 						default:
 						{
+							in[out_pos++] = INTERNAL_SEPARATOR;							
 							in[out_pos]=in[in_pos];
 							break;
 						}
@@ -1133,7 +1023,8 @@ wchar_t *unescape( const wchar_t * orig, int unescape_special )
 				}
 				else 
 				{
-					switch( in[in_pos]){
+					switch( in[in_pos])
+					{
 						case L'~':
 						{
 							if( unescape_special && (in_pos == 0) )
@@ -1275,6 +1166,31 @@ wchar_t *unescape( const wchar_t * orig, int unescape_special )
 			*/
 			case 1:
 			{
+				if( c == L'\\' )
+				{
+					switch( in[++in_pos] )
+					{
+						case '\\':
+						case L'\'':
+						{
+							in[out_pos]=in[in_pos];
+							break;
+						}
+						
+						case 0:
+						{
+							free(in);
+							return 0;
+						}
+						
+						default:
+						{
+							in[out_pos++] = L'\\';
+							in[out_pos]= in[in_pos];
+						}
+					}
+					
+				}
 				if( c == L'\'' )
 				{
 					in[out_pos] = INTERNAL_SEPARATOR;							
@@ -1312,12 +1228,14 @@ wchar_t *unescape( const wchar_t * orig, int unescape_special )
 								return 0;
 							}
 							
+							case '\\':
 							case L'$':
+							case '"':
 							{
 								in[out_pos]=in[in_pos];
 								break;
 							}
-							
+
 							default:
 							{
 								in[out_pos++] = L'\\';
@@ -1466,13 +1384,13 @@ static char *gen_unique_nfs_filename( const char *filename )
 	pidlen = sprint_pid_t( getpid(), newname + orglen + 1 + hnlen + 1 );
 	newname[orglen + 1 + hnlen + 1 + pidlen] = '\0';
 /*	debug( 1, L"gen_unique_nfs_filename returning with: newname = \"%s\"; "
-	L"HOST_NAME_MAX = %d; hnlen = %d; orglen = %d; "
-	L"sizeof(pid_t) = %d; maxpiddigits = %d; malloc'd size: %d", 
-	newname, (int)HOST_NAME_MAX, hnlen, orglen, 
-	(int)sizeof(pid_t), 
-	(int)(0.31 * sizeof(pid_t) * CHAR_BIT + 1),
-	(int)(orglen + 1 + hnlen + 1 + 
-	(int)(0.31 * sizeof(pid_t) * CHAR_BIT + 1) + 1) ); */
+  L"HOST_NAME_MAX = %d; hnlen = %d; orglen = %d; "
+  L"sizeof(pid_t) = %d; maxpiddigits = %d; malloc'd size: %d", 
+  newname, (int)HOST_NAME_MAX, hnlen, orglen, 
+  (int)sizeof(pid_t), 
+  (int)(0.31 * sizeof(pid_t) * CHAR_BIT + 1),
+  (int)(orglen + 1 + hnlen + 1 + 
+  (int)(0.31 * sizeof(pid_t) * CHAR_BIT + 1) + 1) ); */
 	return newname;
 }
 
@@ -1499,6 +1417,9 @@ int acquire_lock_file( const char *lockfile, const int timeout, int force )
 		debug( 1, L"acquire_lock_file: open: %s", strerror( errno ) );
 		goto done;
 	}
+	/*
+	  Don't need to check exit status of close on read-only file descriptors
+	*/
 	close( fd );
 	if( stat( linkfile, &statbuf ) != 0 )
 	{
@@ -1581,10 +1502,15 @@ int acquire_lock_file( const char *lockfile, const int timeout, int force )
 
 void common_handle_winch( int signal )
 {
+#ifdef HAVE_WINSIZE
 	if (ioctl(1,TIOCGWINSZ,&termsize)!=0)
 	{
 		return;
 	}
+#else
+	termsize.ws_col = 80;
+	termsize.ws_row = 24;
+#endif
 }
 
 int common_get_width()
@@ -1598,3 +1524,29 @@ int common_get_height()
 	return termsize.ws_row;
 }
 
+void tokenize_variable_array( const wchar_t *val, array_list_t *out )
+{
+	if( val )
+	{
+		wchar_t *cpy = wcsdup( val );
+		wchar_t *pos, *start;
+
+		if( !cpy )
+		{
+			DIE_MEM();
+		}
+
+		for( start=pos=cpy; *pos; pos++ )
+		{
+			if( *pos == ARRAY_SEP )
+			{
+				*pos=0;
+				al_push( out, start==cpy?cpy:wcsdup(start) );
+				start=pos+1;
+			}
+		}
+		al_push( out, start==cpy?cpy:wcsdup(start) );
+	}
+}
+
+
