@@ -8,39 +8,37 @@
 */
 #define FISH_COMMON_H
 
+#include <stdlib.h>
+#include <stdio.h>
 #include <wchar.h>
 #include <termios.h>
 
 #include "util.h"
 
-/**
-   Under curses, tputs expects an int (*func)(char) as its last
-   parameter, but in ncurses, tputs expects a int (*func)(int) as its
-   last parameter. tputs_arg_t is defined to always be what tputs
-   expects. Hopefully.
-*/
-
-#ifdef NCURSES_VERSION
-typedef int tputs_arg_t;
-#else
-typedef char tputs_arg_t;
-#endif
-
 /**
    Maximum number of bytes used by a single utf-8 character
 */
 #define MAX_UTF8_BYTES 6
 
 /**
-   Color code for set_color. Does not update the color.
+   This is in the unicode private use area.
 */
-
-#define FISH_COLOR_IGNORE -1
+#define ENCODE_DIRECT_BASE 0xf100
 
 /**
-   Color code for set_color. Sets the default color.
+  Highest legal ascii value
 */
-#define FISH_COLOR_RESET -2
+#define ASCII_MAX 127u
+
+/**
+  Highest legal 16-bit unicode value
+*/
+#define UCS2_MAX 0xffffu
+
+/**
+  Highest legal byte value
+*/
+#define BYTE_MAX 0xffu
 
 /** 
 	Save the shell mode on startup so we can restore them on exit
@@ -53,11 +51,6 @@ extern struct termios shell_modes;
 */
 extern wchar_t ellipsis_char;
 
-/**
-   The maximum number of charset convertion errors to report
-*/
-extern int error_max;
-
 /**
    The verbosity of fish
 */
@@ -74,9 +67,51 @@ extern char *profile;
 */
 extern wchar_t *program_name;
 
+/**
+   This macro is used to check that an input argument is not null. It
+   is a bit lika a non-fatal form of assert. Instead of exit-ing on
+   failiure, the current function is ended at once. The second
+   parameter is the exit status of the current function on failiure.
+*/
+#define CHECK( arg, retval )					\
+	if( !(arg) )														\
+	{																	\
+		debug( 1,														\
+			   L"function %s called with null value for argument %s. "	\
+			   L"This is a bug. "										\
+			   L"If you can reproduce it, please send a bug report to %s.", \
+			   __func__,												\
+			   #arg,													\
+			   PACKAGE_BUGREPORT );										\
+		return retval;													\
+	}																	\
+		
+
+/**
+   Exit program at once, leaving an error message about running out of memory
+*/
+#define DIE_MEM()								\
+	{																	\
+		fwprintf( stderr, L"fish: Out of memory on line %d of file %s, shutting down fish\n", __LINE__, __FILE__ );	\
+		exit(1);														\
+	}																	\
+		
+/**
+   Shorthand for wgettext call
+*/
+#define _(wstr) wgettext(wstr)
+
+/**
+   Noop, used to tell xgettext that a string should be translated, even though it is not directly sent to wgettext.
+*/
+#define N_(wstr) wstr
+
 /**
    Take an array_list_t containing wide strings and converts them to a
-   single null-terminated wchar_t **.
+   single null-terminated wchar_t **. The array is allocated using
+   halloc, and uses the \c context parameter as context. If \c context
+   is not noll, all elements of the \c array_list_t are also
+   registered to \c context using \c halloc_register().
 */
 wchar_t **list_to_char_arr( array_list_t *l );
 
@@ -94,23 +129,49 @@ wchar_t **list_to_char_arr( array_list_t *l );
 int fgetws2( wchar_t **buff, int *len, FILE *f );
 
 /**
-   Sorts a list of wide strings according to the wcsfilecmp-function
-   from the util library
+   Sorts an array_list of wide strings according to the
+   wcsfilecmp-function from the util library
 */
 void sort_list( array_list_t *comp );
 
 /**
    Returns a newly allocated wide character string equivalent of the
    specified multibyte character string
+
+   This function encodes illegal character sequences in a reversible
+   way using the private use area.
 */
 wchar_t *str2wcs( const char *in );
 
+/**
+   Converts the narrow character string \c in into it's wide
+   equivalent, stored in \c out. \c out must have enough space to fit
+   the entire string.
+
+   This function encodes illegal character sequences in a reversible
+   way using the private use area.
+*/
+wchar_t *str2wcs_internal( const char *in, wchar_t *out );
+
 /**
    Returns a newly allocated multibyte character string equivalent of
    the specified wide character string
+
+   This function decodes illegal character sequences in a reversible
+   way using the private use area.
 */
 char *wcs2str( const wchar_t *in );
 
+/**
+   Converts the wide character string \c in into it's narrow
+   equivalent, stored in \c out. \c out must have enough space to fit
+   the entire string.
+
+   This function decodes illegal character sequences in a reversible
+   way using the private use area.
+*/
+char *wcs2str_internal( const wchar_t *in, char *out );
+
 /**
    Returns a newly allocated wide character string array equivalent of
    the specified multibyte character string array
@@ -134,101 +195,26 @@ wchar_t *wcsdupcat( const wchar_t *a, const wchar_t *b );
 */
 wchar_t *wcsdupcat2( const wchar_t *a, ... );
 
-/**
-   Returns a newly allocated wide character string wich is a copy of
-   the string in, but of length c or shorter. The returned string is
-   always null terminated, and the null is not included in the string
-   length.
-*/
-wchar_t *wcsndup( const wchar_t *in, int c );
-
-/**
-   Converts from wide char to digit in the specified base. If d is not
-   a valid digit in the specified base, return -1.
-*/
-long convert_digit( wchar_t d, int base );
-
-
-/**
-   Convert a wide character string to a number in the specified
-   base. This functions is the wide character string equivalent of
-   strtol. For bases of 10 or lower, 0..9 are used to represent
-   numbers. For bases below 36, a-z and A-Z are used to represent
-   numbers higher than 9. Higher bases than 36 are not supported.
-*/
-long wcstol(const wchar_t *nptr,
-			wchar_t **endptr,
-			int base);
-
-/**
-   Appends src to string dst of size siz (unlike wcsncat, siz is the
-   full size of dst, not space left).  At most siz-1 characters will be
-   copied.  Always NUL terminates (unless siz <= wcslen(dst)).  Returns
-   wcslen(src) + MIN(siz, wcslen(initial dst)).  If retval >= siz,
-   truncation occurred.
-
-   This is the OpenBSD strlcat function, modified for wide characters,
-   and renamed to reflect this change.
-
-*/
-size_t wcslcat( wchar_t *dst, const wchar_t *src, size_t siz );
-
-/**
-   Copy src to string dst of size siz.  At most siz-1 characters will
-   be copied.  Always NUL terminates (unless siz == 0).  Returns
-   wcslen(src); if retval >= siz, truncation occurred.
-
-   This is the OpenBSD strlcpy function, modified for wide characters,
-   and renamed to reflect this change. 
-*/
-size_t wcslcpy( wchar_t *dst, const wchar_t *src, size_t siz );
-
-/**
-   Create a duplicate string. Wide string version of strdup. Will
-   automatically exit if out of memory.
-*/
-wchar_t *wcsdup(const wchar_t *in);
-
-size_t wcslen(const wchar_t *in);
-
-/**
-   Case insensitive string compare function. Wide string version of
-   strcasecmp.
-
-   This implementation of wcscasecmp does not take into account
-   esoteric locales where uppercase and lowercase do not cleanly
-   transform between each other. Hopefully this should be fine since
-   fish only uses this function with one of the strings supplied by
-   fish and guaranteed to be a sane, english word. Using wcscasecmp on
-   a user-supplied string should be considered a bug.
-*/
-int wcscasecmp( const wchar_t *a, const wchar_t *b );
-
-/**
-   Case insensitive string compare function. Wide string version of
-   strncasecmp.
-
-   This implementation of wcsncasecmp does not take into account
-   esoteric locales where uppercase and lowercase do not cleanly
-   transform between each other. Hopefully this should be fine since
-   fish only uses this function with one of the strings supplied by
-   fish and guaranteed to be a sane, english word. Using wcsncasecmp on
-   a user-supplied string should be considered a bug.
-*/
-int wcsncasecmp( const wchar_t *a, const wchar_t *b, int count );
-
 /**
    Test if the given string is a valid variable name
 */
 
-int wcsvarname( wchar_t *str );
 
 /**
-   The prototype for this function is missing in some libc
-   implementations. Fish has a fallback implementation in case the
-   implementation is missing altogether.
+   Test if the given string is a valid variable name. 
+
+   \return null if this is a valid name, and a pointer to the first invalid character otherwise
 */
-int wcwidth( wchar_t c );
+
+wchar_t *wcsvarname( wchar_t *str );
+
+/**
+   Test if the given string is valid in a variable name 
+
+   \return 1 if this is a valid name, 0 otherwise
+*/
+
+int wcsvarchr( wchar_t chr );
 
 
 /**
@@ -238,8 +224,8 @@ int my_wcswidth( const wchar_t *c );
 
 /**
    This functions returns the end of the quoted substring beginning at
-   \c in. It can handle both single and double quotes. Returns 0 on
-   error.
+   \c in. The type of quoting character is detemrined by examining \c
+   in. Returns 0 on error.
 
    \param in the position of the opening quote
 */
@@ -264,9 +250,11 @@ void error_reset();
 const wchar_t *wsetlocale( int category, const wchar_t *locale );
 
 /**
-   Checks if \c needle is included in the list of strings specified
+   Checks if \c needle is included in the list of strings specified. A warning is printed if needle is zero.
 
    \param needle the string to search for in the list 
+
+   \return zero is needle is not found, of if needle is null, non-zero otherwise
 */
 int contains_str( const wchar_t *needle, ... );
 
@@ -276,22 +264,6 @@ int contains_str( const wchar_t *needle, ... );
 */
 int read_blocked(int fd, void *buf, size_t count);
 
-/**
-   This is for writing process notification messages. Has to write to
-   stdout, so clr_eol and such functions will work correctly. Not an
-   issue since this function is only used in interactive mode anyway.
-*/
-int writeb( tputs_arg_t b );
-
-/**
-   Exit program at once, leaving an error message about running out of memory
-*/
-void die_mem();
-
-/**
-  Clean up 
-*/
-void common_destroy();
 
 /**
    Issue a debug message with printf-style string formating and
@@ -318,8 +290,7 @@ void debug( int level, const wchar_t *msg, ... );
    \return The escaped string, or 0 if there is not enough memory
 */
 
-wchar_t *escape( const wchar_t *in, 
-				 int escape_all );
+wchar_t *escape( const wchar_t *in, int escape_all );
 
 /**
    Expand backslashed escapes and substitute them with their unescaped
@@ -335,16 +306,6 @@ wchar_t *escape( const wchar_t *in,
 wchar_t *unescape( const wchar_t * in, 
 				   int escape_special );
 
-/**
-   Block SIGCHLD. Calls to block/unblock may be nested, and only once the nest count reaches zero wiull the block be removed.
-*/
-void block();
-
-/**
-   undo call to block().
-*/
-void unblock();
-
 /**
    Attempt to acquire a lock based on a lockfile, waiting LOCKPOLLINTERVAL 
    milliseconds between polls and timing out after timeout seconds, 
@@ -381,10 +342,20 @@ int common_get_height();
 void common_handle_winch( int signal );
 
 /**
-   Write paragraph of output to screen. Ignore newlines in message and
-   perform internal line-breaking.
+   Write paragraph of output to the specified stringbuffer, and redo
+   the linebreaks to fit the current screen.
 */
-void write_screen( const wchar_t *msg );
+void write_screen( const wchar_t *msg, string_buffer_t *buff );
+
+/**
+   Tokenize the specified string into the specified array_list_t.
+   Each new element is allocated using malloc and must be freed by the
+   caller.
+   
+   \param val the input string. The contents of this string is not changed.
+   \param out the list in which to place the elements. 
+*/
+void tokenize_variable_array( const wchar_t *val, array_list_t *out );
 
 
 #endif
