@@ -1,10 +1,12 @@
 /** \file complete.h
 	Prototypes for functions related to tab-completion.
 
-	These functions are used for storing and retrieving tab-completion data, as well as for performing tab-completion.
+	These functions are used for storing and retrieving tab-completion
+	data, as well as for performing tab-completion.
 */
 
 #ifndef FISH_COMPLETE_H
+
 /**
    Header guard
 */
@@ -14,39 +16,60 @@
 
 #include "util.h"
 
-/** Use all completions */
+/** 
+	Use all completions 
+*/
 #define SHARED 0
-/** Do not use file completion */
+/** 
+	Do not use file completion 
+*/
 #define NO_FILES 1
-/** Require a parameter after completion */
+/** 
+	Require a parameter after completion 
+*/
 #define NO_COMMON 2
-/** Only use the argument list specifies with completion after option. This is the same as (NO_FILES & NO_COMMON) */
+/** 
+	Only use the argument list specifies with completion after
+	option. This is the same as (NO_FILES & NO_COMMON)
+*/
 #define EXCLUSIVE 3
 
-/** Command is a path */
+/** 
+	Command is a path 
+*/
 #define PATH 1
-/** Command is not a path */
+/** 
+	Command is not a path 
+*/
 #define COMMAND 0
 
-/** Separateor between completion and description*/
+/** 
+	Separator between completion and description
+*/
 #define COMPLETE_SEP L'\004'
-/** Separateor between completion and description*/
+
+/** 
+	Separator between completion and description
+*/
 #define COMPLETE_SEP_STR L"\004"
 
 /**
-   Character that separates the completion and description on programmable completions
+   Separator between completion items in fish_pager. This is used for
+   completion grouping, e.g. when putting completions with the same
+   descriptions on the same line.
+*/
+#define COMPLETE_ITEM_SEP L'\uf500'
+
+/**
+   Character that separates the completion and description on
+   programmable completions
 */
 #define PROG_COMPLETE_SEP L'\t'
 
 /**
-  Initializes various structures used for tab-completion.
+   Terminator for completions sent to the fish_pager
 */
-void complete_init();
-
-/**
-   Destroys various structures used for tab-completion and free()s the memory used by them.
-*/
-void complete_destroy();
+#define COMPLETE_TERMINATOR L'\006'
 
 /**
 
@@ -162,6 +185,6 @@ int complete_is_valid_argument( const wchar_t *str,
    \param cmd the command for which to load command-specific completions
    \param reload should the commands completions be reloaded, even if they where previously loaded. (This is set to true on actual completions, so that changed completion are updated in running shells)
 */
-void complete_load( wchar_t *cmd, int reload );
+void complete_load( const wchar_t *cmd, int reload );
 
 #endif
