@@ -25,10 +25,22 @@
 #include <termio.h>
 #endif
 
+#if HAVE_TERM_H
 #include <term.h>
+#elif HAVE_NCURSES_TERM_H
+#include <ncurses/term.h>
+#endif
+
+#if HAVE_LIBINTL_H
+#include <libintl.h>
+#endif
+
 #include <errno.h>
 
+
+#include "fallback.h"
 #include "util.h"
+
 #include "wutil.h"
 #include "proc.h"
 #include "common.h"
@@ -41,13 +53,13 @@
 #include "env_universal.h"
 #include "input_common.h"
 #include "event.h"
-#include "translate.h"
+
 #include "complete.h"
 
 /**
    Command used to start fishd
 */
-#define FISHD_CMD L"if which fishd >/dev/null; fishd ^/tmp/fishd.%s.log; end"
+#define FISHD_CMD L"if which fishd 2>/dev/null >/dev/null; fishd ^/tmp/fishd.log.%s; end"
 
 /**
    Value denoting a null string
@@ -96,7 +108,11 @@ typedef struct env_node
 typedef struct var_entry
 {
 	int export; /**< Whether the variable should be exported */
-	wchar_t val[0]; /**< The value of the variable */
+#if __STDC_VERSION__ < 199901L
+	wchar_t val[1]; /**< The value of the variable */
+#else
+	wchar_t val[]; /**< The value of the variable */
+#endif
 }
 	var_entry_t;
 
@@ -143,19 +159,6 @@ static buffer_t export_buffer;
 */
 static int has_changed = 1;
 
-/**
-   Free hash key and hash value
-*/
-static void clear_hash_entry( const void *key, const void *data )
-{
-	var_entry_t *entry = (var_entry_t *)data;	
-	if( entry->export )
-		has_changed = 1;
-	
-	free( (void *)key );
-	free( (void *)data );
-}
-
 /**
    This stringbuffer is used to store the value of dynamically
    generated variables, such as history.
@@ -173,6 +176,36 @@ static int get_names_show_exported;
 */
 static int get_names_show_unexported;
 
+/**
+   List of all locale variable names
+*/
+static const wchar_t *locale_variable[] =
+{
+	L"LANG",
+	L"LC_ALL",
+	L"LC_COLLATE",
+	L"LC_CTYPE",
+	L"LC_MESSAGES",
+	L"LC_MONETARY",
+	L"LC_NUMERIC",
+	L"LC_TIME",
+	(void *)0
+}
+	;
+
+/**
+   Free hash key and hash value
+*/
+static void clear_hash_entry( void *key, void *data )
+{
+	var_entry_t *entry = (var_entry_t *)data;	
+	if( entry->export )
+		has_changed = 1;
+	
+	free( (void *)key );
+	free( (void *)data );
+}
+
 /**
    When fishd isn't started, this function is provided to
    env_universal as a callback, it tries to start up fishd. It's
@@ -184,7 +217,7 @@ static void start_fishd()
 {
 	string_buffer_t cmd;
 	struct passwd *pw;
-		
+	
 	sb_init( &cmd );
 	pw = getpwuid(getuid());
 	
@@ -215,15 +248,6 @@ static mode_t get_umask()
 	return res;
 }
 
-/**
-   List of all locale variable names
-*/
-static const wchar_t *locale_variable[] =
-{
-	L"LANG", L"LC_ALL", L"LC_COLLATE", L"LC_CTYPE", L"LC_MESSAGES", L"LC_MONETARY", L"LC_NUMERIC", L"LC_TIME", (void *)0
-}
-	;
-
 /**
    Checks if the specified variable is a locale variable
 */
@@ -251,7 +275,14 @@ static void handle_locale()
 	*/
 	static const int cat[] = 
 		{
-			0, LC_ALL, LC_COLLATE, LC_CTYPE, LC_MESSAGES, LC_MONETARY, LC_NUMERIC, LC_TIME
+			0, 
+			LC_ALL, 
+			LC_COLLATE,
+			LC_CTYPE,
+			LC_MESSAGES,
+			LC_MONETARY,
+			LC_NUMERIC,
+			LC_TIME
 		}
 	;
 	
@@ -277,18 +308,21 @@ static void handle_locale()
 	
 	if( wcscmp( wsetlocale( LC_MESSAGES, (void *)0 ), old ) != 0 )
 	{
-		/* Make change known to gettext.  */
-		{
-			extern int  _nl_msg_cat_cntr;
-			++_nl_msg_cat_cntr;
-		}			
+
+		/* 
+		   Try to make change known to gettext. Both changing
+		   _nl_msg_cat_cntr and calling dcgettext might potentially
+		   tell some gettext implementation that the translation
+		   strings should be reloaded. We do both and hope for the
+		   best.
+		*/
+		extern int _nl_msg_cat_cntr;
+		++_nl_msg_cat_cntr;
+		dcgettext( "fish", "Changing language to English", LC_MESSAGES );
 		
 		if( is_interactive )
 		{
-			complete_destroy();
-			complete_init();
-			
-			debug( 0, _(L"Changing language to english") );
+			debug( 0, _(L"Changing language to English") );
 		}
 	}
 	free( old );
@@ -339,24 +373,104 @@ static void universal_callback( int type,
 	}
 }
 
+/**
+   Make sure the PATH variable contains the essaential directories
+*/
+static void setup_path()
+{
+	wchar_t *path;
+	
+	int i, j;
+	array_list_t l;
+
+	const wchar_t *path_el[] = 
+		{
+			L"/bin",
+			L"/usr/bin",
+			PREFIX L"/bin",
+			0
+		}
+	;
+
+	path = env_get( L"PATH" );
+	if( !path )
+	{
+		env_set( L"PATH", 0, ENV_EXPORT | ENV_GLOBAL );
+		path=0;
+	}	
+		
+	al_init( &l );
+	
+	if( path )
+		tokenize_variable_array( path, &l );
+	
+	for( j=0; path_el[j]; j++ )
+	{
+		int has_el=0;
+		
+		for( i=0; i<al_get_count( &l); i++ )
+		{
+			wchar_t * el = (wchar_t *)al_get( &l, i );
+			size_t len = wcslen( el );
+			while( (len > 0) && (el[len-1]==L'/') )
+				len--;
+			if( (wcslen( path_el[j] ) == len) && (wcsncmp( el, path_el[j], len)==0) )
+			{
+				has_el = 1;
+			}
+		}
+		
+		if( !has_el )
+		{
+			string_buffer_t b;
+
+			debug( 3, L"directory %ls was missing", path_el[j] );
+
+			sb_init( &b );
+			if( path )
+			{
+				sb_append( &b, path );
+			}
+			
+			sb_append2( &b,
+						ARRAY_SEP_STR,
+						path_el[j],
+						(void *)0 );
+			
+			env_set( L"PATH", (wchar_t *)b.buff, ENV_GLOBAL | ENV_EXPORT );
+			
+			sb_destroy( &b );
+			
+			al_foreach( &l, &free );
+			path = env_get( L"PATH" );
+			al_truncate( &l, 0 );
+			tokenize_variable_array( path, &l );			
+		}
+	}
+	
+	al_foreach( &l, &free );
+	al_destroy( &l );
+}
+
 void env_init()
 {
 	char **p;
 	struct passwd *pw;
-	wchar_t *uname, *path;
+	wchar_t *uname;
+	wchar_t *version;
+	
 
 	sb_init( &dyn_var );
-
 	b_init( &export_buffer );
 	
-	
 	/*
 	  These variables can not be altered directly by the user
 	*/
 	hash_init( &env_read_only, &hash_wcs_func, &hash_wcs_cmp );
-
+	
 	hash_put( &env_read_only, L"status", L"" );
 	hash_put( &env_read_only, L"history", L"" );
+	hash_put( &env_read_only, L"version", L"" );
 	hash_put( &env_read_only, L"_", L"" );
 	hash_put( &env_read_only, L"LINES", L"" );
 	hash_put( &env_read_only, L"COLUMNS", L"" );
@@ -387,6 +501,11 @@ void env_init()
 	hash_init( &top->env, &hash_wcs_func, &hash_wcs_cmp );
 	global_env = top;
 	global = &top->env;	
+
+	/*
+	  Now the environemnt variable handling is set up, the next step
+	  is to insert valid data
+	*/
 	
 	/*
 	  Import environment variables
@@ -424,77 +543,14 @@ void env_init()
 		free(key);
 	}
 
-	path = env_get( L"PATH" );
-	if( !path )
-	{
-		env_set( L"PATH", L"/bin" ARRAY_SEP_STR L"/usr/bin", ENV_EXPORT | ENV_GLOBAL );
-		path = env_get( L"PATH" );
-	}
-	else
-	{
-		int i, j;
-		array_list_t l;
-		
-		al_init( &l );
-		expand_variable_array( path, &l );
+	/*
+	  Set up the PATH variable
+	*/
+	setup_path();
 
-		debug( 3, L"PATH is %ls", path );
-		
-		
-		const wchar_t *path_el[] = 
-			{
-				L"/bin",
-				L"/usr/bin",
-				PREFIX L"/bin",
-				0
-			}
-		;
-		
-		for( j=0; path_el[j]; j++ )
-		{
-			int has_el=0;
-
-			debug( 3, L"Check directory %ls", path_el[j] );
-		
-			
-			for( i=0; i<al_get_count( &l); i++ )
-			{
-				wchar_t * el = (wchar_t *)al_get( &l, i );
-				size_t len = wcslen( el );
-				while( (len > 0) && (el[len-1]==L'/') )
-					len--;
-				if( (wcslen( path_el[j] ) == len) && (wcsncmp( el, path_el[j], len)==0) )
-				{
-					has_el = 1;
-				}
-			}
-			
-			if( !has_el )
-			{
-				string_buffer_t b;
-				debug( 3, L"directory %ls was missing", path_el[j] );
-				sb_init( &b );
-				sb_append2( &b, path,
-							ARRAY_SEP_STR,
-							path_el[j],
-							(void *)0 );
-								
-				env_set( L"PATH", (wchar_t *)b.buff, ENV_GLOBAL | ENV_EXPORT );
-				sb_destroy( &b );
-				path = env_get( L"PATH" );
-				
-			}
-		}
-		
-		debug( 3, L"After: PATH is %ls", path );
-
-		al_foreach( &l, (void (*)(const void *))&free );
-		al_destroy( &l );
-		
-	}
-	
-	
-	
+	/*
+	  Set up the USER variable
+	*/
 	pw = getpwuid( getuid() );
 	if( pw )
 	{
@@ -502,6 +558,13 @@ void env_init()
 		env_set( L"USER", uname, ENV_GLOBAL | ENV_EXPORT );
 		free( uname );
 	}
+
+	/*
+	  Set up the version variable
+	*/
+	version = str2wcs( PACKAGE_VERSION );
+	env_set( L"version", version, ENV_GLOBAL );
+	free( version );
 	
 	env_universal_init( env_get( L"FISHD_SOCKET_DIR"), 
 						env_get( L"USER" ),
@@ -561,9 +624,9 @@ static env_node_t *env_get_node( const wchar_t *key )
 	return 0;
 }
 
-void env_set( const wchar_t *key, 
-			  const wchar_t *val, 
-			  int var_mode )
+int env_set( const wchar_t *key, 
+			 const wchar_t *val, 
+			 int var_mode )
 {
 	int free_val = 0;
 	var_entry_t *entry;
@@ -576,10 +639,12 @@ void env_set( const wchar_t *key,
 	event_t ev;
 	int is_universal = 0;	
 
+	CHECK( key, ENV_INVALID );
+		
 	if( (var_mode & ENV_USER ) && 
 		hash_get( &env_read_only, key ) )
 	{
-		return;
+		return ENV_PERM;
 	}
 	
 	if( wcscmp( key, L"umask" ) == 0)
@@ -603,13 +668,13 @@ void env_set( const wchar_t *key,
 		/*
 		  Do not actually create a umask variable, on env_get, it will be calculated dynamically
 		*/
-		return;
+		return 0;
 	}
 
 	/*
 	  Zero element arrays are internaly not coded as null but as this placeholder string
 	*/
-	if( !val && ( var_mode & ENV_USER ) )
+	if( !val )
 	{
 		val = ENV_NULL;
 	}
@@ -696,12 +761,13 @@ void env_set( const wchar_t *key,
 				else
 				{
 					/*
-					  New variable with unspecified scope. The default scope is the innermost scope that is shadowing
+					  New variable with unspecified scope. The default
+					  scope is the innermost scope that is shadowing
 					*/
 					node = top;
 					while( node->next && !node->new_scope )
 						node = node->next;
-				
+					
 				}
 			}
 		}
@@ -709,7 +775,7 @@ void env_set( const wchar_t *key,
 		if( !done )
 		{
 			void *k, *v;
-			hash_remove( &node->env, key, (const void **)&k, (const void **)&v );
+			hash_remove( &node->env, key, &k, &v );
 			free( k );
 			free( v );
 
@@ -749,6 +815,7 @@ void env_set( const wchar_t *key,
 		
 		al_init( &ev.arguments );
 		al_push( &ev.arguments, L"VARIABLE" );
+		al_push( &ev.arguments, L"SET" );
 		al_push( &ev.arguments, key );
 		
 //	debug( 1, L"env_set: fire events on variable %ls", key );	
@@ -761,27 +828,35 @@ void env_set( const wchar_t *key,
 	{
 		handle_locale();
 	}
-		
+
+	return 0;
 }
 
 
-
-
 /**
    Attempt to remove/free the specified key/value pair from the
    specified hash table.
+
+   \return zero if the variable was not found, non-zero otherwise
 */
 static int try_remove( env_node_t *n,
-					   const wchar_t *key )
+					   const wchar_t *key,
+					   int var_mode )
 {
+	void *old_key_void, *old_val_void;
 	wchar_t *old_key, *old_val;
+
 	if( n == 0 )
 		return 0;
 
 	hash_remove( &n->env, 
 				 key,
-				 (const void **)&old_key, 
-				 (const void **)&old_val );
+				 &old_key_void, 
+				 &old_val_void );
+
+	old_key = (wchar_t *)old_key_void;
+	old_val = (wchar_t *)old_val_void;
+	
 	if( old_key != 0 )
 	{
 		var_entry_t * v = (var_entry_t *)old_val;
@@ -795,29 +870,72 @@ static int try_remove( env_node_t *n,
 		return 1;
 	}
 
+	if( var_mode & ENV_LOCAL )
+		return 0;
+	
 	if( n->new_scope )
-		return try_remove( global_env, key );
+		return try_remove( global_env, key, var_mode );
 	else
-		return try_remove( n->next, key );
+		return try_remove( n->next, key, var_mode );
 }
 
 
-void env_remove( const wchar_t *key, int var_mode )
+int env_remove( const wchar_t *key, int var_mode )
 {
+	env_node_t *first_node;
+	int erased = 0;
+	
+	CHECK( key, 1 );
+		
 	if( (var_mode & ENV_USER ) && 
 		hash_get( &env_read_only, key ) )
 	{
-		return;
+		return 2;
 	}
 	
-	if( !try_remove( top, key ) )
+	first_node = top;
+	
+	if( ! (var_mode & ENV_UNIVERSAL ) )
 	{
-		env_universal_remove( key );
+		
+		if( var_mode & ENV_GLOBAL )
+		{
+			first_node = global_env;
+		}
+		
+		if( try_remove( first_node, key, var_mode ) )
+		{		
+			event_t ev;
+			
+			ev.type=EVENT_VARIABLE;
+			ev.param1.variable=key;
+			ev.function_name=0;
+			
+			al_init( &ev.arguments );
+			al_push( &ev.arguments, L"VARIABLE" );
+			al_push( &ev.arguments, L"ERASE" );
+			al_push( &ev.arguments, key );
+			
+			event_fire( &ev );	
+			
+			al_destroy( &ev.arguments );
+			erased = 1;
+		}
+	}
+	
+	if( !erased && 
+		!(var_mode & ENV_GLOBAL) &&
+		!(var_mode & ENV_LOCAL) ) 
+	{
+		erased = !env_universal_remove( key );
 	}
 
 	if( is_locale( key ) )
+	{
 		handle_locale();
-			
+	}
+	
+	return !erased;	
 }
 
 
@@ -827,6 +945,8 @@ wchar_t *env_get( const wchar_t *key )
 	env_node_t *env = top;
 	wchar_t *item;
 	
+	CHECK( key, 0 );
+		
 	if( wcscmp( key, L"history" ) == 0 )
 	{
 		wchar_t *current;
@@ -846,7 +966,10 @@ wchar_t *env_get( const wchar_t *key )
 			wchar_t *next = history_get( i-add_current );
 			if( !next )
 			{
-				debug( 1, _( L"No history item at index %d\n" ), i );
+				/*
+				  This is not an error - it simply means the user has
+				  a short history
+				*/
 				break;
 			}
 			
@@ -916,40 +1039,62 @@ wchar_t *env_get( const wchar_t *key )
 		return item;
 }
 
-int env_exist( const wchar_t *key )
+int env_exist( const wchar_t *key, int mode )
 {
 	var_entry_t *res;
-	env_node_t *env = top;
-	wchar_t *item;
-	
-    if( hash_get( &env_read_only, key ) || hash_get( &env_electric, key ) )
-    {
-        return 1;
-    }
-	
-	while( env != 0 )
+	env_node_t *env;
+	wchar_t *item=0;
+
+	CHECK( key, 0 );
+		
+	/*
+	  Read only variables all exist, and they are all global. A local
+	  version can not exist.
+	*/
+	if( ! (mode & ENV_LOCAL) && ! (mode & ENV_UNIVERSAL) )
 	{
-		res = (var_entry_t *) hash_get( &env->env, 
-										key );
-		if( res != 0 )
+		if( hash_get( &env_read_only, key ) || hash_get( &env_electric, key ) )
 		{
 			return 1;
 		}
-		
-		if( env->new_scope )
-			env = global_env;
-		else
-			env = env->next;
-	}	
-	if( !proc_had_barrier)
+	}
+
+	if( ! (mode & ENV_UNIVERSAL) )
 	{
-		proc_had_barrier=1;
-		env_universal_barrier();
+		env = (mode & ENV_GLOBAL)?global_env:top;
+					
+		while( env != 0 )
+		{
+			res = (var_entry_t *) hash_get( &env->env, 
+											key );
+			if( res != 0 )
+			{
+				return 1;
+			}
+			
+			if( mode & ENV_LOCAL )
+				break;
+			
+			if( env->new_scope )
+				env = global_env;
+			else
+				env = env->next;
+		}	
 	}
 	
-	item = env_universal_get( key );
+	if( ! (mode & ENV_LOCAL) && ! (mode & ENV_GLOBAL) )
+	{
+		if( !proc_had_barrier)
+		{
+			proc_had_barrier=1;
+			env_universal_barrier();
+		}
+		
+		item = env_universal_get( key );
 	
+	}
 	return item != 0;
+
 }
 
 /**
@@ -1031,20 +1176,22 @@ void env_pop()
    Function used with hash_foreach to insert keys of one table into
    another
 */
-static void add_key_to_hash( const void *key, 
-							 const void *data,
+static void add_key_to_hash( void *key, 
+							 void *data,
 							 void *aux )
 {
 	var_entry_t *e = (var_entry_t *)data;
 	if( ( e->export && get_names_show_exported) || 
 		( !e->export && get_names_show_unexported) )
+	{
 		hash_put( (hash_table_t *)aux, key, 0 );
+	}
 }
 
 /**
    Add key to hashtable
 */
-static void add_to_hash( const void *k, void *aux )
+static void add_to_hash( void *k, void *aux )
 {
 	hash_put( (hash_table_t *)aux,
 			  k,
@@ -1054,8 +1201,8 @@ static void add_to_hash( const void *k, void *aux )
 /**
    Add key to list
 */
-static void add_key_to_list( const void * key, 
-							 const void * val, 
+static void add_key_to_list( void * key, 
+							 void * val, 
 							 void *aux )
 {
 	al_push( (array_list_t *)aux, key );
@@ -1071,6 +1218,8 @@ void env_get_names( array_list_t *l, int flags )
 	hash_table_t names;
 	env_node_t *n=top;
 
+	CHECK( l, );
+	
 	get_names_show_exported = 
 		flags & ENV_EXPORT|| (!(flags & ENV_UNEXPORT));
 	get_names_show_unexported = 
@@ -1139,7 +1288,7 @@ void env_get_names( array_list_t *l, int flags )
 /**
    Function used by env_export_arr to iterate over hashtable of variables
 */
-static void export_func1( const void *k, const void *v, void *aux )
+static void export_func1( void *k, void *v, void *aux )
 {
 	var_entry_t *val_entry = (var_entry_t *)v;
 	if( val_entry->export )
@@ -1155,7 +1304,7 @@ static void export_func1( const void *k, const void *v, void *aux )
 /**
    Function used by env_export_arr to iterate over hashtable of variables
 */
-static void export_func2( const void *k, const void *v, void *aux )
+static void export_func2( void *k, void *v, void *aux )
 {
 	wchar_t *key = (wchar_t *)k;
 	wchar_t *val = (wchar_t *)v;
@@ -1169,7 +1318,7 @@ static void export_func2( const void *k, const void *v, void *aux )
 
 	if( !ks || !vs )
 	{
-		die_mem();
+		DIE_MEM();
 	}
 	
 	/*
@@ -1192,7 +1341,7 @@ static void export_func2( const void *k, const void *v, void *aux )
 	free( vs );
 }
 
-char **env_export_arr( int recalc)
+char **env_export_arr( int recalc )
 {
 	if( recalc && !proc_had_barrier)
 	{
