@@ -41,6 +41,16 @@
 */
 #define ENV_UNIVERSAL 32
 
+/**
+   Error code for trying to alter read-only variable
+*/
+enum{
+	ENV_PERM = 1,
+	ENV_INVALID
+}
+;
+
+
 /**
    Initialize environment variable data
 */
@@ -61,9 +71,15 @@ void env_destroy();
    \param val The value
    \param mode The type of the variable. Can be any combination of ENV_GLOBAL, ENV_LOCAL, ENV_EXPORT and ENV_USER. If mode is zero, the current variable space is searched and the current mode is used. If no current variable with the same name is found, ENV_LOCAL is assumed.
 
+   \returns 0 on suicess or an error code on failiure.
+
+   The current error codes are:
+
+   * ENV_PERM, can only be returned when setting as a user, e.g. ENV_USER is set. This means that the user tried to change a read-only variable.
+   * ENV_INVALID, the variable name or mode was invalid
 */
 
-void env_set( const wchar_t *key, 
+int env_set( const wchar_t *key, 
 			  const wchar_t *val,
 			  int mode );
 
@@ -80,16 +96,21 @@ wchar_t *env_get( const wchar_t *key );
 /**
    Returns 1 if the specified key exists. This can't be reliable done
    using env_get, since env_get returns null for 0-element arrays
+
+   \param key The name of the variable to remove
+   \param mode the scope to search in. All scopes are searched if unset
 */
-int env_exist( const wchar_t *key );
+int env_exist( const wchar_t *key, int mode );
 
 /**
    Remove environemnt variable
    
    \param key The name of the variable to remove
-   \param mode should be ENV_USER if this is a remove request from the user, 0 otherwise. If this is a user request, read-only variables can not be removed.
+   \param mode should be ENV_USER if this is a remove request from the user, 0 otherwise. If this is a user request, read-only variables can not be removed. The mode may also specify the scope of the variable that should be erased.
+
+   \return zero if the variable existed, and non-zero if the variable did not exist
 */
-void env_remove( const wchar_t *key, int mode );
+int env_remove( const wchar_t *key, int mode );
 
 /**
   Push the variable stack. Used for implementing local variables for functions and for-loops.
@@ -108,7 +129,6 @@ char **env_export_arr( int recalc );
 
 /**
   Insert all variable names into l. These are not copies of the strings and should not be freed after use.
-
 */
 void env_get_names( array_list_t *l, int flags );
 
