@@ -5,6 +5,8 @@
 	manual, though I the changes performed have been massive.
 */
 
+#include "config.h"
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <sys/types.h>
@@ -21,8 +23,13 @@
 #include <assert.h>
 #include <dirent.h>
 
-#include "config.h"
+#ifdef HAVE_SIGINFO_H
+#include <siginfo.h>
+#endif
+
+#include "fallback.h"
 #include "util.h"
+
 #include "common.h"
 #include "wutil.h"
 #include "proc.h"
@@ -36,7 +43,10 @@
 #include "expand.h"
 #include "signal.h"
 #include "env_universal.h"
-#include "translate.h"
+
+#include "halloc.h"
+#include "halloc_util.h"
+#include "parse_util.h"
 
 /**
    Prototype for the getpgid library function. The prototype for this
@@ -67,6 +77,7 @@ pid_t getpgid( pid_t pid );
 */
 static array_list_t *open_fds=0;
 
+static int set_child_group( job_t *j, process_t *p, int print_errors );
 
 void exec_close( int fd )
 {
@@ -78,20 +89,20 @@ void exec_close( int fd )
 		{
 			debug( 1, FD_ERROR, fd );
 			wperror( L"close" );
+			break;
 		}
 	}
-
+	
 	if( open_fds )
 	{
 		for( i=0; i<al_get_count( open_fds ); i++ )
 		{
-			int n = (int)(long)al_get( open_fds, i );
+			int n = (int)al_get_long( open_fds, i );
 			if( n == fd )
 			{
-				al_set( open_fds,
-						i,
-						al_get( open_fds,
-								al_get_count( open_fds ) -1 ) );
+				al_set_long( open_fds,
+							 i,
+							 al_get_long( open_fds, al_get_count( open_fds ) -1 ) );
 				al_truncate( open_fds, 
 							 al_get_count( open_fds ) -1 );
 				break;
@@ -118,11 +129,11 @@ int exec_pipe( int fd[2])
 	
 	if( open_fds == 0 )
 	{
-		open_fds = al_new();
+		open_fds = al_halloc( global_context );
 	}
 	
-	al_push( open_fds, (void *)(long)fd[0] );
-	al_push( open_fds, (void *)(long)fd[1] );		
+	al_push_long( open_fds, (long)fd[0] );
+	al_push_long( open_fds, (long)fd[1] );		
 	
 	return res;
 }
@@ -156,7 +167,8 @@ static int use_fd_in_pipe( int fd, io_data_t *io )
    the redirection list io. This should make sure that there are no
    stray opened file descriptors in the child.
    
-   \param io the list of io redirections for this job. Pipes mentioned here should not be closed.
+   \param io the list of io redirections for this job. Pipes mentioned
+   here should not be closed.
 */
 static void close_unused_internal_pipes( io_data_t *io )
 {
@@ -166,7 +178,7 @@ static void close_unused_internal_pipes( io_data_t *io )
 	{
 		for( ;i<al_get_count( open_fds ); i++ )
 		{
-			int n = (int)(long)al_get( open_fds, i );
+			int n = (long)al_get_long( open_fds, i );
 			if( !use_fd_in_pipe( n, io) )
 			{
 				debug( 4, L"Close fd %d, used in other context", n );
@@ -177,20 +189,6 @@ static void close_unused_internal_pipes( io_data_t *io )
 	}
 }
 
-void exec_init()
-{
-}
-
-void exec_destroy()
-{
-	if( open_fds )
-	{
-		al_destroy( open_fds );
-		free( open_fds );
-	}
-}
-
-
 /**
    Make sure the fd used by this redirection is not used by i.e. a pipe. 
 */
@@ -238,7 +236,7 @@ void free_fd( io_data_t *io, int fd )
    \param io the list of IO redirections for the child
    \param exit_on_error whether to call exit() on errors
 
-   \return 1 on sucess, 0 on failiure
+   \return 0 on sucess, -1 on failiure
 */
 static int handle_child_io( io_data_t *io, int exit_on_error )
 {
@@ -257,7 +255,7 @@ static int handle_child_io( io_data_t *io, int exit_on_error )
 		if( io->fd > 2 )
 		{
 			/*
-			  Make sure the fd used by this redirection is not used by i.e. a pipe. 
+			  Make sure the fd used by this redirection is not used by e.g. a pipe. 
 			*/
 			free_fd( io, io->fd );
 		}
@@ -265,7 +263,11 @@ static int handle_child_io( io_data_t *io, int exit_on_error )
 		switch( io->io_mode )
 		{
 			case IO_CLOSE:
-				close(io->fd);
+				if( close(io->fd) )
+				{
+					debug( 0, _(L"Failed to close file descriptor %d"), io->fd );
+					wperror( L"close" );
+				}
 				break;
 			case IO_FILE:
 			{
@@ -283,13 +285,17 @@ static int handle_child_io( io_data_t *io, int exit_on_error )
 					}
 					else
 					{
-						return 0;
+						return -1;
 					}					
 				}				
 				else if( tmp != io->fd)
 				{
+					/*
+					  This call will sometimes fail, but that is ok,
+					  this is just a precausion.
+					*/
 					close(io->fd);
-		
+							
 					if(dup2( tmp, io->fd ) == -1 )
 					{
 						debug( 1, 
@@ -302,7 +308,7 @@ static int handle_child_io( io_data_t *io, int exit_on_error )
 						}
 						else
 						{
-							return 0;
+							return -1;
 						}
 					}
 					exec_close( tmp );
@@ -312,6 +318,10 @@ static int handle_child_io( io_data_t *io, int exit_on_error )
 			
 			case IO_FD:
 			{
+				/*
+				  This call will sometimes fail, but that is ok,
+				  this is just a precausion.
+				*/
 				close(io->fd);
 
 				if( dup2( io->param1.old_fd, io->fd ) == -1 )
@@ -326,7 +336,7 @@ static int handle_child_io( io_data_t *io, int exit_on_error )
 					}
 					else
 					{
-						return 0;
+						return -1;
 					}
 				}
 				break;
@@ -337,6 +347,10 @@ static int handle_child_io( io_data_t *io, int exit_on_error )
 			{
 				int fd_to_dup = io->fd;
 
+				/*
+				  This call will sometimes fail, but that is ok,
+				  this is just a precausion.
+				*/
 				close(io->fd);
 				
 				if( dup2( io->param1.pipe_fd[fd_to_dup?1:0], io->fd ) == -1 )
@@ -349,7 +363,7 @@ static int handle_child_io( io_data_t *io, int exit_on_error )
 					}
 					else
 					{
-						return 0;
+						return -1;
 					}					
 
 				}
@@ -373,7 +387,7 @@ static int handle_child_io( io_data_t *io, int exit_on_error )
 	if( env_universal_server.fd >= 0 )
 		exec_close( env_universal_server.fd );
 
-	return 1;
+	return 0;
 	
 }
 
@@ -385,43 +399,26 @@ static int handle_child_io( io_data_t *io, int exit_on_error )
    IO redirections and other file descriptor actions are performed.
 
    \param j the job to set up the IO for
-   \param exit_on_error whether to call exit() on errors
+   \param p the child process to set up
 
-   \return 1 on sucess, 0 on failiure
+   \return 0 on sucess, -1 on failiure
 */
 static int setup_child_process( job_t *j, process_t *p )
 {
-	int res;
+	int res=0;
 	
-	if( is_interactive  && p->type==EXTERNAL )
-    {
-		pid_t pid;
-		/* 
-		   Put the process into the process group and give the process
-		   group the terminal, if appropriate.  This has to be done
-		   both by the shell and in the individual child processes
-		   because of potential race conditions.
-		*/
-		pid = getpid ();
-		if (j->pgid == 0)
-			j->pgid = pid;
-
-		/* Wait till shell puts os in our own group */
-		while( getpgrp() != j->pgid )
-			sleep(0);
-		
-		/* Wait till shell gives us stdin */
-		if ( j->fg )
-		{
-			while( tcgetpgrp( 0 ) != j->pgid )
-				sleep(0);
-		}
+	if( p )
+	{
+		res = set_child_group( j, p, 1 );
+	}
+	
+	if( !res )	
+	{
+		res = handle_child_io( j->io, (p!=0) );
 	}
 	
-	res = handle_child_io( j->io, (p==0) );
-
 	/* Set the handling for job control signals back to the default.  */
-	if( res )
+	if( !res )
 	{
 		signal_reset_handlers();
 	}
@@ -442,8 +439,11 @@ static int setup_child_process( job_t *j, process_t *p )
 */
 static void launch_process( process_t *p )
 {
+//	debug( 1, L"exec '%ls'", p->argv[0] );
 		
-	execve (wcs2str(p->actual_cmd), wcsv2strv( (const wchar_t **) p->argv), env_export_arr( 0 ) );
+	execve ( wcs2str(p->actual_cmd), 
+			 wcsv2strv( (const wchar_t **) p->argv), 
+			 env_export_arr( 0 ) );
 	debug( 0, 
 		   _( L"Failed to execute process '%ls'" ),
 		   p->actual_cmd );
@@ -499,7 +499,7 @@ static io_data_t *io_transmogrify( io_data_t * in )
 	
 	out = malloc( sizeof( io_data_t ) );
 	if( !out )
-		die_mem();
+		DIE_MEM();
 	
 	out->fd = in->fd;
 	out->io_mode = IO_FD;
@@ -572,7 +572,7 @@ static void internal_exec_helper( const wchar_t *def,
 	io_data_t *io_internal = io_transmogrify( io );
 	int is_block_old=is_block;
 	is_block=1;
-
+	
 	/*
 	  Did the transmogrification fail - if so, set error status and return
 	*/
@@ -583,7 +583,7 @@ static void internal_exec_helper( const wchar_t *def,
 	}
 	
 	signal_unblock();
-		
+	
 	eval( def, io_internal, block_type );		
 	
 	signal_block();
@@ -594,65 +594,63 @@ static void internal_exec_helper( const wchar_t *def,
 }
 
 /**
-   This function should be called by the parent process right after
-   fork() has been called. If job control is enabled, the child is put
-   in the jobs group.
+   This function should be called by both the parent process and the
+   child right after fork() has been called. If job control is
+   enabled, the child is put in the jobs group, and if the child is
+   also in the foreground, it is also given control of the
+   terminal. When called in the parent process, this function may
+   fail, since the child might have already finished and called
+   exit. The parent process may safely ignore the exit status of this
+   call.
+
+   Returns 0 on sucess, -1 on failiure. 
 */
-static int handle_new_child( job_t *j, process_t *p )
+static int set_child_group( job_t *j, process_t *p, int print_errors )
 {
+	int res = 0;
 	
-	if( is_interactive  && p->type==EXTERNAL )
+	if( j->job_control )
 	{
-		int new_pgid=0;
-		
 		if (!j->pgid)
 		{
-			new_pgid=1;			
 			j->pgid = p->pid;
 		}
 		
 		if( setpgid (p->pid, j->pgid) )
-		{						
-			if( getpgid( p->pid) != j->pgid )
+		{
+			if( getpgid( p->pid) != j->pgid && print_errors )
 			{
 				debug( 1, 
-					   _( L"Could not send process %d from group %d to group %d" ),
-					   p->pid, 
+					   _( L"Could not send process %d, '%ls' in job %d, '%ls' from group %d to group %d" ),
+					   p->pid,
+					   p->argv[0],
+					   j->job_id,
+					   j->command,
 					   getpgid( p->pid),
 					   j->pgid );
 				wperror( L"setpgid" );
+				res = -1;
 			}
 		}
-
-		if( j->fg )
-		{
-			if( tcsetpgrp (0, j->pgid) )
-			{
-				debug( 1, _( L"Could not send job %d ('%ls') to foreground" ), 
-					   j->job_id, 
-					   j->command );
-				wperror( L"tcsetpgrp" );
-				return -1;
-			}
-		}
-
-		if( j->fg && new_pgid)
-		{
-			if( tcsetpgrp (0, j->pgid) )
-			{
-				debug( 1, _( L"Could not send job %d ('%ls') to foreground" ), 
-					   j->job_id, 
-					   j->command );
-				wperror( L"tcsetpgrp" );
-				return -1;
-			}
-		}		
 	}
 	else
 	{
 		j->pgid = getpid();
 	}
-	return 0;
+
+	if( j->terminal && j->fg )
+	{
+		if( tcsetpgrp (0, j->pgid) && print_errors )
+		{
+			debug( 1, _( L"Could not send job %d ('%ls') to foreground" ), 
+				   j->job_id, 
+				   j->command );
+			wperror( L"tcsetpgrp" );
+			res = -1;
+		}
+	}
+
+	return res;
 }
 
 
@@ -663,8 +661,6 @@ void exec( job_t *j )
 	pid_t pid;
 	int mypipe[2];
 	sigset_t chldset; 
-	sigemptyset( &chldset );
-	sigaddset( &chldset, SIGCHLD );
 	int skip_fork;
 	
 	io_data_t pipe_read, pipe_write;
@@ -677,9 +673,48 @@ void exec( job_t *j )
 	*/
 	int exec_error=0;
 
+	int needs_keepalive = 0;
+	process_t keepalive;
 	
-	debug( 4, L"Exec job %ls with id %d", j->command, j->job_id );	
+
+	CHECK( j, );
+		
+	if( no_exec )
+		return;
 	
+	
+
+	sigemptyset( &chldset );
+	sigaddset( &chldset, SIGCHLD );
+	
+	debug( 4, L"Exec job '%ls' with id %d", j->command, j->job_id );	
+	
+	if( block_io )
+	{
+		if( j->io )
+			j->io = io_add( io_duplicate( j, block_io), j->io );
+		else
+			j->io=io_duplicate( j, block_io);				
+	}
+
+	io_data_t *input_redirect = io_get( j->io, 0 );
+	
+	if( input_redirect && 
+		(input_redirect->io_mode == IO_BUFFER) && 
+		input_redirect->param3.is_input )
+	{
+		/*
+		  Input redirection - create a new gobetween process to take
+		  care of buffering
+		*/
+		
+		process_t *fake = halloc( j, sizeof(process_t) );
+		fake->type = INTERNAL_BUFFER;
+		fake->pipe_fd = 1;
+		fake->next = j->first_process;
+		j->first_process = fake;
+	}
+
 	if( j->first_process->type==INTERNAL_EXEC )
 	{
 		/*
@@ -688,9 +723,10 @@ void exec( job_t *j )
 		signal_block();
 
 		/*
-		  setup_child_process make sure signals are propelry set up
+		  setup_child_process makes sure signals are properly set
+		  up. It will also call signal_unblock
 		*/
-		if( setup_child_process( j, 0 ) )
+		if( !setup_child_process( j, 0 ) )
 		{
 			/*
 			  launch_process never returns
@@ -703,9 +739,8 @@ void exec( job_t *j )
 			j->first_process->completed=1;
 			return;
 		}
-		
-	}
-	
+
+	}	
 
 	pipe_read.fd=0;
 	pipe_write.fd=1;
@@ -717,20 +752,61 @@ void exec( job_t *j )
 	pipe_write.next=0;
 	pipe_write.param1.pipe_fd[0]=pipe_write.param1.pipe_fd[1]=-1;	
 
-	//fwprintf( stderr, L"Run command %ls\n", j->command );
+
 	
-	if( block_io )
-	{
-		if( j->io )
-			j->io = io_add( io_duplicate(block_io), j->io );
-		else
-			j->io=io_duplicate(block_io);				
-	}
+	//fwprintf( stderr, L"Run command %ls\n", j->command );
 	
 	j->io = io_add( j->io, &pipe_write );
 	
 	signal_block();
 
+	/*
+	  See if we need to create a group keepalive process. This is a
+	  process that we create to make sure that the process group
+	  doesn't die accidentally, and is needed when a block/function is
+	  inside a pipeline.
+	*/
+	
+	if( j->job_control )
+	{
+		for( p=j->first_process; p; p = p->next )
+		{
+			if( (p->type == INTERNAL_BLOCK ) || 
+				(p->type == INTERNAL_FUNCTION ) )
+			{
+				if( p->next )
+				{
+					needs_keepalive = 1;
+					break;
+				}
+			}
+		}
+	}
+		
+	if( needs_keepalive )
+	{
+		keepalive.pid = fork();
+
+		if( keepalive.pid == 0 )
+		{
+			keepalive.pid = getpid();
+			set_child_group( j, &keepalive, 1 );
+			pause();			
+			exit(0);
+		}
+		else if( keepalive.pid < 0 )
+		{
+			/* The fork failed. */
+			debug( 0, FORK_ERROR );
+			wperror (L"fork");
+			exit (1);
+		}
+		else
+		{
+			set_child_group( j, &keepalive, 0 );			
+		}
+	}
+	
 	/*
 	  This loop loops over every process_t in the job, starting it as
 	  appropriate. This turns out to be rather complex, since a
@@ -743,7 +819,7 @@ void exec( job_t *j )
 	{
 		mypipe[1]=-1;
 		skip_fork=0;
-
+		
 		pipe_write.fd = p->pipe_fd;
 
 		/* 
@@ -777,22 +853,12 @@ void exec( job_t *j )
 				break;
 			}
 
-/*			fwprintf( stderr,
-  L"Make pipe from %ls to %ls using fds %d and %d\n", 
-  p->actual_cmd,
-  p->next->actual_cmd,
-  mypipe[0],
-  mypipe[1] );
-*/
 			memcpy( pipe_write.param1.pipe_fd, mypipe, sizeof(int)*2);
 		}
 		else
 		{
 			/*
 			  This is the last element of the pipeline.
-			*/
-
-			/*
 			  Remove the io redirection for pipe output.
 			*/
 			j->io = io_remove( j->io, &pipe_write );
@@ -803,39 +869,26 @@ void exec( job_t *j )
 		{
 			case INTERNAL_FUNCTION:
 			{
-				wchar_t **arg;
-				int i;
-				string_buffer_t sb;
-			
-				const wchar_t * def = function_get_definition( p->argv[0] );
-//			fwprintf( stderr, L"run function %ls\n", argv[0] );
+				
+				wchar_t * def = halloc_register( j, wcsdup( function_get_definition( p->argv[0] )));
 				if( def == 0 )
 				{
 					debug( 0, _( L"Unknown function '%ls'" ), p->argv[0] );
 					break;
 				}
+				
 				parser_push_block( FUNCTION_CALL );
 				
-				if( builtin_count_args(p->argv)>1 )
-				{
-					sb_init( &sb );
-				
-					for( i=1, arg=p->argv+1; *arg; i++, arg++ )
-					{
-						if( i != 1 )
-							sb_append( &sb, ARRAY_SEP_STR );
-						sb_append( &sb, *arg );
-					}
-
-					env_set( L"argv", (wchar_t *)sb.buff, ENV_LOCAL );
-					sb_destroy( &sb );
-				}
+				current_block->param2.function_call_process = p;
+				current_block->param1.function_name = halloc_register( current_block, wcsdup( p->argv[0] ) );
+						
+				parse_util_set_argv( p->argv+1 );
+								
 				parser_forbid_function( p->argv[0] );
 
 				if( p->next )
 				{
-//					fwprintf( stderr, L"Function %ls\n", def );
-					io_buffer = io_buffer_create();					
+					io_buffer = io_buffer_create( 0 );					
 					j->io = io_add( j->io, io_buffer );
 				}
 				
@@ -851,16 +904,15 @@ void exec( job_t *j )
 			{
 				if( p->next )
 				{
-//					fwprintf( stderr, L"Block %ls\n", p->argv[0] );
-					io_buffer = io_buffer_create();					
+					io_buffer = io_buffer_create( 0 );					
 					j->io = io_add( j->io, io_buffer );
 				}
-				
+								
 				internal_exec_helper( p->argv[0], TOP, j->io );			
 				break;
 				
 			}
-			
+
 			case INTERNAL_BUILTIN:
 			{
 				int builtin_stdin=0;
@@ -870,6 +922,7 @@ void exec( job_t *j )
 				if( p == j->first_process )
 				{
 					io_data_t *in = io_get( j->io, 0 );
+					
 					if( in )
 					{
 						switch( in->io_mode )
@@ -878,11 +931,6 @@ void exec( job_t *j )
 							case IO_FD:
 							{
 								builtin_stdin = in->param1.old_fd;
-/*								fwprintf( stderr, 
-  L"Input redirection for builtin '%ls' from fd %d\n", 
-  p->argv[0],
-  in->old_fd );
-*/
 								break;
 							}
 							case IO_PIPE:
@@ -893,11 +941,6 @@ void exec( job_t *j )
 							
 							case IO_FILE:
 							{
-/*								
-  fwprintf( stderr, 
-  L"Input redirection for builtin from file %ls\n",
-  in->filename);
-*/
 								builtin_stdin=wopen( in->param1.filename,
                                               in->param2.flags, 0777 );
 								if( builtin_stdin == -1 )
@@ -937,40 +980,37 @@ void exec( job_t *j )
 					exec_error=1;
 					break;
 				}
+				else
+				{
+					builtin_push_io( builtin_stdin );
+					
+					/* 
+					   Since this may be the foreground job, and since a
+					   builtin may execute another foreground job, we need to
+					   pretend to suspend this job while running the builtin.
+					*/
+					
+					builtin_out_redirect = has_fd( j->io, 1 );
+					builtin_err_redirect = has_fd( j->io, 2 );		
+					fg = j->fg;
+					j->fg = 0;
+					
+					signal_unblock();
+					
+					p->status = builtin_run( p->argv );
+					
+					signal_block();
+					
+					/*
+					  Restore the fg flag, which is temporarily set to
+					  false during builtin execution so as not to confuse
+					  some job-handling builtins.
+					*/
+					j->fg = fg;
+				}
 				
-				builtin_push_io( builtin_stdin );
-				
-				/* 
-				   Since this may be the foreground job, and since a
-				   builtin may execute another foreground job, we need to
-				   pretend to suspend this job while running the builtin.
-				*/
-
-				builtin_out_redirect = has_fd( j->io, 1 );
-				builtin_err_redirect = has_fd( j->io, 2 );		
-				fg = j->fg;
-				j->fg = 0;
-				
-				signal_unblock();
-				
-				p->status = builtin_run( p->argv );
-				
-				signal_block();
-				
-				/*
-				  Restore the fg flag, which is temporarily set to
-				  false during builtin execution so as not to confuse
-				  some job-handling builtins.
-				*/
-				j->fg = fg;
-
-
-/*			if( is_interactive )
-  fwprintf( stderr, L"Builtin '%ls' finished\n", j->command );
-*/
 				if( close_stdin )
 				{
-//					fwprintf( stderr, L"Close builtin_stdin\n" );					
 					exec_close( builtin_stdin );
 				}				
 				break;				
@@ -982,6 +1022,7 @@ void exec( job_t *j )
 
 		switch( p->type )
 		{
+
 			case INTERNAL_BLOCK:
 			case INTERNAL_FUNCTION:
 			{
@@ -996,7 +1037,7 @@ void exec( job_t *j )
 				if( !io_buffer )
 				{
 					/*
-					  No buffer, se we exit directly. This means we
+					  No buffer, so we exit directly. This means we
 					  have to manually set the exit status.
 					*/
 					if( p->next == 0 )
@@ -1014,6 +1055,7 @@ void exec( job_t *j )
 				if( io_buffer->param2.out_buffer->used != 0 )
 				{
 					pid = fork();
+
 					if( pid == 0 )
 					{
 						/*
@@ -1041,12 +1083,19 @@ void exec( job_t *j )
 						   it control over the terminal.
 						*/
 						p->pid = pid;						
-						if( handle_new_child( j, p ) )
-							exit(1);
-						
+						set_child_group( j, p, 0 );
+												
 					}					
 					
 				}
+				else
+				{
+					if( p->next == 0 )
+					{
+						proc_set_last_status( j->negate?(status?0:1):status);
+					}
+					p->completed = 1;
+				}
 				
 				io_buffer_destroy( io_buffer );
 				
@@ -1054,6 +1103,46 @@ void exec( job_t *j )
 				break;
 				
 			}
+
+
+			case INTERNAL_BUFFER:
+			{
+		
+				pid = fork();
+				
+				if( pid == 0 )
+				{
+					/*
+					  This is the child process. Write out the contents of the pipeline.
+					*/
+					p->pid = getpid();
+					setup_child_process( j, p );
+					
+					write( 1,
+						   input_redirect->param2.out_buffer->buff, 
+						   input_redirect->param2.out_buffer->used );
+					exit( 0 );
+				}
+				else if( pid < 0 )
+				{
+					/* The fork failed. */
+					debug( 0, FORK_ERROR );
+					wperror (L"fork");
+					exit (1);
+				}
+				else
+				{
+					/* 
+					   This is the parent process. Store away
+					   information on the child, and possibly give
+					   it control over the terminal.
+					*/
+					p->pid = pid;						
+					set_child_group( j, p, 0 );	
+				}	
+
+				break;				
+			}
 			
 			case INTERNAL_BUILTIN:
 			{
@@ -1081,7 +1170,6 @@ void exec( job_t *j )
 					( sb_out->used ) && 
 					( buffer_stdout ) )
 				{
-//				fwprintf( stderr, L"Skip fork of %ls\n", j->command );				
 					char *res = wcs2str( (wchar_t *)sb_out->buff );				
 					b_append( io->param2.out_buffer, res, strlen( res ) );
 					skip_fork = 1;
@@ -1104,9 +1192,13 @@ void exec( job_t *j )
 				pid = fork();
 				if( pid == 0 )
 				{
+
 					/*
-					  This is the child process. 
+					  This is the child process. Setup redirections,
+					  print correct output to stdout and stderr, and
+					  then exit.
 					*/
+
 					p->pid = getpid();
 					setup_child_process( j, p );
 					if( sb_out->used )
@@ -1128,14 +1220,13 @@ void exec( job_t *j )
 				{
 					/* 
 					   This is the parent process. Store away
-					   information on the child, and possibly fice
+					   information on the child, and possibly give
 					   it control over the terminal.
 					*/
 					p->pid = pid;
 						
-					if( handle_new_child( j, p ) )
-						exit( 1 );
-					
+					set_child_group( j, p, 0 );
+										
 				}					
 				
 				break;
@@ -1152,7 +1243,7 @@ void exec( job_t *j )
 					p->pid = getpid();
 					setup_child_process( j, p );
 					launch_process( p );
-
+					
 					/*
 					  launch_process _never_ returns...
 					*/
@@ -1173,20 +1264,16 @@ void exec( job_t *j )
 					*/
 					p->pid = pid;
 
-					if( handle_new_child( j, p ) )
-						exit( 1 );
-					
-										
+					set_child_group( j, p, 0 );
+															
 				}
 				break;
 			}
 			
 		}
 
-
 		if( p->type == INTERNAL_BUILTIN )
-			builtin_pop_io();			
-		
+			builtin_pop_io();
 				
 		/* 
 		   Close the pipe the current process uses to read from the previous process_t 
@@ -1200,8 +1287,11 @@ void exec( job_t *j )
 			pipe_read.param1.pipe_fd[0] = mypipe[0];
 		
 		/* 
-		   If there is a next process, close the output end of the
-		   pipe (the child subprocess already has a copy of the pipe).
+		   If there is a next process in the pipeline, close the
+		   output end of the current pipe (the surrent child
+		   subprocess already has a copy of the pipe - this makes sure
+		   we don't leak file descriptors either in the shell or in
+		   the children).
 		*/
 		if( p->next )
 		{
@@ -1209,6 +1299,15 @@ void exec( job_t *j )
 		}		
 	}
 
+	/*
+	  The keepalive process is no longer needed, so we terminate it
+	  with extreme prejudice
+	*/
+	if( needs_keepalive )
+	{
+		kill( keepalive.pid, SIGKILL );
+	}
+	
 	signal_unblock();	
 
 	debug( 3, L"Job is constructed" );
@@ -1224,14 +1323,12 @@ void exec( job_t *j )
 	{
 		proc_last_bg_pid = j->pgid;
 	}
-	
+
 	if( !exec_error )
 	{
 		job_continue (j, 0);
 	}
 	
-	debug( 3, L"End of exec()" );
-	
 }
 
 int exec_subshell( const wchar_t *cmd, 
@@ -1252,7 +1349,7 @@ int exec_subshell( const wchar_t *cmd,
 	}
 	
 	is_subshell=1;	
-	io_buffer= io_buffer_create();
+	io_buffer= io_buffer_create( 0 );
 	
 	prev_status = proc_get_last_status();
 	
@@ -1281,24 +1378,30 @@ int exec_subshell( const wchar_t *cmd,
 					{
 						wchar_t *el = str2wcs( begin );
 						if( el )
+						{
 							al_push( l, el );
+						}
+						else
+						{
+							debug( 2, L"Got null string on line %d of file %s", __LINE__, __FILE__ );
+						}
 					}				
 					io_buffer_destroy( io_buffer );
-				
+					
 					return status;
 				
 				case '\n':
 				{
 					wchar_t *el;
 					*end=0;
-					el = str2wcs( begin );				
-					if( !el )
+					el = str2wcs( begin );
+					if( el )
 					{
-						debug( 0, _( L"Subshell '%ls' returned illegal string, discarded one entry" ), cmd );
+						al_push( l, el );
 					}
 					else
 					{
-						al_push( l, el );
+						debug( 2, L"Got null string on line %d of file %s", __LINE__, __FILE__ );
 					}
 					begin = end+1;
 					break;
@@ -1309,5 +1412,6 @@ int exec_subshell( const wchar_t *cmd,
 	}
 	
 	io_buffer_destroy( io_buffer );
+
 	return status;	
 }
