@@ -20,9 +20,9 @@
 #include "util.h"
 
 /**
-   Flag specifying that subshell expansion should be skipped
+   Flag specifying that cmdsubst expansion should be skipped
 */
-#define EXPAND_SKIP_SUBSHELL 1
+#define EXPAND_SKIP_CMDSUBST 1
 
 /**
    Flag specifying that variable expansion should be skipped
@@ -112,99 +112,92 @@ enum
 /** String containing the character for separating two array elements */
 #define ARRAY_SEP_STR L"\x1e"
 
+/**
+   Error issued on array out of bounds
+*/
+#define ARRAY_BOUNDS_ERR _(L"Array index out of bounds")
 
 
 /**
    Perform various forms of expansion on in, such as tilde expansion
-   (~USER becomes the users home directory), variable expansion
-   ($VAR_NAME becomes the value of the environment variable VAR_NAME),
-   subshell expansion and wildcard expansion. The results are inserted
+   (\~USER becomes the users home directory), variable expansion
+   (\$VAR_NAME becomes the value of the environment variable VAR_NAME),
+   cmdsubst expansion and wildcard expansion. The results are inserted
    into the list out.
    
    If the parameter does not need expansion, it is copied into the list
    out. If expansion is performed, the original parameter is freed and
    newly allocated strings are inserted into the list out.
-   
+
+   If \c context is non-null, all the strings contained in the
+   array_list_t \c out will be registered to be free'd when context is
+   free'd.
+ 
+   \param context the halloc context to use for automatic deallocation
    \param in The parameter to expand
-   \param flag Specifies if any expansion pass should be skipped. Legal values are any combination of EXPAND_SKIP_SUBSHELL EXPAND_SKIP_VARIABLES and EXPAND_SKIP_WILDCARDS
+   \param flag Specifies if any expansion pass should be skipped. Legal values are any combination of EXPAND_SKIP_CMDSUBST EXPAND_SKIP_VARIABLES and EXPAND_SKIP_WILDCARDS
    \param out The list to which the result will be appended.
-   \return One of EXPAND_OK, EXPAND_ERROR, EXPAND_WILDCARD_MATCH and EXPAND_WILDCARD_NO_MATCH
+   \return One of EXPAND_OK, EXPAND_ERROR, EXPAND_WILDCARD_MATCH and EXPAND_WILDCARD_NO_MATCH. EXPAND_WILDCARD_NO_MATCH and EXPAND_WILDCARD_MATCH are normal exit conditions used only on strings containing wildcards to tell if the wildcard produced any matches.
 */
-int expand_string( wchar_t *in, array_list_t *out, int flag );
+int expand_string( void *context, wchar_t *in, array_list_t *out, int flag );
 
 /**
    expand_one is identical to expand_string, except it will fail if in
    expands to more than one string. This is used for expanding command
    names.
 
+   If \c context is non-null, the returning string ill be registered
+   to be free'd when context is free'd.
+ 
+   \param context the halloc context to use for automatic deallocation   
    \param in The parameter to expand
-   \param flag Specifies if any expansion pass should be skipped. Legal values are any combination of EXPAND_SKIP_SUBSHELL EXPAND_SKIP_VARIABLES and EXPAND_SKIP_WILDCARDS
+   \param flag Specifies if any expansion pass should be skipped. Legal values are any combination of EXPAND_SKIP_CMDSUBST EXPAND_SKIP_VARIABLES and EXPAND_SKIP_WILDCARDS
    \return The expanded parameter, or 0 on failiure
 */
-wchar_t *expand_one( wchar_t *in, int flag );
-
-/**
-   Expand backslashed escapes and substitute them with their unescaped
-   counterparts. Also optionally change the wildcards, the tilde
-   character and a few more into constants which are defined in a
-   private use area of Unicode. This assumes wchar_t is a unicode
-   character.  character set.
-
-   The result must be free()d. The original string is not modified. If
-   an invalid sequence is specified, 0 is returned.
-
-*/
-wchar_t *expand_unescape( const wchar_t * in, int escape_special );
-
-/**
-   Replace special characters with escape sequences. Newline is
-   replaced with \n, etc. 
-
-   The result must be free()d. The original string is not modified.
-
-   \param in The string to be escaped
-   \param escape_all Whether all characters wich hold special meaning in fish (Pipe, semicolon, etc,) should be escaped, or only unprintable characters
-   \return The escaped string
-*/
-wchar_t *expand_escape( const wchar_t *in, int escape_all );
+wchar_t *expand_one( void *context, wchar_t *in, int flag );
 
 /**
    Convert the variable value to a human readable form, i.e. escape things, handle arrays, etc. Suitable for pretty-printing.
+
+   \param in the value to escape
 */
 wchar_t *expand_escape_variable( const wchar_t *in );
 
-
 /**
    Perform tilde expansion and nothing else on the specified string.
 
    If tilde expansion is needed, the original string is freed and a
    new string, allocated using malloc, is returned.
+
+   \param in the string to tilde expand
 */
 wchar_t *expand_tilde(wchar_t *in);
 
-/**
-   Locate the first subshell in the specified string.
-   
-   \param in the string to search for subshells
-   \param begin the starting paranthesis of the subshell
-   \param end the ending paranthesis of the subshell
-   \param flags set this variable to ACCEPT_INCOMPLETE if in tab_completion mode
-   \return -1 on syntax error, 0 if no subshells exist and 1 on sucess
-*/
-int expand_locate_subshell( wchar_t *in, 
-							wchar_t **begin, 
-							wchar_t **end,
-							int flags );
-
 
 /**
-   Tokenize the specified string into the specified array_list_t.
-   Each new element is allocated using malloc and must be freed by the
-   caller.
-   
-   \param val the input string. The contents of this string is not changed.
-   \param out the list in which to place the elements. 
+   Test if the specified argument is clean, i.e. it does not contain
+   any tokens which need to be expanded or otherwise altered. Clean
+   strings can be passed through expand_string and expand_one without
+   changing them. About two thirds of all strings are clean, so
+   skipping expansion on them actually does save a small amount of
+   time, since it avoids multiple memory allocations during the
+   expansion process.
+
+   \param in the string to test
 */
-void expand_variable_array( const wchar_t *val, array_list_t *out );
+int expand_is_clean( const wchar_t *in );
+
+/**
+   Perform error reporting for a syntax error related to the variable
+   expansion beginning at the specified character of the specified
+   token. This function will call the error function with an
+   explanatory string about what is wrong with the specified token.
+
+   \param token The token containing the error
+   \param token_pos The position where the expansion begins
+   \param error_pos The position on the line to report to the error function.
+*/
+void expand_variable_error( const wchar_t *token, int token_pos, int error_pos );
+
 
 #endif
