@@ -9,7 +9,14 @@
 
 #include <sys/types.h>
 #include <sys/stat.h>
+#ifdef HAVE_SYS_TERMIOS_H
+#include <sys/termios.h>
+#endif
+
+#ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
+#endif
+
 #include <sys/time.h>
 #include <sys/wait.h>
 #include <dirent.h>
@@ -27,16 +34,26 @@
 #include <termio.h>
 #endif
 
+#if HAVE_TERM_H
 #include <term.h>
+#elif HAVE_NCURSES_TERM_H
+#include <ncurses/term.h>
+#endif
+
 #include <signal.h>
 
+
+#include "fallback.h"
 #include "util.h"
+
 #include "wutil.h"
 #include "common.h"
 #include "complete.h"
 #include "output.h"
 #include "input_common.h"
 #include "env_universal.h"
+#include "halloc.h"
+#include "halloc_util.h"
 
 enum 
 {
@@ -69,6 +86,7 @@ static struct termios pager_modes;
 
 static int is_ca_mode = 0;
 
+static buffer_t *pager_buffer;
 
 /**
    The environment variables used to specify the color of different
@@ -86,20 +104,55 @@ static wchar_t *hightlight_var[] =
 static string_buffer_t out_buff;
 static FILE *out_file;
 
-
-int get_color( int highlight )
+/**
+   Data structure describing one or a group of related completions
+ */
+typedef struct 
 {
+	/**
+	   The list of all completin strings this entry applies to
+	*/
+	array_list_t *comp;
+	/**
+	   The description
+	*/
+	wchar_t *desc;
+	/**
+	   On-screen width of the completion string
+	*/
+	int comp_width;	
+	/**
+	   On-screen width of the description information
+	*/
+	int desc_width;	
+	/**
+	   Preffered total width
+	*/
+	int pref_width;
+	/**
+	   Minimum acceptable width
+	*/
+	int min_width;
+}
+	comp_t;
+
+static int get_color( int highlight )
+{
+	wchar_t *val;
+
 	if( highlight < 0 )
 		return FISH_COLOR_NORMAL;
 	if( highlight >= (4) )
 		return FISH_COLOR_NORMAL;
 	
-	wchar_t *val = env_universal_get( hightlight_var[highlight]);
+	val = wgetenv( hightlight_var[highlight]);
+
+	if( !val )
+	{
+		val = env_universal_get( hightlight_var[highlight]);
+	}
 	
-	if( val == 0 )
-		return FISH_COLOR_NORMAL;
-	
-	if( val == 0 )
+	if( !val )
 	{
 		return FISH_COLOR_NORMAL;
 	}
@@ -107,7 +160,20 @@ int get_color( int highlight )
 	return output_color_code( val );	
 }
 
-int try_sequence( char *seq )
+static void recalc_width( array_list_t *l, const wchar_t *prefix )
+{
+	int i;
+	for( i=0; i<al_get_count( l ); i++ )
+	{
+		comp_t *c = (comp_t *)al_get( l, i );
+		
+		c->min_width = mini( c->desc_width, maxi(0,termsize.ws_col/3 - 2)) +
+			mini( c->desc_width, maxi(0,termsize.ws_col/5 - 4)) +4;
+	}
+	
+}
+
+static int try_sequence( char *seq )
 {
 	int j, k;
 	wint_t c=0;
@@ -190,6 +256,110 @@ static wint_t readch()
 }
 
 
+static int pager_buffered_writer( char c)
+{
+	b_append( pager_buffer, &c, 1 );
+	return 0;
+}
+
+static void pager_flush()
+{
+	write( 1, pager_buffer->buff, pager_buffer->used );
+	pager_buffer->used = 0;
+}
+
+
+static int print_max( const wchar_t *str, int max, int has_more )
+{
+	int i;
+	int written = 0;
+	for( i=0; str[i]; i++ )
+	{
+		if( written + wcwidth(str[i]) > max )
+			break;
+		if( ( written + wcwidth(str[i]) == max) && (has_more || str[i+1]) )
+		{
+			writech( ellipsis_char );
+			written += wcwidth(ellipsis_char );
+			break;
+		}
+			
+		writech( str[i] );
+		written+= wcwidth( str[i] );
+	}
+	return written;
+}
+
+static void completion_print_item( const wchar_t *prefix, comp_t *c, int width )
+{
+	int comp_width=0, desc_width=0;
+	int i;
+	int written=0;
+	
+	if( c->pref_width <= width )
+	{
+		/*
+		  The entry fits, we give it as much space as it wants
+		*/
+		comp_width = c->comp_width;
+		desc_width = c->desc_width;
+	}
+	else
+	{
+		/*
+		  The completion and description won't fit on the
+		  allocated space. Give a maximum of 2/3 of the
+		  space to the completion, and whatever is left to
+		  the description.
+		*/
+		int desc_all = c->desc_width?c->desc_width+4:0;
+		
+		comp_width = maxi( mini( c->comp_width,
+								 2*(width-4)/3 ),
+						   width - desc_all );
+		if( c->desc_width )
+			desc_width = width-comp_width-4;
+		else
+			c->desc_width=0;
+		
+	}
+	
+	for( i=0; i<al_get_count( c->comp ); i++ )
+	{
+		const wchar_t *comp = (const wchar_t *)al_get( c->comp, i );
+		if( i != 0 )
+			written += print_max( L"  ", comp_width - written, 2 );
+		set_color( get_color(HIGHLIGHT_PAGER_PREFIX),FISH_COLOR_NORMAL );
+		written += print_max( prefix, comp_width - written, comp[0]?1:0 );
+		set_color( get_color(HIGHLIGHT_PAGER_COMPLETION),FISH_COLOR_IGNORE );
+		written += print_max( comp, comp_width - written, i!=(al_get_count(c->comp)-1) );
+	}
+
+
+	if( desc_width )
+	{
+		while( written < (width-desc_width-2))
+		{
+			written++;
+			writech( L' ');
+		}
+		written += print_max( L"(", 1, 0 );
+		set_color( get_color( HIGHLIGHT_PAGER_DESCRIPTION ),
+				   FISH_COLOR_IGNORE );
+		written += print_max( c->desc, desc_width, 0 );
+		written += print_max( L")", 1, 0 );
+	}
+	else
+	{
+		while( written < width )
+		{
+			written++;
+			writech( L' ');
+		}
+	}
+	
+}
+
 /**
    Print the specified part of the completion list, using the
    specified column offsets and quoting style.
@@ -214,277 +384,29 @@ static void completion_print( int cols,
 
 	int rows = (al_get_count( l )-1)/cols+1;
 	int i, j;
-	int prefix_width= my_wcswidth(prefix);
 
 	for( i = row_start; i<row_stop; i++ )
 	{
 		for( j = 0; j < cols; j++ )
 		{
-			wchar_t *el, *el_end;
+			comp_t *el;
 
+			int is_last = (j==(cols-1));
+			
 			if( al_get_count( l ) <= j*rows + i )
 				continue;
 
-			el = (wchar_t *)al_get( l, j*rows + i );
-			el_end= wcschr( el, COMPLETE_SEP );
-
-			set_color( get_color(HIGHLIGHT_PAGER_PREFIX),FISH_COLOR_NORMAL );
-
-			writestr( prefix );
-
-			set_color( get_color(HIGHLIGHT_PAGER_COMPLETION),FISH_COLOR_IGNORE );
-
-			if( el_end == 0 )
-			{
-				/* We do not have a description for this completion */
-				int written = 0;
-				int max_written = width[j] - prefix_width - (j==cols-1?0:2);
-
-				if( is_quoted )
-				{
-					for( i=0; i<max_written; i++ )
-					{
-						if( !el[i] )
-							break;
-						writech( el[i] );
-						written+= wcwidth( el[i] );
-					}
-				}
-				else
-				{
-					written = write_escaped_str( el, max_written );
-				}
-
-				set_color( get_color( HIGHLIGHT_PAGER_DESCRIPTION ),
-						   FISH_COLOR_IGNORE );
-
-				writespace( width[j]-
-							written-
-							prefix_width );
-			}
-			else
-			{
-				int whole_desc_width = my_wcswidth(el_end+1);
-				int whole_comp_width;
-
-				/*
-				  Temporarily drop the description so that wcswidth et
-				  al only calculate the width of the completion.
-				*/
-				*el_end = L'\0';
-
-				/*
-				  Calculate preferred completion width
-				*/
-				if( is_quoted )
-				{
-					whole_comp_width = my_wcswidth(el);
-				}
-				else
-				{
-					wchar_t *tmp = escape( el, 1 );
-					whole_comp_width = my_wcswidth( tmp );
-					free(tmp);
-				}
-
-				/*
-				  Calculate how wide this entry 'wants' to be
-				*/
-				int pref_width = whole_desc_width + 4 + prefix_width + 2 -
-					(j==cols-1?2:0) + whole_comp_width;
-
-				int comp_width, desc_width;
-
-				if( pref_width <= width[j] )
-				{
-					/*
-					  The entry fits, we give it as much space as it wants
-					*/
-					comp_width = whole_comp_width;
-					desc_width = whole_desc_width;
-				}
-				else
-				{
-					/*
-					  The completion and description won't fit on the
-					  allocated space. Give a maximum of 2/3 of the
-					  space to the completion, and whatever is left to
-					  the description.
-					*/
-					int sum = width[j] - prefix_width - 4 - 2 + (j==cols-1?2:0);
-
-					comp_width = maxi( mini( whole_comp_width,
-											 2*sum/3 ),
-									   sum - whole_desc_width );
-					desc_width = sum-comp_width;
-				}
-
-				/* First we must print the completion. */
-				if( is_quoted )
-				{
-					writestr_ellipsis( el, comp_width);
-				}
-				else
-				{
-					write_escaped_str( el, comp_width );
-				}
-
-				/* Put the description back */
-				*el_end = COMPLETE_SEP;
-
-				/* And print it */
-				set_color( get_color(HIGHLIGHT_PAGER_DESCRIPTION),
-						   FISH_COLOR_IGNORE );
-				writespace( maxi( 2,
-								  width[j]
-								  - comp_width
-								  - desc_width
-								  - 4
-								  - prefix_width
-								  + (j==cols-1?2:0) ) );
-				/* Print description */
-				writestr(L"(");
-				writestr_ellipsis( el_end+1, desc_width);
-				writestr(L")");
-
-				if( j != cols-1)
-					writestr( L"  " );
-
-			}
+			el = (comp_t *)al_get( l, j*rows + i );
+			
+			completion_print_item( prefix, el, width[j] - (is_last?0:2) );
+			
+			if( !is_last)
+				writestr( L"  " );
 		}
 		writech( L'\n' );
 	}
 }
 
-/**
-   Calculates how long the specified string would be when printed on the command line.
-
-   \param str The string to be printed.
-   \param is_quoted Whether the string would be printed quoted or unquoted
-   \param pref_width the preferred width for this item
-   \param min_width the minimum width for this item
-*/
-static void printed_length( wchar_t *str,
-							int is_quoted,
-							int *pref_width,
-							int *min_width )
-{
-	if( is_quoted )
-	{
-		wchar_t *sep = wcschr(str,COMPLETE_SEP);
-		if( sep )
-		{
-			*sep=0;
-			int cw = my_wcswidth( str );
-			int dw = my_wcswidth(sep+1);
-
-			if( termsize.ws_col > 80 )
-				dw = mini( dw, termsize.ws_col/3 );
-
-
-			*pref_width = cw+dw+4;
-
-			if( dw > termsize.ws_col/3 )
-			{
-				dw = termsize.ws_col/3;
-			}
-
-			*min_width=cw+dw+4;
-
-			*sep= COMPLETE_SEP;
-			return;
-		}
-		else
-		{
-			*pref_width=*min_width= my_wcswidth( str );
-			return;
-		}
-
-	}
-	else
-	{
-		int comp_len=0, desc_len=0;
-		int has_description = 0;
-		while( *str != 0 )
-		{
-			switch( *str )
-			{
-				case L'\n':
-				case L'\b':
-				case L'\r':
-				case L'\e':
-				case L'\t':
-				case L'\\':
-				case L'&':
-				case L'$':
-				case L' ':
-				case L'#':
-				case L'^':
-				case L'<':
-				case L'>':
-				case L'(':
-				case L')':
-				case L'[':
-				case L']':
-				case L'{':
-				case L'}':
-				case L'?':
-				case L'*':
-				case L'|':
-				case L';':
-				case L':':
-				case L'\'':
-				case L'"':
-				case L'%':
-				case L'~':
-					
-					if( has_description )
-						desc_len++;
-					else
-						comp_len+=2;
-					break;
-
-				case COMPLETE_SEP:
-					has_description = 1;
-					break;
-
-				default:
-					if( has_description )
-						desc_len+= wcwidth(*str);
-					else
-						comp_len+= wcwidth(*str);
-					break;
-			}
-			str++;
-		}
-		if( has_description )
-		{
-			/*
-			  Mangle long descriptions to make formating look nicer
-			*/
-			debug( 3, L"Desc, width = %d %d\n", comp_len, desc_len );
-//			if( termsize.ws_col > 80 )
-//				desc_len = mini( desc_len, termsize.ws_col/3 );
-
-			*pref_width = comp_len+ desc_len+4;;
-
-			comp_len = mini( comp_len, maxi(0,termsize.ws_col/3 - 2));
-			desc_len = mini( desc_len, maxi(0,termsize.ws_col/5 - 4));
-
-			*min_width = comp_len+ desc_len+4;
-			return;
-		}
-		else
-		{
-			debug( 3, L"No desc, width = %d\n", comp_len );
-			
-			*pref_width=*min_width= comp_len;
-			return;
-		}
-
-	}
-}
-
 
 /**
    Try to print the list of completions l with the prefix prefix using
@@ -532,8 +454,6 @@ static int completion_try_print( int cols,
 	
 	int pref_tot_width=0;
 	int min_tot_width = 0;
-	int prefix_width = my_wcswidth( prefix );
-	
 	int res=0;
 	/*
 	  Skip completions on tiny terminals
@@ -541,25 +461,24 @@ static int completion_try_print( int cols,
 	
 	if( termsize.ws_col < 16 )
 		return 1;
-
+	
 	memset( pref_width, 0, sizeof(pref_width) );
 	memset( min_width, 0, sizeof(min_width) );
-
-	/* Calculated how wide the list would be */
+	
+	/* Calculate how wide the list would be */
 	for( j = 0; j < cols; j++ )
 	{
 		for( i = 0; i<rows; i++ )
 		{
 			int pref,min;
-			wchar_t *el;
+			comp_t *c;
 			if( al_get_count( l ) <= j*rows + i )
 				continue;
 
-			el = (wchar_t *)al_get( l, j*rows + i );
-			printed_length( el, is_quoted, &pref, &min );
-
-			pref += prefix_width;
-			min += prefix_width;
+			c = (comp_t *)al_get( l, j*rows + i );
+			pref = c->pref_width;
+			min = c->min_width;
+			
 			if( j != cols-1 )
 			{
 				pref += 2;
@@ -641,6 +560,7 @@ static int completion_try_print( int cols,
 			}
 			
 			completion_print( cols, width, 0, rows, prefix, is_quoted, l);
+			pager_flush();
 		}
 		else
 		{
@@ -663,16 +583,16 @@ static int completion_try_print( int cols,
 			*/
 			while(do_loop)
 			{
-				wchar_t msg[10];
-				int percent = 100*pos/(rows-termsize.ws_row+1);
+				wchar_t msg[30];
+				int percent = 100*(pos+rows)/(2*rows-termsize.ws_row+1);
 				set_color( FISH_COLOR_BLACK,
 						   get_color(HIGHLIGHT_PAGER_PROGRESS) );
-				swprintf( msg, 12,
-						  L" %ls(%d%%) \r",
-						  percent==100?L"":(percent >=10?L" ": L"  "),
-						  percent );
+				swprintf( msg, 30,
+						  L" %d to %d of %d \r",
+						  pos, pos+termsize.ws_row-1, rows );
 				writestr(msg);
 				set_color( FISH_COLOR_NORMAL, FISH_COLOR_NORMAL );
+				pager_flush();
 				int c = readch();
 
 				switch( c )
@@ -734,7 +654,8 @@ static int completion_try_print( int cols,
 						}
 						else
 						{
-							writembs( flash_screen );
+							if( flash_screen )
+								writembs( flash_screen );
 						}
 
 						break;
@@ -758,7 +679,8 @@ static int completion_try_print( int cols,
 						}
 						else
 						{
-							writembs( flash_screen );
+							if( flash_screen )
+								writembs( flash_screen );
 						}
 						break;
 					}
@@ -786,7 +708,9 @@ static int completion_try_print( int cols,
 }
 
 /**
-   Substitute any series of tabs, newlines, etc. with a single space character in completion description
+   Substitute any series of whitespace with a single space character
+   inside completion descriptions. Remove all whitespace from
+   beginning/end of completion descriptions.
 */
 static void mangle_descriptions( array_list_t *l )
 {
@@ -795,7 +719,7 @@ static void mangle_descriptions( array_list_t *l )
 	{
 		wchar_t *next = (wchar_t *)al_get(l, i);
 		wchar_t *in, *out;
-		skip=0;
+		skip=1;
 		
 		while( *next != COMPLETE_SEP && *next )
 			next++;
@@ -824,6 +748,124 @@ static void mangle_descriptions( array_list_t *l )
 	}
 }
 
+/**
+   Merge multiple completions with the same description to the same line
+*/
+static void join_completions( array_list_t *l )
+{
+	long i;
+	int in, out;
+	hash_table_t desc_table;
+	
+	hash_init( &desc_table, &hash_wcs_func, &hash_wcs_cmp );
+
+	for( i=0; i<al_get_count(l); i++ )
+	{
+		wchar_t *item = (wchar_t *)al_get( l, i );
+		wchar_t *desc = wcschr( item, COMPLETE_SEP );
+		long prev_idx;
+		
+		if( !desc )
+			continue;
+		desc++;
+		prev_idx = ((long)hash_get( &desc_table, desc) )-1;
+		if( prev_idx == -1 )
+		{
+			hash_put( &desc_table, halloc_wcsdup( global_context, desc), (void *)(i+1));
+		}
+		else
+		{
+			string_buffer_t foo;
+			wchar_t *old = (wchar_t *)al_get( l, prev_idx );
+			wchar_t *old_end = wcschr( old, COMPLETE_SEP );
+			
+			if( old_end )
+			{
+				*old_end = 0;
+				
+				sb_init( &foo );
+				sb_append( &foo, old );
+				sb_append_char( &foo, COMPLETE_ITEM_SEP );
+				sb_append( &foo, item );
+
+				free( (void *)al_get( l, prev_idx ) );
+				al_set( l, prev_idx, foo.buff );
+
+				free( (void *)al_get( l, i ) );
+				al_set( l, i, 0 );
+			}
+			
+		}
+		
+	}	
+	hash_destroy( &desc_table );
+
+	out=0;
+	for( in=0; in < al_get_count(l); in++ )
+	{
+		const void * d =  al_get( l, in );
+		
+		if( d )
+		{
+			al_set( l, out++, d );
+		}
+	}
+	al_truncate( l, out );	
+}
+
+/**
+   Replace completion strings with a comp_t structure
+*/
+static void mangle_completions( array_list_t *l, const wchar_t *prefix )
+{
+	int i;
+	
+	for( i=0; i<al_get_count( l ); i++ )
+	{
+		wchar_t *next = (wchar_t *)al_get( l, i );
+		wchar_t *start, *end;
+		comp_t *comp = halloc( global_context, sizeof( comp_t ) );
+		comp->comp = al_halloc( global_context );
+		
+		for( start=end=next; 1; end++ )
+		{
+			wchar_t c = *end;
+			
+			if( (c == COMPLETE_ITEM_SEP) || (c==COMPLETE_SEP) || !c)
+			{
+				*end = 0;
+				wchar_t * str = escape( start, 1 );
+				comp->comp_width += my_wcswidth( str );
+				halloc_register( global_context, str );
+				al_push( comp->comp, str );
+				start = end+1;
+			}
+
+			if( c == COMPLETE_SEP )
+			{
+				comp->desc = halloc_wcsdup( global_context, start );
+				break;
+			}	
+			
+			if( !c )
+				break;
+			
+		}
+
+		comp->comp_width  += my_wcswidth(prefix)*al_get_count(comp->comp) + 2*(al_get_count(comp->comp)-1);
+		comp->desc_width = comp->desc?my_wcswidth( comp->desc ):0;
+		
+		comp->pref_width = comp->comp_width + comp->desc_width + (comp->desc_width?4:0);
+		
+		free( next );
+		al_set( l, i, comp );
+	}
+	
+	recalc_width( l, prefix );
+}
+
+
+
 /**
    Respond to a winch signal by checking the terminal size
 */
@@ -846,24 +888,49 @@ static void init()
 	struct sigaction act;
 	program_name = L"fish_pager";
 	wsetlocale( LC_ALL, L"" );
-	
+
+	/*
+	  Make fd 1 output to screen, and use some other fd for writing
+	  the resulting output back to the caller
+	*/
 	int out = dup( 1 );
+	int in = dup( 0 );
 	close(1);
-	if( open( ttyname(0), O_WRONLY ) != 1 )
+	close(0);
+
+	if( (in = open( ttyname(2), O_RDWR )) != -1 )
 	{
 		if( dup2( 2, 1 ) == -1 )
 		{			
-			debug( 0, L"Could not set up file descriptors for pager" );
+			debug( 0, L"Could not set up output file descriptors for pager" );
+			exit( 1 );
+		}
+		
+		if( dup2( in, 0 ) == -1 )
+		{			
+			debug( 0, L"Could not set up input file descriptors for pager %d", in );
 			exit( 1 );
 		}
 	}
+	else
+	    {
+		debug( 0, L"Could not open tty for pager" );
+		exit( 1 );
+	    }
 	out_file = fdopen( out, "w" );
+
+	/**
+	   Init the stringbuffer used to keep any output in
+	*/
 	sb_init( &out_buff );
 
-	output_init();
+	halloc_util_init();
 	env_universal_init( 0, 0, 0, 0);
 	input_common_init( &interrupt_handler );
-	
+	output_set_writer( &pager_buffered_writer );
+	pager_buffer = halloc( global_context, sizeof( buffer_t ) );
+	halloc_register_function( global_context, (void (*)(void *))&b_destroy, pager_buffer );
+
 	sigemptyset( & act.sa_mask );
 	act.sa_flags=0;
 	act.sa_handler=SIG_DFL;
@@ -875,35 +942,8 @@ static void init()
 		exit(1);
 	}
 	
-	/* Loop until we are in the foreground.  */
-	while (tcgetpgrp( 0 ) != getpid())
-	{
-		kill (- getpid(), SIGTTIN);
-	}
-
-	/* Put ourselves in our own process group.  */
-	if( getpgrp() != getpid() )
-	{
-		if (setpgid (getpid(), getpid()) < 0)
-		{
-			debug( 1,
-				   L"Couldn't put the shell in its own process group");
-			wperror( L"setpgid" );
-			exit (1);
-		}
-	}
-
-	/* Grab control of the terminal.  */
-	if( tcsetpgrp (STDIN_FILENO, getpid()) )
-	{
-		debug( 1,
-			   L"Couldn't grab control of terminal" );
-		wperror( L"tcsetpgrp" );
-		exit(1);
-	}
-
 	handle_winch( 0 );                /* Set handler for window change events */
-	
+
 	tcgetattr(0,&pager_modes);        /* get the current terminal modes */
 	memcpy( &saved_modes,
 			&pager_modes,
@@ -914,6 +954,9 @@ static void init()
     pager_modes.c_cc[VMIN]=1;
     pager_modes.c_cc[VTIME]=0;
 
+	/*
+	  
+	*/
     if( tcsetattr(0,TCSANOW,&pager_modes))      /* set the new modes */
     {
         wperror(L"tcsetattr");
@@ -932,45 +975,102 @@ void destroy()
 {
 	env_universal_destroy();
 	input_common_destroy();
-	output_destroy();
-	del_curterm( cur_term );
+	halloc_util_destroy();
+	if( del_curterm( cur_term ) == ERR )
+	{
+		debug( 0, _(L"Error while closing terminfo") );		
+	}
+	
 	sb_destroy( &out_buff );
 	fclose( out_file );
+}
 
+#define BUFSIZE 1024
+void read_array( FILE* file, array_list_t *comp )
+{
+	char buffer[BUFSIZE];
+	int c;
+	int i;
+	wchar_t *wcs, *unescaped;
+
+	while( !feof( file ) )
+	{
+		i = 0;
+		while( i < BUFSIZE-1 )
+		{
+		    c = getc( file );
+			if( c == EOF ) 
+			{
+				return;
+				
+			}
+			if( c == '\n' )
+			{
+				break;
+			}
+			
+			buffer[ i++ ] = c;
+		}
+
+		buffer[ i ] = '\0';
+
+		wcs = str2wcs( buffer );
+		if( wcs ) 
+		{
+			unescaped = unescape( wcs, 0 );
+			al_push( comp, unescaped );
+			free( wcs );
+		}
+	}
 }
 
 int main( int argc, char **argv )
 {
 	int i;
 	int is_quoted=0;	
-	array_list_t comp;
+	array_list_t *comp;
 	wchar_t *prefix;
-
 		
-	init();
 	if( argc < 3 )
 	{
 		debug( 0, L"Insufficient arguments" );
 	}
 	else
 	{
+		comp = al_halloc( global_context );
 		prefix = str2wcs( argv[2] );
 		is_quoted = strcmp( "1", argv[1] )==0;
-	
+		is_quoted = 0;
+		
 		debug( 3, L"prefix is '%ls'", prefix );
-	
-		al_init( &comp );
-	
-		for( i=3; i<argc; i++ )
+		
+	    if( argc > 3 )
 		{
-			al_push( &comp, str2wcs( argv[i] ) );
+		    for( i=3; i<argc; i++ )
+			{
+			    wchar_t *wcs = str2wcs( argv[i] );
+			    if( wcs )
+				{
+				    al_push( comp, wcs );
+				}
+			}
+		}
+	    else
+		{
+		    read_array( stdin, comp );
 		}
-	
-		mangle_descriptions( &comp );
 
+
+	    init();
+
+	    mangle_descriptions( comp );
+	    if( wcscmp( prefix, L"-" ) == 0 )
+		join_completions( comp );
+	    mangle_completions( comp, prefix );
+	
 		for( i = 6; i>0; i-- )
 		{
-			switch( completion_try_print( i, prefix, is_quoted, &comp ) )
+			switch( completion_try_print( i, prefix, is_quoted, comp ) )
 			{
 				case 0:
 					break;
@@ -984,15 +1084,16 @@ int main( int argc, char **argv )
 		
 		}
 	
-		al_foreach( &comp, (void(*)(const void *))&free );
-		al_destroy( &comp );	
 		free(prefix );
 
 		fwprintf( out_file, L"%ls", (wchar_t *)out_buff.buff );
 		if( is_ca_mode )
+		{
 			writembs(exit_ca_mode);
+			pager_flush();
+		}
 	}
-		
+	
 	destroy();
 }
 
