@@ -69,4 +69,15 @@ int function_exists( const wchar_t *name );
 void function_get_names( array_list_t *list, 
 						 int get_hidden );
 
+/**
+   Returns tha absolute path of the file where the specified function
+   was defined. Returns 0 if the file was defined on the commandline.
+*/
+const wchar_t *function_get_definition_file( const wchar_t *name );
+
+/**
+   Returns the linenumber where the definition of the specified function started
+*/
+int function_get_definition_offset( const wchar_t *name );
+
 #endif
