@@ -1,6 +1,8 @@
 /** \file history.c
   History functions, part of the user interface.
 */
+#include "config.h"
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <wchar.h>
@@ -11,14 +13,16 @@
 #include <unistd.h>
 
 
-#include "config.h"
+#include "fallback.h"
 #include "util.h"
+
 #include "wutil.h"
 #include "history.h"
 #include "common.h"
 #include "reader.h"
 #include "env.h"
 #include "sanity.h"
+#include "signal.h"
 
 /*
   The history is implemented using a linked list. Searches are done
@@ -37,14 +41,21 @@ typedef struct
 	   Number of entries
 	*/
 	int count;
+
 	/**
 	   Last history item
 	*/
 	ll_node_t *last;
+
 	/**
 	   The last item loaded from file
 	*/
 	ll_node_t *last_loaded;
+
+	/**
+	   Set to one if the file containing the saved history has been loaded
+	*/
+	int is_loaded;
 }
 	history_data;
 
@@ -73,7 +84,9 @@ static int past_end =1;
 static int history_count=0;
 
 /**
-   The name of the current history list. The name is used to switch between history lists for different commands as sell as for deciding the name of the file to save the history in.
+   The name of the current history list. The name is used to switch
+   between history lists for different commands as sell as for
+   deciding the name of the file to save the history in.
 */
 static wchar_t *mode_name;
 
@@ -82,25 +95,37 @@ static wchar_t *mode_name;
 */
 static hash_table_t history_table;
 
+/**
+   Flag, set to 1 once the history file has been loaded
+*/
+static int is_loaded=0;
 
 /**
-   Load history from ~/.fish_history
+   Load history from file
 */
-static void history_load( wchar_t *name )
+static int history_load()
 {
 	wchar_t *fn;
 	wchar_t *buff=0;
 	int buff_len=0;
 	FILE *in_stream;
 	hash_table_t used;
+	int res = 0;
 	
-	block();
+	if( !mode_name )
+	{
+		return -1;
+	}	
+
+	is_loaded = 1;
+		
+	signal_block();
 	hash_init2( &used, 
 				&hash_wcs_func,
 				&hash_wcs_cmp,
 				4096 );
 
-	fn = wcsdupcat2( env_get(L"HOME"), L"/.", name, L"_history", 0 );
+	fn = wcsdupcat2( env_get(L"HOME"), L"/.", mode_name, L"_history", 0 );
 	
 	in_stream = wfopen( fn, "r" );
 	
@@ -111,13 +136,13 @@ static void history_load( wchar_t *name )
 			int buff_read = fgetws2( &buff, &buff_len, in_stream );
 			if( buff_read == -1 )
 			{
-				debug( 1, L"The following non-fatal error occurred while reading command history from \'%ls\':", name );
+				debug( 1, L"The following non-fatal error occurred while reading command history from \'%ls\':", mode_name );
 				wperror( L"fgetws2" );
 				fclose( in_stream );
 				free( fn );
 				free( buff );
-				unblock();				
-				return;
+				signal_unblock();				
+				return -1;
 			}
 
 			/*
@@ -135,7 +160,7 @@ static void history_load( wchar_t *name )
 				history_current = malloc( sizeof( ll_node_t ) );
 				if( !history_current )
 				{
-					die_mem();
+					DIE_MEM();
 					
 				}
 				
@@ -161,8 +186,10 @@ static void history_load( wchar_t *name )
 	{
 		if( errno != ENOENT )
 		{
-			debug( 1, L"The following non-fatal error occurred while reading command history from \'%ls\':", name );
+			debug( 1, L"The following non-fatal error occurred while reading command history from \'%ls\':", mode_name );
 			wperror( L"fopen" );
+			res = -1;
+			
 		}
 		
 	}	
@@ -173,7 +200,9 @@ static void history_load( wchar_t *name )
 	free( buff );
 	free( fn );
 	last_loaded = history_last;
-	unblock();
+	signal_unblock();
+	return res;
+	
 }
 
 void history_init()
@@ -190,6 +219,10 @@ static void history_to_hash()
 {
 	history_data *d;
 	
+	if( !mode_name )
+		return;
+	
+
 	d = (history_data *)hash_get( &history_table, 
 								  mode_name );
 	
@@ -198,6 +231,7 @@ static void history_to_hash()
 	d->last=history_last;
 	d->last_loaded=last_loaded;
 	d->count=history_count;
+	d->is_loaded = is_loaded;
 	
 	hash_put( &history_table, 
 			  mode_name,
@@ -212,24 +246,38 @@ void history_set_mode( wchar_t *name )
 	
 	if( mode_name )
 	{		
+		/*
+		  Move the current history to the hashtable
+		*/
 		history_to_hash();		
 	}
+	
+	/*
+	  See if the new history already exists
+	*/
 	curr = (history_data *)hash_get( &history_table,
 									 name );
 	if( curr )
 	{
+		/*
+		  Yes. Restore it.
+		*/
 		mode_name = (wchar_t *)hash_get_key( &history_table,
 											 name );
 		history_current = history_last = curr->last;
 		last_loaded = curr->last_loaded;
 		history_count = curr->count;
+		is_loaded = curr->is_loaded;
 	}
 	else
 	{
+		/*
+		  Nope. Create a new history list.
+		*/
 		history_count=0;
 		history_last = history_current = last_loaded=0;
 		mode_name = wcsdup( name );
-		history_load( name );
+		is_loaded = 0;
 	}
 	
 	past_end=1;
@@ -259,6 +307,9 @@ static void history_save()
 	/* First we save this sessions history in local variables */
 	ll_node_t *real_pos = history_last, *real_first = last_loaded;
 
+	if( !is_loaded )
+		return;
+	
 	if( !real_first )
 	{
 		real_first = history_current;
@@ -274,7 +325,9 @@ static void history_save()
 	history_count=0;
 	past_end=1;
 
-	history_load( mode_name );
+	if( !history_load() )
+	{
+		
 	if( real_pos != 0 )
 	{
 		/* 
@@ -332,6 +385,8 @@ static void history_save()
 		}
 		free( fn );	
 	}
+	}
+	
 	
 }
 
@@ -339,7 +394,7 @@ static void history_save()
    Save the specified mode to file
 */
 
-static void history_destroy_mode( const void *name, const void *link )
+static void history_destroy_mode( void *name, void *link )
 {
 	mode_name = (wchar_t *)name;
 	history_data *d = (history_data *)link;
@@ -349,7 +404,7 @@ static void history_destroy_mode( const void *name, const void *link )
 	past_end=1;
 
 //	fwprintf( stderr, L"Destroy history mode \'%ls\'\n", mode_name );
-		
+
 	if( history_last )
 	{
 		history_save();	
@@ -368,11 +423,12 @@ static void history_destroy_mode( const void *name, const void *link )
 
 void history_destroy()
 {
+	
 	/**
 	   Make sure current mode is in table
 	*/
 	history_to_hash();		
-
+		
 	/**
 	   Save all modes in table
 	*/
@@ -390,6 +446,15 @@ void history_destroy()
 */
 static ll_node_t *history_find( ll_node_t *n, const wchar_t *s )
 {
+	if( !is_loaded )
+	{
+		if( history_load() )
+		{
+			return 0;
+		}
+	}
+	
+
 	if( n == 0 )
 		return 0;
 	
@@ -406,6 +471,15 @@ void history_add( const wchar_t *str )
 {
 	ll_node_t *old_node;
 
+	if( !is_loaded )
+	{
+		if( history_load() )
+		{
+			return;
+		}
+	}
+	
+
 	if( wcslen( str ) == 0 )
 		return;
 	
@@ -455,11 +529,20 @@ static int history_test( const wchar_t *needle, const wchar_t *haystack )
 /*
 	return wcsncmp( haystack, needle, wcslen(needle) )==0;
 */
-	return (int)wcsstr( haystack, needle );
+	return !!wcsstr( haystack, needle );
 }
 
 const wchar_t *history_prev_match( const wchar_t *str )
 {
+	if( !is_loaded )
+	{
+		if( history_load() )
+		{
+			return 0;
+		}
+	}
+	
+
 	if( history_current == 0 )
 		return str;
 	
@@ -481,6 +564,15 @@ const wchar_t *history_prev_match( const wchar_t *str )
 
 const wchar_t *history_next_match( const wchar_t *str)
 {
+	if( !is_loaded )
+	{
+		if( history_load() )
+		{
+			return 0;
+		}
+	}
+	
+
 	if( history_current == 0 )
 		return str;
 
@@ -507,15 +599,27 @@ void history_reset()
 */
 void history_first()
 {
-	while( history_current->prev )
-		history_current = history_current->prev;
+	if( is_loaded )
+		while( history_current->prev )
+			history_current = history_current->prev;
 	
 }
 
 wchar_t *history_get( int idx )
 {
-	ll_node_t *n=history_last;
+	ll_node_t *n;
 	int i;
+	
+	if( !is_loaded )
+	{
+		if( history_load() )
+		{
+			return 0;
+		}
+	}
+	
+	n = history_last;
+	
 	if( idx<0)
 	{
 		debug( 1, L"Tried to access negative history index %d", idx );
