@@ -1,5 +1,5 @@
 /** \file input.c
-	
+a	
 Functions for reading a character of input from stdin, using the
 inputrc information for key bindings.
 
@@ -17,7 +17,15 @@ implementation in fish is as of yet incomplete.
 #include <termios.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+
+#ifdef HAVE_SYS_TERMIOS_H
+#include <sys/termios.h>
+#endif
+
+#ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
+#endif
+
 #include <unistd.h>
 #include <wchar.h>
 
@@ -32,13 +40,21 @@ implementation in fish is as of yet incomplete.
 #include <termio.h>
 #endif
 
+#if HAVE_TERM_H
 #include <term.h>
+#elif HAVE_NCURSES_TERM_H
+#include <ncurses/term.h>
+#endif
+
 #include <signal.h>
 #include <dirent.h>
 #include <wctype.h>
 
 
+
+#include "fallback.h"
 #include "util.h"
+
 #include "wutil.h"
 #include "reader.h"
 #include "proc.h"
@@ -50,7 +66,10 @@ implementation in fish is as of yet incomplete.
 #include "env.h"
 #include "expand.h"
 #include "event.h"
-#include "translate.h"
+#include "signal.h"
+
+#include "output.h"
+#include "intern.h"
 
 static void input_read_inputrc( wchar_t *fn );
 
@@ -61,9 +80,9 @@ static void input_read_inputrc( wchar_t *fn );
 
 typedef struct
 {
-	wchar_t *seq; /**< Character sequence which generates this event */
-	wchar_t *seq_desc; /**< Description of the character sequence suitable for printing on-screen */
-	wchar_t *command; /**< command that should be evaluated by this mapping */
+	const wchar_t *seq; /**< Character sequence which generates this event */
+	const wchar_t *seq_desc; /**< Description of the character sequence suitable for printing on-screen */
+	const wchar_t *command; /**< command that should be evaluated by this mapping */
 		
 }
 	mapping;
@@ -80,7 +99,7 @@ typedef struct
 /**
    Names of all the readline functions supported
 */
-const wchar_t *name_arr[] = 
+static const wchar_t *name_arr[] = 
 {
 	L"beginning-of-line",
 	L"end-of-line",
@@ -109,14 +128,17 @@ const wchar_t *name_arr[] =
 	L"history-token-search-backward",
 	L"history-token-search-forward",
 	L"self-insert",
-	L"null"
+	L"null",
+	L"eof",
+	L"vi-arg-digit"
 }
 	;
 
 /**
    Description of each supported readline function
 */
-const wchar_t *desc_arr[] =
+/*
+static const wchar_t *desc_arr[] =
 {
 	L"Move to beginning of line",
 	L"Move to end of line",
@@ -145,14 +167,16 @@ const wchar_t *desc_arr[] =
 	L"Search backward through list of previous commands for matching token",
 	L"Search forward through list of previous commands for matching token",
 	L"Insert the pressed key",
-	L"Do nothing"
+	L"Do nothing",
+	L"End of file",
+	L"Repeat command"
 }
 	;
-
+*/
 /**
    Internal code for each supported readline function
 */
-const wchar_t code_arr[] = 
+static const wchar_t code_arr[] = 
 {
 	R_BEGINNING_OF_LINE,
 	R_END_OF_LINE,
@@ -181,7 +205,9 @@ const wchar_t code_arr[] =
 	R_HISTORY_TOKEN_SEARCH_BACKWARD,
 	R_HISTORY_TOKEN_SEARCH_FORWARD,
 	R_SELF_INSERT,
-	R_NULL
+	R_NULL,
+	R_EOF,
+	R_VI_ARG_DIGIT
 }
 	;
 
@@ -191,7 +217,18 @@ const wchar_t code_arr[] =
 */
 static hash_table_t all_mappings;
 
-static array_list_t *current_mode_mappings, *current_application_mappings, *global_mappings;
+/**
+   Mappings for the current input mode
+*/
+static array_list_t *current_mode_mappings; 
+/**
+   Mappings for the current application
+*/
+static array_list_t *current_application_mappings;
+/**
+   Global mappings
+*/
+static array_list_t *global_mappings;
 
 /**
    Number of nested conditional statement levels that are not evaluated
@@ -207,7 +244,25 @@ static int inputrc_block_count=0;
 */
 static int inputrc_error = 0;
 
-wchar_t input_get_code( wchar_t *name )
+/**
+   Set to one when the input subsytem has been initialized. 
+*/
+static int is_init = 0;
+
+
+/**
+   This is the variable telling us how many timew the next command
+   should bne repeated. Only actually used in vi-mode.
+*/
+static int repeat_count = 1;
+
+/**
+   This is the type of the first command in a vi-mode two-part combo
+   like 'dw' or '3d3l'.
+*/
+static wint_t first_command = 0;
+
+wchar_t input_get_code( const wchar_t *name )
 {
 
 	int i;
@@ -310,18 +365,16 @@ void add_mapping( const wchar_t *mode,
 		mapping *m = (mapping *)al_get( mappings, i );
 		if( wcscmp( m->seq, s ) == 0 )
 		{
-			free( m->command );
-			free( m->seq_desc );
-			m->seq_desc = wcsdup(d );
-			m->command=wcsdup(c);
+			m->seq_desc = intern(d);
+			m->command = intern(c);
 			return;
 		}
 	}
 	
 	mapping *m = malloc( sizeof( mapping ) );
-	m->seq = wcsdup( s );
-	m->seq_desc = wcsdup(d );
-	m->command=wcsdup(c);
+	m->seq = intern( s );
+	m->seq_desc = intern(d );
+	m->command = intern(c);
 	al_push( mappings, m );	
 }
 
@@ -969,7 +1022,7 @@ void input_parse_inputrc_line( wchar_t *cmd )
 		{
 			if( wcscmp( key, L"editing-mode" ) == 0 )
 			{
-//				current_mode_mappings = get_mapping( value );
+				current_mode_mappings = get_mapping( value );
 			}
 		}
 		
@@ -1069,7 +1122,7 @@ static void input_read_inputrc( wchar_t *fn )
 	int error=0;
 //	fwprintf( stderr, L"read %ls\n", fn );
 		
-	block();
+	signal_block();
 	rc = wfopen( fn, "r" );
 	
 	if( rc )
@@ -1102,9 +1155,12 @@ static void input_read_inputrc( wchar_t *fn )
 			}
 		}
 		free( buff );
+		/*
+		  Don't need to check exit status of fclose on read-only stream
+		*/
 		fclose( rc );
 	}
-	unblock();
+	signal_unblock();
 	
 	inputrc_skip_block_count=0;
 	inputrc_block_count=0;	
@@ -1244,7 +1300,14 @@ static void add_emacs_bindings()
 	add_escaped_mapping( L"emacs", (L"\\C-h"), L"Control-h", L"backward-delete-char" );
 	add_escaped_mapping( L"emacs", (L"\\C-e"), L"Control-e", L"end-of-line" );
 	add_escaped_mapping( L"emacs", (L"\\C-w"), L"Control-w", L"backward-kill-word" );
+	add_escaped_mapping( L"emacs", (L"\\C-p"), L"Control-p", L"history-search-backward" );
+	add_escaped_mapping( L"emacs", (L"\\C-n"), L"Control-n", L"history-search-forward" );
+	add_escaped_mapping( L"emacs", (L"\\C-f"), L"Control-f", L"forward-char" );
+	add_escaped_mapping( L"emacs", (L"\\C-b"), L"Control-b", L"backward-char" );
 	add_escaped_mapping( L"emacs", (L"\e\x7f"), L"Alt-backspace", L"backward-kill-word" );
+	add_escaped_mapping( L"emacs", (L"\eb"), L"Alt-b", L"backward-word" );
+	add_escaped_mapping( L"emacs", (L"\ef"), L"Alt-f", L"forward-word" );
+	add_escaped_mapping( L"emacs", (L"\ed"), L"Alt-d", L"forward-kill-word" );
 	add_terminfo_mapping( L"emacs", (key_ppage), L"Page Up", L"beginning-of-history" );
 	add_terminfo_mapping( L"emacs", (key_npage), L"Page Down", L"end-of-history" );
 }
@@ -1266,12 +1329,29 @@ static void add_vi_bindings()
 	add_mapping( L"vi-command", L"$", L"$", L"end-of-line" );
 	add_mapping( L"vi-command", L"^", L"^", L"beginning-of-line" );
 	add_mapping( L"vi-command", L"0", L"0", L"beginning-of-line" );
+
 	add_mapping( L"vi-command", L"b", L"b", L"backward-word" );
 	add_mapping( L"vi-command", L"B", L"B", L"backward-word" );
 	add_mapping( L"vi-command", L"w", L"w", L"forward-word" );
 	add_mapping( L"vi-command", L"W", L"W", L"forward-word" );
+
 	add_mapping( L"vi-command", L"x", L"x", L"delete-char" );
-	
+
+	add_mapping( L"vi-command", L"1", L"1", L"vi-arg-digit" );
+	add_mapping( L"vi-command", L"2", L"2", L"vi-arg-digit" );
+	add_mapping( L"vi-command", L"3", L"3", L"vi-arg-digit" );
+	add_mapping( L"vi-command", L"4", L"4", L"vi-arg-digit" );
+	add_mapping( L"vi-command", L"5", L"5", L"vi-arg-digit" );
+	add_mapping( L"vi-command", L"6", L"6", L"vi-arg-digit" );
+	add_mapping( L"vi-command", L"7", L"7", L"vi-arg-digit" );
+	add_mapping( L"vi-command", L"8", L"8", L"vi-arg-digit" );
+	add_mapping( L"vi-command", L"9", L"9", L"vi-arg-digit" );
+
+
+	add_mapping( L"vi-command", L"d", L"d", L"vi-delete-to" );
+	add_mapping( L"vi-command", L"D", L"D", L"vi-delete-to" );
+
+
 /*
   movement ("h", "l"), word movement
   ("b", "B", "w", "W", "e", "E"), moving to beginning and end of line
@@ -1292,12 +1372,32 @@ static int interrupt_handler()
 	  Fire any pending events
 	*/
 	event_fire( 0 );	
+	
+	/*
+	  Reap stray processes, including printing exit status messages
+	*/
 	if( job_reap( 1 ) )
 		repaint();
+	
+	/*
+	  Check if we should exit
+	*/
+	if( exit_status() )
+	{
+		return R_EXIT;
+	}
+	
+	/*
+	  Tell the reader an event occured
+	*/
 	if( reader_interupted() )
 	{
+		/*
+		  Return 3, i.e. the character read by a Control-C.
+		*/
 		return 3;
 	}
+
 	return 0;	
 }
 
@@ -1305,6 +1405,11 @@ int input_init()
 {
 	wchar_t *fn;
 	
+	if( is_init )
+		return 1;
+	
+	is_init = 1;
+
 	input_common_init( &interrupt_handler );
 	
 	if( setupterm( 0, STDOUT_FILENO, 0) == ERR )
@@ -1360,22 +1465,13 @@ int input_init()
 /**
    Free memory used by the specified mapping
 */
-static void destroy_mapping( const void *key, const void *val )
+static void destroy_mapping( void *key, void *val )
 {
-	int i;
 	array_list_t *mappings = (array_list_t *)val;
 	
-	for( i=0; i<al_get_count( mappings ); i++ )
-	{
-		mapping *m = (mapping *)al_get( mappings, i );
-		free( m->seq );
-		free( m->seq_desc );
-		
-		free( m->command );
-		free(m );
-	}
-		
+	al_foreach( mappings, &free );		
 	al_destroy( mappings );
+
 	free((void *)key);
 	free((void *)val);	
 }
@@ -1383,12 +1479,21 @@ static void destroy_mapping( const void *key, const void *val )
 
 void input_destroy()
 {
+	if( !is_init )
+		return;
+	
+	is_init=0;
+	
 	input_common_destroy();
 	
 	hash_foreach( &all_mappings, &destroy_mapping );	
 	hash_destroy( &all_mappings );
 	
-	del_curterm( cur_term );
+	if( del_curterm( cur_term ) == ERR )
+	{
+		debug( 0, _(L"Error while closing terminfo") );
+	}
+	
 }
 
 /**
@@ -1396,6 +1501,8 @@ void input_destroy()
 */
 static wint_t input_exec_binding( mapping *m, const wchar_t *seq )
 {
+	int i;
+	
 //	fwprintf( stderr, L"Binding %ls\n", m->command );
 	wchar_t code = input_get_code( m->command );
 	if( code != -1 )
@@ -1404,15 +1511,63 @@ static wint_t input_exec_binding( mapping *m, const wchar_t *seq )
 		{
 			case R_DUMP_FUNCTIONS:
 			{
-				dump_functions();
+				for( i=0; i<repeat_count; i++ )
+					dump_functions();
+				repeat_count = 1;				
 				return R_NULL;							
 			}
+
 			case R_SELF_INSERT:
 			{
+				for( i=1; i<repeat_count; i++ )
+					input_unreadch( seq[0] );							
+				repeat_count = 1;				
 				return seq[0];							
 			}
+
+			case R_VI_ARG_DIGIT:
+			{
+				int repeat = seq[0]-L'0';
+				if( repeat > 0 && repeat <= 9 )
+					repeat_count *= repeat;
+				
+				return R_NULL;
+			}
+			
+			case R_VI_DELETE_TO:
+			{
+				first_command = R_VI_DELETE_TO;
+				return R_NULL;
+			}
+
 			default:
+			{
+				
+				if( first_command )
+				{
+					switch( first_command )
+					{
+						case R_VI_DELETE_TO:
+						{
+						
+							break;
+							
+						}
+					}
+				}
+				else
+				{
+					for( i=1; i<repeat_count; i++ )
+					{
+						input_unreadch( code );							
+					}
+				}
+				
+				repeat_count = 1;				
+				first_command = 0;
 				return code;
+			}
+			
 		}	
 	}
 	else
@@ -1451,6 +1606,13 @@ static wint_t input_try_mapping( mapping *m)
 {
 	int j, k;
 	wint_t c=0;
+
+	c = input_common_readch( 0 );
+	if( c == input_get_code( m->command ) )
+	{
+		return input_exec_binding( m, m->seq );
+	}
+	input_unreadch( c );
 	
 	if( m->seq != 0 )
 	{
@@ -1464,9 +1626,14 @@ static wint_t input_try_mapping( mapping *m)
 		}
 		else
 		{
+			/*
+			  Return the read characters
+			*/
 			input_unreadch(c);
-			for(k=j-1; k>=0; k--)
-				input_unreadch(m->seq[k]);
+			for( k=j-1; k>=0; k-- )
+			{
+				input_unreadch( m->seq[k] );
+			}
 		}
 	}		
 	return 0;
