@@ -9,6 +9,7 @@ inputrc information for key bindings.
 #define FISH_INPUT_H
 
 #include <wchar.h>
+#include "input_common.h"
 
 /**
    Key codes for inputrc-style keyboard functions that are passed on
@@ -16,7 +17,7 @@ inputrc information for key bindings.
 */
 enum
 {
-	R_BEGINNING_OF_LINE = R_NULL+1, 
+	R_BEGINNING_OF_LINE = R_NULL+10,  /* This give input_common ten slots for lowlevel keycodes */
 	R_END_OF_LINE,
 	R_FORWARD_CHAR,
 	R_BACKWARD_CHAR,
@@ -43,6 +44,8 @@ enum
 	R_HISTORY_TOKEN_SEARCH_BACKWARD,
 	R_HISTORY_TOKEN_SEARCH_FORWARD,
 	R_SELF_INSERT,
+	R_VI_ARG_DIGIT,
+	R_VI_DELETE_TO
 }
 ;
 
@@ -77,8 +80,8 @@ wint_t input_readch();
 /**
    Push a character or a readline function onto the stack of unread
    characters that input_readch will return before actually reading from fd
-   0.
-*/
+   0.   
+ */
 void input_unreadch( wint_t ch );
 
 
@@ -88,7 +91,7 @@ void input_unreadch( wint_t ch );
    \param mode the name of the mapping mode to add this mapping to
    \param s the sequence
    \param d a description of the sequence
-   \param c am input function that will be run whenever the key sequence occurs
+   \param cmd an input function that will be run whenever the key sequence occurs
 */
 void add_mapping( const wchar_t *mode, const wchar_t *s, const wchar_t * d, const wchar_t *cmd );
 
@@ -110,6 +113,6 @@ void input_parse_inputrc_line( wchar_t *cmd );
 /**
    Returns the function for the given function name.
 */
-wchar_t input_get_code( wchar_t *name );
+wchar_t input_get_code( const wchar_t *name );
 
 #endif
