@@ -5,14 +5,22 @@ Implementation file for the low level input library
 */
 #include "config.h"
 
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
+#include <sys/time.h>
+#include <sys/types.h>
 #include <unistd.h>
 #include <wchar.h>
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
 
+#include "fallback.h"
 #include "util.h"
+
 #include "common.h"
 #include "wutil.h"
 #include "input_common.h"
@@ -28,7 +36,7 @@ Implementation file for the low level input library
 /**
    Characters that have been read and returned by the sequence matching code
 */
-static wint_t lookahead_arr[32];
+static wint_t lookahead_arr[1024];
 
 /**
    Number of entries in lookahead_arr
@@ -58,7 +66,7 @@ static wint_t readb()
 {
 	unsigned char arr[1];
 	int do_loop = 0;
-
+	
 	do
 	{
 		fd_set fd;	
@@ -83,28 +91,25 @@ static wint_t readb()
 				case EINTR:
 				case EAGAIN:
 				{
-//					wperror( L"select" );
 					if( interrupt_handler )
 					{
 						int res = interrupt_handler();
-						
-/*						debug( 0,
-							   L"interrupt, %d is %ls", 
-							   res, 
-							   (res==R_NULL?L"good": L"Bad") );
-*/
 						if( res )
+						{
 							return res;
+						}
 					}
 					
+					
 					do_loop = 1;
 					break;
 				}
 				default:
 				{
-					debug( 0, L"Error while reading input from keyboard, shutting down" );
-					wperror(L"read");
-					exit(1);
+					/*
+					  The teminal has been closed. Save and exit.
+					*/
+					return R_EOF;
 				}
 			}	
 		}
@@ -116,17 +121,17 @@ static wint_t readb()
 				{
 					debug( 3, L"Wake up on universal variable event" );					
 					env_universal_read_all();
-					debug( 3, L"Return R_NULL" );					
-					return R_NULL;
+					do_loop = 1;
 				}				
 			}
 			if( FD_ISSET( 0, &fd ) )
 			{
-				if( read_blocked( 0, arr, 1 ) == -1 )
+				if( read_blocked( 0, arr, 1 ) != 1 )
 				{
-					debug( 0, L"Error while reading input from keyboard, shutting down" );
-					wperror(L"read");
-					exit(1);
+					/*
+					  The teminal has been closed. Save and exit.
+					*/
+					return R_EOF;
 				}
 				do_loop = 0;
 			}				
@@ -137,6 +142,7 @@ static wint_t readb()
 	return arr[0];
 }
 
+
 wchar_t input_common_readch( int timed )
 {
 	if( lookahead_count == 0 )
@@ -180,8 +186,8 @@ wchar_t input_common_readch( int timed )
 			
 			int sz;
 			
-			if( b == R_NULL )
-				return R_NULL;
+			if( (b == R_NULL) || (b == R_EOF) )
+				return b;
 
 			bb=b;
 			
