@@ -6,6 +6,8 @@
 	with the X clipboard.
 */
 
+#include "config.h"
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <wchar.h>
@@ -17,17 +19,18 @@
 #include <dirent.h>
 
 
-#include "config.h"
+#include "fallback.h"
 #include "util.h"
+
 #include "wutil.h"
 #include "kill.h"
 #include "proc.h"
 #include "sanity.h"
 #include "common.h"
 #include "env.h"
-#include "expand.h"
 #include "exec.h"
 #include "parser.h"
+#include "halloc.h"
 
 /**
    Maximum entries in killring
@@ -46,14 +49,11 @@ static wchar_t *cut_buffer=0;
 */
 static int has_xsel()
 {
-	wchar_t *path = get_filename( L"xsel" );
-	if( path)
-	{
-		free(path);
-		return 1;
-	}
-	else
-		return 0;
+	void *context = halloc(0, 0);
+	wchar_t *path = parser_get_filename( context, L"xsel" );
+	int res = !!path;
+	halloc_free( context );
+	return res;
 }
 
 
@@ -95,7 +95,7 @@ void kill_add( wchar_t *str )
 	wchar_t *disp;
 	if( (disp = env_get( L"DISPLAY" )) )
 	{
-		wchar_t *escaped_str = expand_escape( str, 1 );
+		wchar_t *escaped_str = escape( str, 1 );
 		wchar_t *cmd = wcsdupcat2(L"echo ", escaped_str, L"|xsel -b",0);
 		exec_subshell( cmd, 0 );
 		free( cut_buffer );
@@ -137,7 +137,7 @@ static void kill_check_x_buffer()
 		
 		for( i=0; i<al_get_count( &list ); i++ )
 		{
-			wchar_t *next_line = expand_escape( (wchar_t *)al_get( &list, i ), 0);
+			wchar_t *next_line = escape( (wchar_t *)al_get( &list, i ), 0 );
 			if( i==0 )
 			{
 				new_cut_buffer = next_line;
@@ -179,7 +179,7 @@ static void kill_check_x_buffer()
 			}
 		}
 		
-		al_foreach( &list, (void (*)(const void *))&free );
+		al_foreach( &list, &free );
 		al_destroy( &list );
 	}
 }
