@@ -3,8 +3,7 @@ Copyright (C) 2005 Axel Liljencrantz
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
-as published by the Free Software Foundation; either version 2
-of the License, or (at your option) any later version.
+as published by the Free Software Foundation.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -23,6 +22,7 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 
 #include "config.h"
 
+
 #include <stdlib.h>
 #include <stdio.h>
 #include <wchar.h>
@@ -40,7 +40,9 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include <locale.h>
 #include <signal.h>
 
+#include "fallback.h"
 #include "util.h"
+
 #include "common.h"
 #include "reader.h"
 #include "builtin.h"
@@ -56,7 +58,14 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 #include "exec.h"
 #include "event.h"
 #include "output.h"
-#include "translate.h"
+
+#include "halloc_util.h"
+#include "history.h"
+
+/**
+   The string describing the single-character options accepted by the main fish binary
+*/
+#define GETOPT_STRING "hilnvc:p:d:"
 
 /**
    Parse init files
@@ -65,23 +74,23 @@ static int read_init()
 {
 	char cwd[4096];
 	wchar_t *wcwd;
-	
+
 	if( !getcwd( cwd, 4096 ) )
 	{
 		wperror( L"getcwd" );		
 		return 0;
 	}
 
-	env_set( L"__fish_help_dir", DOCDIR, 0);	
- 	
-	eval( L"builtin cd " SYSCONFDIR L" 2>/dev/null; . fish 2>/dev/null", 0, TOP );
-	eval( L"builtin cd 2>/dev/null;. .fish 2>/dev/null", 0, TOP );
-	
+	eval( L"builtin cd " DATADIR L"/fish 2>/dev/null; and builtin . fish 2>/dev/null", 0, TOP );
+	eval( L"builtin cd " SYSCONFDIR L" 2>/dev/null; and builtin . fish 2>/dev/null", 0, TOP );
+	eval( L"builtin cd 2>/dev/null; and builtin . .fish 2>/dev/null", 0, TOP );
+
 	if( chdir( cwd ) == -1 )
 	{
-//		fwprintf( stderr, L"Invalid directory: %s\n", cwd );
-//		wperror( L"chdir" );
-//		return 0;
+		/*
+		  If we can't change back to previos directory, we'll stay in
+		  ~. Should be a sane default behavior.
+		*/
 	}	
 	wcwd = str2wcs( cwd );
 	if( wcwd )
@@ -106,13 +115,14 @@ int main( int argc, char **argv )
 	
 	char *cmd=0;	
 
+	halloc_util_init();	
+
 	wsetlocale( LC_ALL, L"" );
 	is_interactive_session=1;
 	program_name=L"fish";
 	
 	while( 1 )
 	{
-#ifdef HAVE_GETOPT_LONG
 		static struct option
 			long_options[] =
 			{
@@ -120,10 +130,22 @@ int main( int argc, char **argv )
 					"command", required_argument, 0, 'c' 
 				}
 				,
+				{
+					"debug-level", required_argument, 0, 'd' 
+				}
+				,
 				{
 					"interactive", no_argument, 0, 'i' 
 				}
 				,
+				{
+					"login", no_argument, 0, 'l' 
+				}
+				,
+				{
+					"no-execute", no_argument, 0, 'n' 
+				}
+				,
 				{
 					"profile", required_argument, 0, 'p' 
 				}
@@ -146,84 +168,119 @@ int main( int argc, char **argv )
 		
 		int opt = getopt_long( argc,
 							   argv, 
-							   "hivc:p:", 
+							   GETOPT_STRING,
 							   long_options, 
 							   &opt_index );
 		
-#else	
-		int opt = getopt( argc,
-						  argv, 
-						  "hivc:p:" );
-#endif
 		if( opt == -1 )
 			break;
 		
 		switch( opt )
 		{
 			case 0:
+			{
 				break;
-				
+			}
+			
 			case 'c':		
+			{
 				cmd = optarg;				
 				is_interactive_session = 0;
 				break;
-
-			case 'h':
-				cmd = "help";
-				//interactive=0;
-				
+			}
+			
+			case 'd':		
+			{
+				char *end;
+				int tmp = strtol(optarg, &end, 10);
+				if( tmp >= 0 && tmp <=10 && !*end )
+				{
+					debug_level=tmp;
+				}
+				else
+				{
+					debug( 0, _(L"Invalid value '%s' for debug level switch"), optarg );
+					exit(1);
+				}
 				break;
-
+			}
+			
+			case 'h':
+			{
+				cmd = "help";
+				break;
+			}
+			
 			case 'i':
+			{
 				force_interactive = 1;
 				break;				
-				
+			}
+			
+			case 'l':
+			{
+				is_login=1;
+				break;				
+			}
+			
+			case 'n':
+			{
+				no_exec=1;
+				break;				
+			}
+			
 			case 'p':
+			{
 				profile = optarg;
 				break;				
-				
+			}
+			
 			case 'v':
+			{
 				fwprintf( stderr, 
-						  L"%s, version %s\n", 
+						  _(L"%s, version %s\n"), 
 						  PACKAGE_NAME,
 						  PACKAGE_VERSION );
 				exit( 0 );				
-
+			}
+			
 			case '?':
+			{
 				return 1;
-				
+			}
+			
 		}		
 	}
 
 	my_optind = optind;
 	
-	is_login |= strcmp( argv[0], "-fish") == 0;
-//	fwprintf( stderr, L"%s\n", argv[0] );
+	is_login |= (strcmp( argv[0], "-fish") == 0);
 		
 	is_interactive_session &= (cmd == 0);
 	is_interactive_session &= (my_optind == argc);
 	is_interactive_session &= isatty(STDIN_FILENO);	
+	is_interactive_session |= force_interactive;
+
+	/*
+	  No-exec is prohibited when in interactive mode
+	*/
+	if( is_interactive_session && no_exec)
+	{
+		debug( 1, _(L"Can not use the no-execute mode when running an interactive session") );
+		no_exec = 0;
+	}
 	
-//	fwprintf( stderr, L"%d %d %d\n", cmd==0, my_optind == argc, isatty(STDIN_FILENO) );
 
-	if( force_interactive )
-		is_interactive_session=1;	
-
-	translate_init();	
 	proc_init();	
-	output_init();	
 	event_init();	
-	exec_init();	
 	wutil_init();
 	parser_init();
 	builtin_init();
 	function_init();
 	env_init();
-	complete_init();
 	reader_init();
-	
-	reader_push_current_filename( L"(internal)" );
-	
+	history_init();
+
 	if( read_init() )
 	{
 		if( cmd != 0 )
@@ -231,15 +288,13 @@ int main( int argc, char **argv )
 			wchar_t *cmd_wcs = str2wcs( cmd );
 			res = eval( cmd_wcs, 0, TOP );
 			free(cmd_wcs);
-			reader_exit(0);			
+			reader_exit(0, 0);
 		}
 		else
 		{
 			if( my_optind == argc )
 			{
-				reader_push_current_filename( L"(stdin)" );
 				res = reader_read( 0 );				
-				reader_pop_current_filename();
 			}
 			else
 			{
@@ -248,17 +303,18 @@ int main( int argc, char **argv )
 				int i; 
 				string_buffer_t sb;
 				int fd;
-				
+				wchar_t *rel_filename, *abs_filename;
+								
 				if( ( fd = open(file, O_RDONLY) ) == -1 )
 				{
 					wperror( L"open" );
 					return 1;
 				}
 
-				sb_init( &sb );
-				
 				if( *(argv+2))
 				{
+					sb_init( &sb );
+				
 					for( i=1,ptr = argv+2; *ptr; i++, ptr++ )
 					{
 						if( i != 1 )
@@ -271,40 +327,46 @@ int main( int argc, char **argv )
 					env_set( L"argv", (wchar_t *)sb.buff, 0 );
 					sb_destroy( &sb );
 				}
-				
-				reader_push_current_filename( str2wcs( file ) );
+
+				rel_filename = str2wcs( file );
+				abs_filename = wrealpath( rel_filename, 0 );
+				if( !abs_filename )
+					abs_filename = wcsdup(rel_filename);
+				reader_push_current_filename( intern( abs_filename ) );
+				free( rel_filename );
+				free( abs_filename );
+
 				res = reader_read( fd );
 
 				if( res )
 				{
 					debug( 1, 
-						   L"Error while reading file %ls\n", 
-						   reader_current_filename() );
+						   _(L"Error while reading file %ls\n"), 
+						   reader_current_filename()?reader_current_filename(): _(L"Standard input") );
 				}				
-				free(reader_pop_current_filename());
+				reader_pop_current_filename();
 			}
 		}
 	}
 
 	proc_fire_event( L"PROCESS_EXIT", EVENT_EXIT, getpid(), res );
 	
-	reader_pop_current_filename();	
-	
+
+	history_destroy();
 	proc_destroy();
-	env_destroy();
 	builtin_destroy();
 	function_destroy();
-	complete_destroy();
 	reader_destroy();
 	parser_destroy();
 	wutil_destroy();
-	common_destroy();
-	exec_destroy();	
 	event_destroy();
-	output_destroy();
-	translate_destroy();	
-	
+
+	halloc_util_destroy();
+
+	env_destroy();
+
 	intern_free_all();
 
+
 	return res;	
 }
