@@ -10,21 +10,6 @@
 
 #include <wchar.h>
 
-enum
-{
-	HIGHLIGHT_NORMAL,
-	HIGHLIGHT_COMMAND,
-	HIGHLIGHT_SUBSHELL,
-	HIGHLIGHT_REDIRECTION, 
-	HIGHLIGHT_END, 
-	HIGHLIGHT_ERROR,
-	HIGHLIGHT_PARAM,
-	HIGHLIGHT_COMMENT,
-	HIGHLIGHT_MATCH,
-	HIGHLIGHT_SEARCH_MATCH,
-}
-	;
-
 /**
    Constants for various colors as used by the set_color function. 
 */
@@ -39,10 +24,20 @@ enum
 	FISH_COLOR_CYAN,
 	FISH_COLOR_WHITE,
 	/** The default fg color of the terminal */
-	FISH_COLOR_NORMAL
+	FISH_COLOR_NORMAL,
+	FISH_COLOR_IGNORE,
+	FISH_COLOR_RESET
 }
 ;
- 
+
+/**
+   The value to send to set_color to tell it to use a bold font
+*/
+#define FISH_COLOR_BOLD 0x80
+/**
+   The value to send to set_color to tell it to underline the text
+*/
+#define FISH_COLOR_UNDERLINE 0x100
 
 /**
    Sets the fg and bg color. May be called as often as you like, since
@@ -80,7 +75,9 @@ void set_color( int c, int c2 );
 
 /**
    Write a char * narrow string to FD 1, needed for the terminfo
-   strings.
+   strings. This is usually just a wrapper aound tputs, using writeb
+   as the sending function. But a weird bug on PPC Linux means that on
+   this platform, write is instead used directly.
 */
 int writembs( char *str );
 
@@ -117,13 +114,25 @@ int writespace( int c );
 int output_color_code( const wchar_t *val );
 
 /**
-   Initialize static data
+   perm_left_cursor and parm_right_cursor don't seem to be defined
+   very often so we use cursor_left and cursor_right as a fallback.
 */
-void output_init();
+void move_cursor( int steps );
 
 /**
-  Destroy static data
+   This is for writing process notification messages. Has to write to
+   stdout, so clr_eol and such functions will work correctly. Not an
+   issue since this function is only used in interactive mode anyway.
 */
-void output_destroy();
+int writeb( tputs_arg_t b );
+
+/**
+   Set the function used for writing in move_cursor, writespace and
+   set_color and all other output functions in this library. By
+   default, the write call is used to give completely unbuffered
+   output to stdout.
+*/
+void output_set_writer( int (*writer)(char) );
+
 
 #endif
