@@ -10,7 +10,11 @@
 #include "proc.h"
 #include "util.h"
 #include "parser.h"
+#include "event.h"
 
+/**
+   event_block_t represents a block on events of the specified type
+*/
 typedef struct event_block
 {
 	/**
@@ -60,7 +64,9 @@ typedef struct block
 		wchar_t *for_variable; /**< Name of the variable to loop over */
 		int if_state; /**< The state of the if block */
 		wchar_t *switch_value; /**< The value to test in a switch block */
-		wchar_t *function_name; /**< The name of the function to define */
+		wchar_t *function_name; /**< The name of the function to define or the function called*/
+		const wchar_t *source_dest; /**< The name of the file to source*/
+		event_t *event; /**<The event that triggered this block */		
 	} param1;
 
 	/**
@@ -71,6 +77,7 @@ typedef struct block
 		array_list_t for_vars; /**< List of values for a for block */	
 		int switch_taken; /**< Whether a switch match has already been found */
 		wchar_t *function_description; /**< The description of the function to define */
+		process_t *function_call_process;		/**< The process representing this function call */
 	} param2;
 
 	/**
@@ -88,6 +95,16 @@ typedef struct block
 	{
 		array_list_t *function_events;
 	} param4;
+
+	/**
+	   Name of file that created this block
+	*/
+	const wchar_t *src_filename;
+	
+	/**
+	   Line number where this block was created
+	*/
+	int src_lineno;
 	
 	/**
 	   Some naming confusion. This is a pointer to the first element in the list of all event blocks.
@@ -115,6 +132,8 @@ enum block_type
 	SUBST, /**< Command substitution scope */
 	TOP, /**< Outermost block */
 	BEGIN, /**< Unconditional block */
+	SOURCE, /**< Block created by the . (source) builtin */
+	EVENT, /**< Block created on event notifier invocation */
 }
 ;
 
@@ -160,9 +179,9 @@ enum parser_error
 	*/
 	EVAL_ERROR,
 	/**
-	   Error while evaluating subshell
+	   Error while evaluating cmdsubst
 	*/
-	SUBSHELL_ERROR,
+	CMDSUBST_ERROR,
 }
 ;
 
@@ -172,10 +191,6 @@ extern block_t *current_block;
 /** Global event blocks */
 extern event_block_t *global_event_block;
 
-
-/** The current error code */
-extern int error_code;
-
 /**
    Current block level io redirections 
 */
@@ -185,23 +200,24 @@ extern io_data_t *block_io;
   Finds the full path of an executable in a newly allocated string.
   
   \param cmd The name of the executable.
+  \param context the halloc context to use for memory allocations
   \return 0 if the command can not be found, the path of the command otherwise.
 */
-wchar_t *get_filename( const wchar_t *cmd );
+wchar_t *parser_get_filename( void *context, const wchar_t *cmd );
 
 /**
   Evaluate the expressions contained in cmd.
 
   \param cmd the string to evaluate
-  \param out buffer to insert output to. May be null.
-  \param the type of block to push onto the scope stack
+  \param io io redirections to perform on all started jobs
   \param block_type The type of block to push on the block stack
-  \return 0 on success.
+
+  \return 0 on success, 1 otherwise
 */
 int eval( const wchar_t *cmd, io_data_t *io, int block_type );
 
 /**
-  Evaluate line as a list of parameters, i.e. tokenize it and perform parameter expansion and subshell execution on the tokens.
+  Evaluate line as a list of parameters, i.e. tokenize it and perform parameter expansion and cmdsubst execution on the tokens.
   The output is inserted into output, and should be freed by the caller.
 
   \param line Line to evaluate
@@ -235,10 +251,16 @@ int parser_is_subcommand( const wchar_t *cmd );
    command scope, like 'for', 'end' or 'command' or 'exec'. These
    functions may not be overloaded, so their names are reserved.
 
-   \param cmd The command name to test
+   \param word The command name to test
    \return 1 of the command parameter is a command, 0 otherwise
 */
-int parser_is_reserved( wchar_t *word);
+int parser_is_reserved( const wchar_t *word );
+
+/**
+   Test if the specified string is command that opens a new block
+*/
+
+int parser_is_block( const wchar_t *word);
 
 /**
    Returns a string describing the current parser pisition in the format 'FILENAME (line LINE_NUMBER): LINE'.
@@ -248,6 +270,11 @@ int parser_is_reserved( wchar_t *word);
 */
 wchar_t *parser_current_line();
 
+/**
+   Returns the current line number
+*/
+int parser_get_lineno();
+
 /**
    Returns the current position in the latest string of the tokenizer.
 */
@@ -285,10 +312,20 @@ const wchar_t *parser_get_block_desc( int block );
 
 
 /**
-   Test if the specified string can be parsed, or if more bytes need to be read first. 
-   The result has the first bit set if the string contains errors, and the second bit is set if the string contains an unclosed block.
+   Test if the specified string can be parsed, or if more bytes need
+   to be read first.  The result has the first bit set if the string
+   contains errors, and the second bit is set if the string contains
+   an unclosed block.
 */
-int parser_test( wchar_t * buff, int babble );
+int parser_test( const wchar_t * buff, string_buffer_t *out, const wchar_t *prefix );
+
+/**
+   Test if the specified string can be parsed as an argument list,
+   e.g. sent to eval_args.  The result has the first bit set if the
+   string contains errors, and the second bit is set if the string
+   contains an unclosed block.
+*/
+int parser_test_args( const wchar_t * buff, string_buffer_t *out, const wchar_t *prefix );
 
 /**
    Returns the full path of the specified directory. If the \c in is a
@@ -297,7 +334,7 @@ int parser_test( wchar_t * buff, int babble );
    directories i the CDPATH, the full path is returned. If no
    directory can be found, 0 is returned.
 */
-wchar_t *parser_cdpath_get( wchar_t *in );
+wchar_t *parser_cdpath_get( void *context, wchar_t *in );
 
 /**
    Tell the parser that the specified function may not be run if not
@@ -328,4 +365,17 @@ void parser_destroy();
 */
 int parser_is_help( wchar_t *s, int min_match );
 
+/**
+   Returns the file currently evaluated by the parser. This can be
+   different than reader_current_filename, e.g. if we are evaulating a
+   function defined in a different file than the one curently read.
+*/
+const wchar_t *parser_current_filename();
+
+/**
+   Write a stack trace starting at the specified block to the specified string_buffer_t
+*/
+void parser_stack_trace( block_t *b, string_buffer_t *buff);
+
+
 #endif
