@@ -20,7 +20,15 @@ Some of the code in this file is based on code from the Glibc manual.
 #include <termios.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+
+#ifdef HAVE_SYS_TERMIOS_H
+#include <sys/termios.h>
+#endif
+
+#ifdef HAVE_SYS_IOCTL_H
 #include <sys/ioctl.h>
+#endif
+
 #include <unistd.h>
 #include <signal.h>
 #include <dirent.h>
@@ -36,9 +44,23 @@ Some of the code in this file is based on code from the Glibc manual.
 #include <termio.h>
 #endif
 
+#if HAVE_TERM_H
 #include <term.h>
+#elif HAVE_NCURSES_TERM_H
+#include <ncurses/term.h>
+#endif
 
+#ifdef HAVE_SIGINFO_H
+#include <siginfo.h>
+#endif
+
+#ifdef HAVE_SYS_SELECT_H
+#include <sys/select.h>
+#endif
+
+#include "fallback.h"
 #include "util.h"
+
 #include "wutil.h"
 #include "proc.h"
 #include "common.h"
@@ -48,7 +70,10 @@ Some of the code in this file is based on code from the Glibc manual.
 #include "parser.h"
 #include "signal.h"
 #include "event.h"
-#include "translate.h"
+
+#include "halloc.h"
+#include "halloc_util.h"
+#include "output.h"
 
 /**
    Size of message buffer 
@@ -71,7 +96,7 @@ static int last_status=0;
 static sig_atomic_t got_signal=0;
 
 job_t *first_job=0;
-int is_interactive=0;
+int is_interactive=-1;
 int is_interactive_session=0;
 int is_subshell=0;
 int is_block=0;
@@ -79,6 +104,9 @@ int is_login=0;
 int is_event=0;
 int proc_had_barrier;
 pid_t proc_last_bg_pid = 0;
+int job_control_mode = JOB_CONTROL_INTERACTIVE;
+int no_exec=0;
+
 
 /**
    The event variable used to send all process event
@@ -95,43 +123,21 @@ static string_buffer_t event_pid;
 */
 static string_buffer_t event_status;
 
+/**
+   A stack containing the values of is_interactive. Used by proc_push_interactive and proc_pop_interactive.
+*/
+static array_list_t *interactive_stack;
 
 void proc_init()
 {
+	interactive_stack = al_halloc( global_context );
+	proc_push_interactive( 0 );
 	al_init( &event.arguments );
 	sb_init( &event_pid );
 	sb_init( &event_status );
 }
 
 
-/**
-   Recursively free a process and those following it
-*/
-static void free_process( process_t *p )
-{
-	wchar_t **arg;
-	
-	if( p==0 )
-		return;
-
-	
-
-	free_process( p->next );
-	debug( 3, L"Free process %ls", p->actual_cmd );
-	free( p->actual_cmd );
-	if( p->argv != 0 )
-	{
-		debug( 3, L"Process has argument vector" );
-		for( arg=p->argv; *arg; arg++ )
-		{
-			debug( 3, L"Free argument %ls", *arg );
-			free( *arg );
-		}		
-		free(p->argv );
-	}
-	free( p );
-}
-
 /**
    Remove job from list of jobs 
 */
@@ -165,30 +171,8 @@ static int job_remove( job_t *j )
 */
 void job_free( job_t * j )
 {
-	io_data_t *io, *ionext;
-	
-//	fwprintf( stderr, L"Remove job %d (%ls)\n", j->job_id, j->command );	
-
 	job_remove( j );
-	
-	/* Then free ourselves */
-	free_process( j->first_process);
-	
-	if( j->command != 0 )
-		free( j->command );
-	
-	for( io=j->io; io; io=ionext )
-	{
-		ionext = io->next;
-//		fwprintf( stderr, L"Kill redirect %d of type %d\n", ionext, io->io_mode );
-		if( io->io_mode == IO_FILE )
-		{
-			free( io->param1.filename );
-		}
-		free( io );		
-	}
-	
-	free( j );
+	halloc_free( j );
 }
 
 void proc_destroy()
@@ -205,6 +189,7 @@ void proc_destroy()
 
 void proc_set_last_status( int s )
 {
+
 	last_status = s;
 	//	fwprintf( stderr, L"Set last status to %d\n", s );
 }
@@ -216,15 +201,19 @@ int proc_get_last_status()
 
 job_t *job_create()
 {
-	int free_id=0;
+	int free_id=1;
 	job_t *res;
 	
 	while( job_get( free_id ) != 0 )
 		free_id++;
-	res = calloc( 1, sizeof(job_t) );
+	res = halloc( 0, sizeof(job_t) );
 	res->next = first_job;
 	res->job_id = free_id;
 	first_job = res;
+
+	res->job_control = (job_control_mode==JOB_CONTROL_ALL) || 
+		((job_control_mode == JOB_CONTROL_INTERACTIVE) && (is_interactive));
+
 //	if( res->job_id > 2 )
 //		fwprintf( stderr, L"Create job %d\n", res->job_id );	
 	return res;
@@ -234,7 +223,7 @@ job_t *job_create()
 job_t *job_get( int id )
 {
 	job_t *res = first_job;
-	if( id == -1 )
+	if( id <= 0 )
 	{
 		return res;
 	}
@@ -283,7 +272,7 @@ int job_is_stopped( const job_t *j )
 
 
 /* 
-   Return true if all processes in the job have completed.  
+   Return true if the last processes in the job has completed.  
 
    \param j the job to test
 */
@@ -291,32 +280,13 @@ int job_is_completed( const job_t *j )
 {
 	process_t *p;
 	
-	for (p = j->first_process; p; p = p->next)
-	{
-		if (!p->completed)
-		{
-//			fwprintf( stderr, L"Process %ls not finished\n", p->argv[0] );
-			return 0;
-		}
-	}
-	return 1;
-}
-
-/**
-   Return true if all processes in the job have completed.  
-
-   \param j the job to test
-*/
-static int job_last_is_completed( const job_t *j )
-{
-	process_t *p;
-	
 	for (p = j->first_process; p->next; p = p->next)
 		;
+	
 	return p->completed;
+	
 }
 
-
 /**
    Store the status of the process pid that was returned by waitpid.
    Return 0 if all went well, nonzero otherwise.  
@@ -384,7 +354,7 @@ static void handle_child_status( pid_t pid, int status )
   "Process %d is %ls from job %ls\n",
   (int) pid, p->actual_cmd, j->command );
   write( 2, mess, strlen(mess ));
-*/			
+*/		
 				
 				mark_process_status ( j, p, status);
 				if( p->completed && prev != 0  )
@@ -484,7 +454,7 @@ void job_handle_signal ( int signal, siginfo_t *info, void *con )
 				handle_child_status( pid, status );
 				break;
 		}
-	}
+	}	
 	kill( 0, SIGIO );
 	errno=errno_old;
 }
@@ -533,6 +503,11 @@ int job_reap( int interactive )
 	static int locked = 0;
 	
 	locked++;	
+	
+	/*
+	  job_read may fire an event handler, we do not want to call
+	  ourselves recursively (to avoid infinite recursion).
+	*/
 	if( locked>1 )
 		return 0;
 		
@@ -541,6 +516,11 @@ int job_reap( int interactive )
 		process_t *p;
 		jnext = j->next;
 		
+		/*
+		  If we are reaping only jobs who do not need status messages
+		  sent to the console, do not consider reaping jobs that need
+		  status messages
+		*/
 		if( (!j->skip_notification) && (!interactive) && (!j->fg))
 		{
 			continue;
@@ -579,7 +559,7 @@ int job_reap( int interactive )
 									  j->job_id, 
 									  j->command,
 									  sig2wcs(WTERMSIG(p->status)),
-									  sig_description( WTERMSIG(p->status) ) );
+									  signal_get_desc( WTERMSIG(p->status) ) );
 						else
 							fwprintf( stdout,
 									  _( L"%ls: Process %d, \'%ls\' from job %d, \'%ls\' terminated by signal %ls (%ls)" ),
@@ -589,7 +569,7 @@ int job_reap( int interactive )
 									  j->job_id,
 									  j->command,
 									  sig2wcs(WTERMSIG(p->status)),
-									  sig_description( WTERMSIG(p->status) ) );
+									  signal_get_desc( WTERMSIG(p->status) ) );
 						tputs(clr_eol,1,&writeb);
 						fwprintf (stdout, L"\n" );
 						found=1;						
@@ -646,13 +626,18 @@ int job_reap( int interactive )
 
 
 #ifdef HAVE__PROC_SELF_STAT
+
+/**
+   Maximum length of a /proc/[PID]/stat filename
+*/
+#define FN_SIZE 256
+
 /**
    Get the CPU time for the specified process
 */
 unsigned long proc_get_jiffies( process_t *p )
 {
-	wchar_t fn[256];
-	//char stat_line[1024];
+	wchar_t fn[FN_SIZE];
 
 	char state;
 	int pid, ppid, pgrp, 
@@ -674,7 +659,7 @@ unsigned long proc_get_jiffies( process_t *p )
 	if( p->pid <= 0 )
 		return 0;
 	
-	swprintf( fn, 512, L"/proc/%d/stat", p->pid );
+	swprintf( fn, FN_SIZE, L"/proc/%d/stat", p->pid );
 	
 	FILE *f = wfopen( fn, "r" );
 	if( !f )
@@ -722,6 +707,10 @@ unsigned long proc_get_jiffies( process_t *p )
 	{
 		return 0;
 	}
+
+	/*
+	  Don't need to check exit status of fclose on read-only streams
+	*/
 	fclose( f );
 	return utime+stime+cutime+cstime;
 	
@@ -840,28 +829,11 @@ static void read_try( job_t *j )
 			else
 			{
 				b_append( buff->param2.out_buffer, b, l );
-			}
-			
+			}			
 		}
 	}
 }
 
-/**
-   Test if a specified job contains external commands
-
-   \param j the job to test
-*/
-static int job_is_external( job_t *j )
-{
-	process_t *p;
-	for( p=j->first_process; p; p=p->next )
-	{
-		if( p->type == EXTERNAL )
-			return 1;
-	}
-	return 0;
-}
-
 
 void job_continue (job_t *j, int cont)
 {
@@ -872,47 +844,43 @@ void job_continue (job_t *j, int cont)
 	j->next = first_job;
 	first_job = j;
 	j->notified = 0;
-	
-	debug( 3,
-		   L"Continue on job %d (%ls), %ls, %ls",
+
+	debug( 4,
+		   L"Continue job %d (%ls), %ls, %ls",
 		   j->job_id, 
 		   j->command, 
 		   job_is_completed( j )?L"COMPLETED":L"UNCOMPLETED", 
 		   is_interactive?L"INTERACTIVE":L"NON-INTERACTIVE" );
-
+	
 	if( !job_is_completed( j ) )
 	{
-		if( is_interactive  && job_is_external( j ) )
-		{
-							
+		if( j->terminal && j->fg )
+		{							
 			/* Put the job into the foreground.  */
-			if(  j->fg )
+			signal_block();
+			if( tcsetpgrp (0, j->pgid) )
 			{
-				signal_block();
-				if( tcsetpgrp (0, j->pgid) )
-				{
-					debug( 1, 
-						   _( L"Could not send job %d ('%ls') to foreground" ), 
-						   j->job_id, 
-						   j->command );
-					wperror( L"tcsetpgrp" );
-					return;
-				}
-				
-				if( cont )
-				{  
-					if( tcsetattr (0, TCSADRAIN, &j->tmodes))
-					{
-						debug( 1,
-							   _( L"Could not send job %d ('%ls') to foreground" ),
-							   j->job_id,
-							   j->command );
-						wperror( L"tcsetattr" );
-						return;
-					}										
-				}
-				signal_unblock();
+				debug( 1, 
+					   _( L"Could not send job %d ('%ls') to foreground" ), 
+					   j->job_id, 
+					   j->command );
+				wperror( L"tcsetpgrp" );
+				return;
 			}
+			
+			if( cont )
+			{  
+				if( tcsetattr (0, TCSADRAIN, &j->tmodes))
+				{
+					debug( 1,
+						   _( L"Could not send job %d ('%ls') to foreground" ),
+						   j->job_id,
+						   j->command );
+					wperror( L"tcsetattr" );
+					return;
+				}										
+			}
+			signal_unblock();		
 		}
 
 		/* 
@@ -921,15 +889,28 @@ void job_continue (job_t *j, int cont)
 		if( cont )
 		{
 			process_t *p;
+
 			for( p=j->first_process; p; p=p->next )
 				p->stopped=0;
-			for( p=j->first_process; p; p=p->next )
+
+			if( j->job_control )
 			{
-				if (kill ( p->pid, SIGCONT) < 0)
+				if( killpg( j->pgid, SIGCONT ) )
 				{
-					wperror (L"kill (SIGCONT)");
+					wperror( L"killpg (SIGCONT)" );
 					return;
-				}		
+				}
+			}
+			else
+			{
+				for( p=j->first_process; p; p=p->next )
+				{
+					if (kill ( p->pid, SIGCONT) < 0)
+					{
+						wperror (L"kill (SIGCONT)");
+						return;
+					}		
+				}
 			}
 		}
 	
@@ -937,6 +918,7 @@ void job_continue (job_t *j, int cont)
 		{
 			int quit = 0;
 		
+//			debug( 1, L"wait loop" );	
 			/* 
 			   Wait for job to report. Looks a bit ugly because it has to
 			   handle the possibility that a signal is dispatched while
@@ -947,13 +929,13 @@ void job_continue (job_t *j, int cont)
 				do
 				{
 					got_signal = 0;
-					quit = job_is_stopped( j ) || job_last_is_completed( j );
+					quit = job_is_stopped( j ) || job_is_completed( j );
 				}
 				while( got_signal && !quit );
 				if( !quit )
 				{
 					
-					debug( 3, L"select_try()" );	
+//					debug( 1, L"select_try()" );	
 					switch( select_try(j) )
 					{
 						case 1:			
@@ -973,6 +955,7 @@ void job_continue (job_t *j, int cont)
 							  short-lived jobs.
 							*/
 							int status;						
+//							debug( 1, L"waitpid" );	
 							pid_t pid = waitpid(-1, &status, WUNTRACED );
 							if( pid > 0 )
 								handle_child_status( pid, status );
@@ -984,7 +967,7 @@ void job_continue (job_t *j, int cont)
 			}
 		}	
 	}
-			
+	
 	if( j->fg )
 	{
 		
@@ -1010,7 +993,7 @@ void job_continue (job_t *j, int cont)
 		/* 
 		   Put the shell back in the foreground.  
 		*/
-		if( is_interactive  && job_is_external( j ) )
+		if( j->terminal && j->fg )
 		{
 			signal_block();
 			if( tcsetpgrp (0, getpid()) )
@@ -1042,6 +1025,7 @@ void job_continue (job_t *j, int cont)
 			signal_unblock();
 		}
 	}
+	
 }
 
 void proc_sanity_check()
@@ -1070,7 +1054,7 @@ void proc_sanity_check()
 		/*
 		  More than one foreground job?
 		*/
-		if( j->fg && !(job_is_stopped(j) || job_last_is_completed(j) ) )
+		if( j->fg && !(job_is_stopped(j) || job_is_completed(j) ) )
 		{
 			if( fg_job != 0 )
 			{
@@ -1117,3 +1101,19 @@ void proc_sanity_check()
 	}	
 }
 
+void proc_push_interactive( int value )
+{
+	int old = is_interactive;
+	al_push_long( interactive_stack, (long)is_interactive );
+	is_interactive = value;
+	if( old != value )
+		signal_set_handlers();
+}
+
+void proc_pop_interactive()
+{
+	int old = is_interactive;
+	is_interactive= (int)al_pop_long(interactive_stack);
+	if( is_interactive != old )
+		signal_set_handlers();
+}
