@@ -14,6 +14,7 @@
 #include <wchar.h>
 #include <signal.h>
 #include <unistd.h>
+#include <sys/time.h>
 
 #include "util.h"
 #include "io.h"
@@ -44,9 +45,21 @@ enum
 	   The exec builtin
 	*/
 	INTERNAL_EXEC,
+	/**
+	   A buffer
+	*/
+	INTERNAL_BUFFER,
+      
 }
 	;
 
+enum
+{
+	JOB_CONTROL_ALL, 
+	JOB_CONTROL_INTERACTIVE,
+	JOB_CONTROL_NONE,
+}
+	;
 
 /** 
 	A structure representing a single fish process. Contains variables
@@ -58,7 +71,7 @@ enum
 	be the result of an exec command. The role of this process_t is
 	determined by the type field, which can be one of EXTERNAL,
 	INTERNAL_BUILTIN, INTERNAL_FUNCTION, INTERNAL_BLOCK and
-	INTERNAL_EXEC.
+	INTERNAL_EXEC, INTERNAL_BUFFER
 
 	The process_t contains information on how the process should be
 	started, such as command name and arguments, as well as runtime
@@ -71,13 +84,13 @@ enum
 	argument array and actual_cmd is the absolute path of the command
 	to execute.
 
-	If the process is of type ITERNAL_BUILTIN, argv is the argument
+	If the process is of type INTERNAL_BUILTIN, argv is the argument
 	vector, and argv[0] is the name of the builtin command.
 
-	If the process is of type ITERNAL_FUNCTION, argv is the argument
+	If the process is of type INTERNAL_FUNCTION, argv is the argument
 	vector, and argv[0] is the name of the shellscript function.
 
-	If the process is of type ITERNAL_BLOCK, argv has exactly one
+	If the process is of type INTERNAL_BLOCK, argv has exactly one
 	element, which is the block of commands to execute.
 
 */
@@ -85,8 +98,8 @@ typedef struct process
 {
 	/** 
 		Type of process. Can be one of \c EXTERNAL, \c
-		INTERNAL_BUILTIN, \c INTERNAL_FUNCTION, \c INTERNAL_BLOCK or
-		INTERNAL_EXEC
+		INTERNAL_BUILTIN, \c INTERNAL_FUNCTION, \c INTERNAL_BLOCK,
+		INTERNAL_EXEC, or INTERNAL_BUFFER
 	*/
 	int type;
 	/** argv parameter for for execv, builtin_run, etc. */
@@ -154,7 +167,13 @@ typedef struct job
 	
 	/** Skip executing this job. This flag is set by the short-circut builtins, i.e. and and or */
 	int skip;
-	
+
+	/** Whether the job is under job control */
+	int job_control;
+			
+	/** Whether the job wants to own the terminal when in the foreground */
+	int terminal;
+			
 	/** Pointer to the next job */
 	struct job *next;           
 } 
@@ -207,8 +226,22 @@ extern job_t *first_job;
 */
 extern int proc_had_barrier;
 
+/**
+   Pid of last process to be started in the background
+*/
 extern pid_t proc_last_bg_pid;
 
+/**
+   Can be one of JOB_CONTROL_ALL, JOB_CONTROL_INTERACTIVE and JOB_CONTROL_NONE
+*/
+extern int job_control_mode;
+
+/**
+   If this flag is set, fish will never fork or run execve.
+*/
+extern int no_exec;
+
+
 /**
    Sets the status of the last process to exit
 */
@@ -225,13 +258,15 @@ int proc_get_last_status();
 void job_free( job_t* j );
 
 /**
-   Create a new job
+   Create a new job. Job struct is allocated using halloc, so anything
+   that should be freed with the job can uset it as a halloc context
+   when allocating.
 */
 job_t *job_create();
 
 /**
   Return the job with the specified job id.
-  If id is -1, return the last job used.
+  If id is 0 or less, return the last job used.
 */
 job_t *job_get(int id);
 
@@ -246,7 +281,7 @@ job_t *job_get_from_pid(int pid);
 int job_is_stopped( const job_t *j );
 
 /**
-   Tests if the job has completed
+   Tests if the job has completed, i.e. if the last process of the pipeline has ended.
  */
 int job_is_completed( const job_t *j );
 
@@ -261,20 +296,19 @@ int job_is_completed( const job_t *j );
 void job_continue( job_t *j, int cont );
 
 /**
-   Notify user of nog events.  Notify the user about stopped or
-   terminated jobs.  Delete terminated jobs from the active job list.
+   Notify the user about stopped or terminated jobs. Delete terminated
+   jobs from the job list.
 
    \param interactive whether interactive jobs should be reaped as well
 */
 int job_reap( int interactive );
 
 /**
-   Signal handler for SIGCHLD.  Mark any processes with relevant
+   Signal handler for SIGCHLD. Mark any processes with relevant
    information.
 */
 void job_handle_signal( int signal, siginfo_t *info, void *con );
 
-
 #ifdef HAVE__PROC_SELF_STAT
 /**
    Use the procfs filesystem to look up how many jiffies of cpu time
@@ -299,11 +333,12 @@ void proc_update_jiffies();
 void proc_sanity_check();
 
 /**
-   Send of an process/job exit event notification. This function is a conveniance wrapper around event_fire().
+   Send a process/job exit event notification. This function is a
+   conveniance wrapper around event_fire().
 */
 void proc_fire_event( const wchar_t *msg, int type, pid_t pid, int status );
 
-/*
+/**
   Initializations
 */
 void proc_init();
@@ -313,4 +348,16 @@ void proc_init();
 */
 void proc_destroy();
 
+/**
+   Set new value for is_interactive flag, saving previous value. If
+   needed, update signal handlers.
+*/
+void proc_push_interactive( int value );
+
+/**
+   Set is_interactive flag to the previous value. If needed, update
+   signal handlers.
+*/
+void proc_pop_interactive();
+
 #endif
