@@ -21,7 +21,7 @@ int reader_read( int fd);
 /**
   Tell the shell that it should exit after the currently running command finishes.
 */
-void reader_exit( int do_exit );
+void reader_exit( int do_exit, int force );
 
 /**
    Check that the reader is in a sane state
@@ -48,7 +48,7 @@ wchar_t *reader_current_filename();
    
    \param fn The fileanme to push
 */
-void reader_push_current_filename( wchar_t *fn );
+void reader_push_current_filename( const wchar_t *fn );
 /**
    Pop the current filename from the stack of read files
  */
@@ -72,7 +72,7 @@ void repaint();
    Run the specified command with the correct terminal modes, and
    while taking care to perform job notification, set the title, etc.
 */
-void reader_run_command( wchar_t *buff );
+void reader_run_command( const wchar_t *buff );
 
 /**
    Get the string of character currently entered into the command
@@ -152,27 +152,6 @@ void reader_set_prompt( wchar_t *prompt );
 */
 int exit_status();
 
-/**
-   Find the beginning and the end of the current subshell
-*/
-void reader_current_subshell_extent( wchar_t **a, wchar_t **b );
-
-/**
-   Find the beginning and the end of the job under the cursor 
-*/
-void reader_current_job_extent( wchar_t **a, wchar_t **b );
-
-/**
-   Find the beginning and the end of the process under the cursor 
-*/
-void reader_current_process_extent( wchar_t **a, wchar_t **b );
-
-/**
-   Find the beginning and the end of the token under the curor and the token before the cursor
-*/
-
-void reader_current_token_extent( wchar_t **a, wchar_t **b, wchar_t **pa, wchar_t **pb );
-
 /**
    Replace the current token with the specified string
 */
@@ -183,5 +162,10 @@ void reader_replace_current_token( wchar_t *new_token );
 */
 void reader_handle_int( int signal );
 
+/**
+   This function returns true if fish is exiting by force, i.e. because stdin died
+*/
+int reader_exit_forced();
+
 
 #endif
