@@ -15,29 +15,30 @@
 #include <string.h>
 #include <unistd.h>
 
+
+#include "fallback.h"
 #include "util.h"
+
 #include "wutil.h"
 #include "tokenizer.h"
 #include "common.h"
 #include "wildcard.h"
-#include "translate.h"
+
 
 /**
    Error string for unexpected end of string
 */
 #define EOL_ERROR _( L"Unexpected end of token" )
+
 /**
    Error string for mismatched parenthesis
 */
 #define PARAN_ERROR _( L"Parenthesis mismatch" )
+
 /**
    Error string for invalid redirections
 */
 #define REDIRECT_ERROR _( L"Invalid redirection" )
-/**
-   Error string for invalid input
-*/
-#define INPUT_ERROR _( L"Invalid input" )
 
 /**
    Error string for when trying to pipe from fd 0
@@ -54,6 +55,9 @@
 */
 #define FD_STR_MAX_LEN 16
 
+/**
+   Descriptions of all tokenizer errors
+*/
 static const wchar_t *tok_desc[] =
 {
 	N_(L"Tokenizer not yet initialized"),
@@ -112,29 +116,21 @@ static void tok_error( tokenizer *tok, const wchar_t *err )
 
 void tok_init( tokenizer *tok, const wchar_t *b, int flags )
 {
-//	fwprintf( stderr, L"CREATE: \'%ls\'\n", b );
-		
+
+	CHECK( tok, );
 
 	memset( tok, 0, sizeof( tokenizer) );
-	
+
+	CHECK( b, );
+
+
 	tok->accept_unfinished = flags & TOK_ACCEPT_UNFINISHED;
 	tok->show_comments = flags & TOK_SHOW_COMMENTS;
-	tok->has_next=1;	
-
-	/* 
-	   Before we copy the buffer we need to check that it is not
-	   null. But before that, we need to init the tokenizer far enough
-	   so that errors can be properly flagged
-	*/
-	if( !b )
-	{
-		tok_error( tok, INPUT_ERROR );
-		return;
-	}
+	tok->has_next=1;
 
 	tok->has_next = (*b != L'\0');
 	tok->orig_buff = tok->buff = (wchar_t *)(b);
-	
+
 	if( tok->accept_unfinished )
 	{
 		int l = wcslen( tok->orig_buff );
@@ -146,7 +142,7 @@ void tok_init( tokenizer *tok, const wchar_t *b, int flags )
 				tok->orig_buff = tok->buff = wcsdup( tok->orig_buff );
 				if( !tok->orig_buff )
 				{
-					die_mem();		
+					DIE_MEM();
 				}
 				tok->orig_buff[l-1] = L'\0';
 			}
@@ -159,6 +155,8 @@ void tok_init( tokenizer *tok, const wchar_t *b, int flags )
 
 void tok_destroy( tokenizer *tok )
 {
+	CHECK( tok, );
+	
 	free( tok->last );
 	if( tok->free_orig )
 		free( tok->orig_buff );
@@ -166,16 +164,27 @@ void tok_destroy( tokenizer *tok )
 
 int tok_last_type( tokenizer *tok )
 {
+	CHECK( tok, TOK_ERROR );
+	CHECK( tok->buff, TOK_ERROR );
+	
 	return tok->last_type;
 }
 
 wchar_t *tok_last( tokenizer *tok )
 {
+	CHECK( tok, 0 );
+	
 	return tok->last;
 }
 
 int tok_has_next( tokenizer *tok )
 {
+	/*
+	  Return 1 on broken tokenizer
+	*/
+	CHECK( tok, 1 );
+	CHECK( tok->buff, 1 );
+
 /*	fwprintf( stderr, L"has_next is %ls \n", tok->has_next?L"true":L"false" );*/
 	return 	tok->has_next;
 }
@@ -220,7 +229,7 @@ static void read_string( tokenizer *tok )
 
 	while( 1 )
 	{
-		
+
 		if( !myal( *tok->buff ) )
 		{
 //			debug(1, L"%lc", *tok->buff );
@@ -257,7 +266,7 @@ static void read_string( tokenizer *tok )
 						mode = 1;
 						break;
 					}
-					
+
 					case L'[':
 					{
 						if( tok->buff != start )
@@ -268,24 +277,24 @@ static void read_string( tokenizer *tok )
 					case L'\'':
 					case L'"':
 					{
-						
-						wchar_t *end = quote_end( tok->buff );
+
+						const wchar_t *end = quote_end( tok->buff );
 						tok->last_quote = *tok->buff;
 						if( end )
 						{
-							tok->buff=end;
+							tok->buff=(wchar_t *)end;
 						}
 						else
 						{
 							tok->buff += wcslen( tok->buff );
-							
+
 							if( (!tok->accept_unfinished) )
 							{
 								tok_error( tok, EOL_ERROR );
 								return;
 							}
 							do_loop = 0;
-							
+
 						}
 						break;
 					}
@@ -300,7 +309,7 @@ static void read_string( tokenizer *tok )
 				}
 				break;
 			}
-		
+
 			case 3:
 			case 1:
 				switch( *tok->buff )
@@ -308,10 +317,10 @@ static void read_string( tokenizer *tok )
 					case L'\'':
 					case L'\"':
 					{
-						wchar_t *end = quote_end( tok->buff );
+						const wchar_t *end = quote_end( tok->buff );
 						if( end )
 						{
-							tok->buff=end;
+							tok->buff=(wchar_t *)end;
 						}
 						else
 							do_loop = 0;
@@ -352,7 +361,7 @@ static void read_string( tokenizer *tok )
 				break;
 		}
 		}
-		
+
 
 		if( !do_loop )
 			break;
@@ -400,12 +409,12 @@ static void read_comment( tokenizer *tok )
 }
 
 /**
-   Read a FD redirect.
+   Read a FD redirection.
 */
 static void read_redirect( tokenizer *tok, int fd )
 {
 	int mode = -1;
-	
+
 	if( (*tok->buff == L'>') ||
 		(*tok->buff == L'^') )
 	{
@@ -427,7 +436,7 @@ static void read_redirect( tokenizer *tok, int fd )
 				tok_error( tok, PIPE_ERROR );
 				return;
 			}
-			check_size( tok, FD_STR_MAX_LEN );						
+			check_size( tok, FD_STR_MAX_LEN );
 			tok->buff++;
 			swprintf( tok->last, FD_STR_MAX_LEN, L"%d", fd );
 			tok->last_type = TOK_PIPE;
@@ -443,14 +452,14 @@ static void read_redirect( tokenizer *tok, int fd )
 	{
 		tok_error( tok, REDIRECT_ERROR);
 	}
-	
+
 	if( !check_size( tok, 2 ))
 	{
 		return;
 	}
-	
+
 	swprintf( tok->last, tok->last_len, L"%d", fd );
-	
+
 	if( *tok->buff == L'&' )
 	{
 		tok->buff++;
@@ -464,6 +473,8 @@ static void read_redirect( tokenizer *tok, int fd )
 
 wchar_t tok_last_quote( tokenizer *tok )
 {
+	CHECK( tok, 0 );
+	
 	return tok->last_quote;
 }
 
@@ -482,21 +493,26 @@ static int my_iswspace( wchar_t c )
 
 const wchar_t *tok_get_desc( int type )
 {
-
+	if( type < 0 || type >= sizeof( tok_desc ) )
+	{
+		return _(L"Invalid token type");
+	}
 	return _(tok_desc[type]);
 }
 
 
 void tok_next( tokenizer *tok )
 {
-//	fwprintf( stderr, L"tok_next on %ls (prev=%ls)\n", tok->orig_buff, tok_desc[tok->last_type] );
 
+	CHECK( tok, );
+	CHECK( tok->buff, );
+	
 	if( tok_last_type( tok ) == TOK_ERROR )
 	{
-		tok->has_next=0;	
-		return;	
+		tok->has_next=0;
+		return;
 	}
-	
+
 	if( !tok->has_next )
 	{
 /*		wprintf( L"EOL\n" );*/
@@ -504,8 +520,22 @@ void tok_next( tokenizer *tok )
 		return;
 	}
 
-	while( my_iswspace(*(tok->buff) ) )
-		tok->buff++;
+	while( 1 )
+	{
+		if( my_iswspace(*(tok->buff) ) )
+		{
+			tok->buff++;
+		}
+		else
+		{
+			if(( *(tok->buff) == L'\\') &&( *(tok->buff+1) == L'\n') )
+			{
+				tok->buff+=2;
+			}
+			break;
+		}
+	}
+	
 
 	if( *tok->buff == L'#')
 	{
@@ -548,9 +578,9 @@ void tok_next( tokenizer *tok )
 
 		case L'|':
 			check_size( tok, 2 );
-			
+
 			tok->last[0]=L'1';
-			tok->last[1]=L'\0';			
+			tok->last[1]=L'\0';
 			tok->last_type = TOK_PIPE;
 			tok->buff++;
 			break;
@@ -564,9 +594,9 @@ void tok_next( tokenizer *tok )
 
 		default:
 		{
-			
+
 			if( iswdigit( *tok->buff ) )
-			{				
+			{
 				wchar_t *orig = tok->buff;
 				int fd = 0;
 				while( iswdigit( *tok->buff ) )
@@ -580,18 +610,18 @@ void tok_next( tokenizer *tok )
 						read_redirect( tok, fd );
 						return;
 				}
-				tok->buff = orig;				
+				tok->buff = orig;
 			}
 			read_string( tok );
 		}
-		
+
 	}
 
 }
 
 wchar_t *tok_string( tokenizer *tok )
 {
-	return tok->orig_buff;
+	return tok?tok->orig_buff:0;
 }
 
 wchar_t *tok_first( const wchar_t *str )
@@ -599,8 +629,10 @@ wchar_t *tok_first( const wchar_t *str )
 	tokenizer t;
 	wchar_t *res=0;
 
-	tok_init( &t, str, 0 );
+	CHECK( str, 0 );
 	
+	tok_init( &t, str, 0 );
+
 	switch( tok_last_type( &t ) )
 	{
 		case TOK_STRING:
@@ -618,12 +650,16 @@ wchar_t *tok_first( const wchar_t *str )
 
 int tok_get_pos( tokenizer *tok )
 {
+	CHECK( tok, 0 );
+	
 	return tok->last_pos + (tok->free_orig?1:0);
 }
 
 
 void tok_set_pos( tokenizer *tok, int pos )
 {
+	CHECK( tok, );
+	
 	tok->buff = tok->orig_buff + pos;
 	tok->has_next = 1;
 	tok_next( tok );
