@@ -6,6 +6,7 @@
 
 #include "config.h"
 
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <wchar.h>
@@ -22,7 +23,9 @@
 #include <errno.h>
 
 
+#include "fallback.h"
 #include "util.h"
+
 #include "common.h"
 #include "wutil.h"
 
@@ -47,18 +50,6 @@
 */
 #define SB_MAX_SIZE 32767
 
-float minf( float a,
-			float b )
-{
-	return a<b?a:b;
-}
-
-
-float maxf( float a,
-			float b )
-{
-	return a>b?a:b;
-}
 
 int mini( int a,
 		  int b )
@@ -168,16 +159,12 @@ int q_empty( dyn_queue_t *q )
 	return q->put_pos == q->get_pos;
 }
 
-/* Stack functions */
-
-
-
 
 /* Hash table functions */
 
 void hash_init2( hash_table_t *h,
-				int (*hash_func)(const void *key),
-				 int (*compare_func)(const void *key1, const void *key2),
+				int (*hash_func)(void *key),
+				 int (*compare_func)(void *key1, void *key2),
 				 size_t capacity)
 {
 	int i;
@@ -193,8 +180,8 @@ void hash_init2( hash_table_t *h,
 }
 
 void hash_init( hash_table_t *h,
-				int (*hash_func)(const void *key),
-				int (*compare_func)(const void *key1, const void *key2) )
+				int (*hash_func)(void *key),
+				int (*compare_func)(void *key1, void *key2) )
 {
 	h->arr = 0;
 	h->size = 0;
@@ -214,7 +201,7 @@ void hash_destroy( hash_table_t *h )
    \return index in the table, or to the first free index if the key is not in the table
 */
 static int hash_search( hash_table_t *h,
-						const void *key )
+						void *key )
 {
 	int hv;
 	int pos;
@@ -291,38 +278,43 @@ int hash_put( hash_table_t *h,
 		}
 	}
 
-	pos = hash_search( h, key );
+	pos = hash_search( h, (void *)key );
 
 	if( h->arr[pos].key == 0 )
 	{
 		h->count++;
 	}
 
-	h->arr[pos].key = key;
-	h->arr[pos].data = data;
+	h->arr[pos].key = (void *)key;
+	h->arr[pos].data = (void *)data;
 	return 1;
 }
 
-const void *hash_get( hash_table_t *h,
-					  const void *key )
+void *hash_get( hash_table_t *h,
+				const void *key )
 {
 	if( !h->count )
 		return 0;
 	
-	int pos = hash_search( h, key );	
+	int pos = hash_search( h, (void *)key );	
 	if( h->arr[pos].key == 0 )
+	{
 		return 0;
+	}
 	else
-		return h->arr[pos].data;
+	{
+		void *res =h->arr[pos].data;
+		return res;
+	}
 }
 
-const void *hash_get_key( hash_table_t *h,
-						  const void *key )
+void *hash_get_key( hash_table_t *h,
+					const void *key )
 {	
 	if( !h->count )
 		return 0;
 	
-	int pos = hash_search( h, key );
+	int pos = hash_search( h, (void *)key );
 	if( h->arr[pos].key == 0 )
 		return 0;
 	else
@@ -336,8 +328,8 @@ int hash_get_count( hash_table_t *h)
 
 void hash_remove( hash_table_t *h,
 				  const void *key,
-				  const void **old_key,
-				  const void **old_val )
+				  void **old_key,
+				  void **old_val )
 {
 	if( !h->count )
 	{
@@ -349,7 +341,7 @@ void hash_remove( hash_table_t *h,
 		return;
 	}
 
-	int pos = hash_search( h, key );
+	int pos = hash_search( h, (void *)key );
 	int next_pos;
 
 	if( h->arr[pos].key == 0 )
@@ -408,15 +400,15 @@ int hash_contains( hash_table_t *h,
 	if( !h->count )
 		return 0;
 	
-	int pos = hash_search( h, key );
+	int pos = hash_search( h, (void *)key );
 	return h->arr[pos].key != 0;
 }
 
 /**
    Push hash value into array_list_t
 */
-static void hash_put_data( const void *key,
-						   const void *data,
+static void hash_put_data( void *key,
+						   void *data,
 						   void *al )
 {
 	al_push( (array_list_t *)al,
@@ -433,7 +425,7 @@ void hash_get_data( hash_table_t *h,
 /**
    Push hash key into array_list_t
 */
-static void hash_put_key( const void *key, const void *data, void *al )
+static void hash_put_key( void *key, void *data, void *al )
 {
 	al_push( (array_list_t *)al, key );
 }
@@ -446,7 +438,7 @@ void hash_get_keys( hash_table_t *h,
 }
 
 void hash_foreach( hash_table_t *h,
-				   void (*func)(const void *, const void *) )
+				   void (*func)( void *, void *) )
 {
 	int i;
 	for( i=0; i<h->size; i++ )
@@ -459,7 +451,7 @@ void hash_foreach( hash_table_t *h,
 }
 
 void hash_foreach2( hash_table_t *h,
-					void (*func)( const void *, const void *, void * ),
+					void (*func)( void *, void *, void * ),
 					void *aux )
 {
 	int i;
@@ -473,7 +465,7 @@ void hash_foreach2( hash_table_t *h,
 }
 
 
-int hash_str_cmp( const void *a, const void *b )
+int hash_str_cmp( void *a, void *b )
 {
 	return strcmp((char *)a,(char *)b) == 0;
 }
@@ -481,13 +473,13 @@ int hash_str_cmp( const void *a, const void *b )
 /**
    Helper function for hash_wcs_func
 */
-static uint rotl5( uint in )
+static unsigned long rotl5( unsigned long in )
 {
-	return (in<<5|in>>27);
+	return ((in<<5|in>>27))&0xffffffff;
 }
 
 
-int hash_str_func( const void *data )
+int hash_str_func( void *data )
 {
 	int res = 0x67452301u;
 	const char *str = data;	
@@ -498,7 +490,7 @@ int hash_str_func( const void *data )
 	return res;
 }
 
-int hash_wcs_func( const void *data )
+int hash_wcs_func( void *data )
 {
 	int res = 0x67452301u;
 	const wchar_t *str = data;	
@@ -510,11 +502,25 @@ int hash_wcs_func( const void *data )
 }
 
 
-int hash_wcs_cmp( const void *a, const void *b )
+int hash_wcs_cmp( void *a, void *b )
 {
 	return wcscmp((wchar_t *)a,(wchar_t *)b) == 0;
 }
 
+int hash_ptr_func( void *data )
+{
+	return (int)(long) data;
+}
+
+/**
+   Hash comparison function suitable for direct pointer comparison
+*/
+int hash_ptr_cmp( void *a,
+                  void *b )
+{
+	return a == b;
+}
+
 void pq_init( priority_queue_t *q,
 			  int (*compare)(void *e1, void *e2) )
 {
@@ -651,7 +657,7 @@ array_list_t *al_new()
 {
 	array_list_t *res = malloc( sizeof( array_list_t ) );
 	if( !res )
-		die_mem();
+		DIE_MEM();
 	al_init( res );
 	return res;
 }
@@ -667,12 +673,12 @@ void al_destroy( array_list_t *l )
 	free( l->arr );
 }
 
-int al_push( array_list_t *l, const void *o )
+static int al_push_generic( array_list_t *l, anything_t o )
 {
 	if( l->pos >= l->size )
 	{
 		int new_size = l->pos == 0 ? MIN_SIZE : 2 * l->pos;
-		void *tmp = realloc( l->arr, sizeof( void *)*new_size );
+		void *tmp = realloc( l->arr, sizeof( anything_t )*new_size );
 		if( tmp == 0 )
 			return 0;
 		l->arr = tmp;
@@ -682,6 +688,28 @@ int al_push( array_list_t *l, const void *o )
 	return 1;
 }
 
+int al_push( array_list_t *l, const void *o )
+{
+	anything_t v;
+	v.ptr_val = (void *)o;
+	return al_push_generic( l, v );
+}
+
+int al_push_long( array_list_t *l, long val )
+{
+	anything_t v;
+	v.long_val = val;
+	return al_push_generic( l, v );
+}
+
+int al_push_func( array_list_t *l, func_ptr_t f )
+{
+	anything_t v;
+	v.func_val = f;
+	return al_push_generic( l, v );
+}
+
+
 int al_push_all( array_list_t *a, array_list_t *b )
 {
 	int k;
@@ -694,7 +722,7 @@ int al_push_all( array_list_t *a, array_list_t *b )
 }
 
 
-int al_set( array_list_t *l, int pos, const void *o )
+static int al_set_generic( array_list_t *l, int pos, anything_t v )
 {
 	int old_pos;
 	
@@ -702,48 +730,88 @@ int al_set( array_list_t *l, int pos, const void *o )
 		return 0;
 	if( pos < l->pos )
 	{
-		l->arr[pos] = o;
+		l->arr[pos]=v;
 		return 1;
 	}
 	old_pos=l->pos;
 	
 	l->pos = pos;
-	if( al_push( l, o ) )
+	if( al_push_generic( l, v ) )
 	{
 /*		fwprintf( stderr, L"Clearing from index %d to index %d\n", 
 				  old_pos, pos );
 */	
 		memset( &l->arr[old_pos], 
 				0,
-				sizeof(void *) * (pos - old_pos) );
+				sizeof(anything_t) * (pos - old_pos) );
 		return 1;		
 	}
 	return 0;	
 }
 
-const void *al_get( array_list_t *l, int pos )
+int al_set( array_list_t *l, int pos, const void *o )
 {
-	if( pos < 0 )
-		return 0;
-	if( pos >= l->pos )
-		return 0;
-	return l->arr[pos];
+	anything_t v;
+	v.ptr_val = (void *)o;
+	return al_set_generic( l, pos, v );
 }
 
+int al_set_long( array_list_t *l, int pos, long o )
+{
+	anything_t v;
+	v.long_val = o;
+	return al_set_generic( l, pos, v );
+}
+
+int al_set_func( array_list_t *l, int pos, func_ptr_t o )
+{
+	anything_t v;
+	v.func_val = o;
+	return al_set_generic( l, pos, v );
+}
+
+static anything_t al_get_generic( array_list_t *l, int pos )
+{
+	anything_t res;
+	res.ptr_val=0;
+	
+	if( (pos >= 0) && (pos < l->pos) )
+		res = l->arr[pos];
+
+	return res;
+}
+
+void *al_get( array_list_t *l, int pos )
+{
+	return al_get_generic(l,pos).ptr_val;
+}
+
+long al_get_long( array_list_t *l, int pos )
+{
+	return al_get_generic(l,pos).long_val;
+}
+
+func_ptr_t al_get_func( array_list_t *l, int pos )
+{
+	return al_get_generic(l,pos).func_val;
+}
+
+
+
 void al_truncate( array_list_t *l, int new_sz )
 {
 	l->pos = new_sz;
 }
 
-const void *al_pop( array_list_t *l )
+static anything_t al_pop_generic( array_list_t *l )
 {
-	const void *e = l->arr[--l->pos];
+	anything_t e = l->arr[--l->pos];
 	if( (l->pos*3 < l->size) && (l->size < MIN_SIZE) )
 	{
-		const void ** old_arr = l->arr;
+		anything_t *old_arr = l->arr;
 		int old_size = l->size;
 		l->size = l->size/2;
-		l->arr = realloc( l->arr, sizeof(void*)*l->size );
+		l->arr = realloc( l->arr, sizeof(anything_t)*l->size );
 		if( l->arr == 0 )
 		{
 			l->arr = old_arr;
@@ -753,10 +821,43 @@ const void *al_pop( array_list_t *l )
 	return e;
 }
 
-const void *al_peek( array_list_t *l )
+void *al_pop( array_list_t *l )
 {
+	return al_pop_generic(l).ptr_val;	
+}
 
-	return l->pos>0?l->arr[l->pos-1]:0;
+long al_pop_long( array_list_t *l )
+{
+	return al_pop_generic(l).long_val;	
+}
+
+func_ptr_t al_pop_func( array_list_t *l )
+{
+	return al_pop_generic(l).func_val;	
+}
+
+static anything_t al_peek_generic( array_list_t *l )
+{
+	anything_t res;
+	res.ptr_val=0;
+	if( l->pos>0)
+		res = l->arr[l->pos-1];
+	return res;
+}
+
+void *al_peek( array_list_t *l )
+{
+	return al_peek_generic(l).ptr_val;	
+}
+
+long al_peek_long( array_list_t *l )
+{
+	return al_peek_generic(l).long_val;	
+}
+
+func_ptr_t al_peek_func( array_list_t *l )
+{
+	return al_peek_generic(l).func_val;	
 }
 
 int al_empty( array_list_t *l )
@@ -770,18 +871,18 @@ int al_get_count( array_list_t *l )
 	return l->pos;
 }
 
-void al_foreach( array_list_t *l, void (*func)( const void * ))
+void al_foreach( array_list_t *l, void (*func)( void * ))
 {
 	int i;
 	for( i=0; i<l->pos; i++ )
-		func( l->arr[i] );
+		func( l->arr[i].ptr_val );
 }
 
-void al_foreach2( array_list_t *l, void (*func)( const void *, void *), void *aux)
+void al_foreach2( array_list_t *l, void (*func)( void *, void *), void *aux)
 {
 	int i;
 	for( i=0; i<l->pos; i++ )
-		func( l->arr[i], aux );
+		func( l->arr[i].ptr_val, aux );
 }
 
 int wcsfilecmp( const wchar_t *a, const wchar_t *b )
@@ -796,6 +897,7 @@ int wcsfilecmp( const wchar_t *a, const wchar_t *b )
 	{
 		return 1;
 	}
+
 	int secondary_diff=0;
 	if( iswdigit( *a ) && iswdigit( *b ) )
 	{
@@ -836,6 +938,12 @@ int wcsfilecmp( const wchar_t *a, const wchar_t *b )
 void sb_init( string_buffer_t * b)
 {
 	wchar_t c=0;
+
+	if( !b )
+	{
+		return;
+	}
+
 	memset( b, 0, sizeof(string_buffer_t) );	
 	b_append( b, &c, sizeof( wchar_t));
 	b->used -= sizeof(wchar_t);
@@ -845,7 +953,7 @@ string_buffer_t *sb_new()
 {
 	string_buffer_t *res = malloc( sizeof( string_buffer_t ) );
 	if( !res )
-		die_mem();
+		DIE_MEM();
 	sb_init( res );
 	return res;
 }
@@ -856,8 +964,14 @@ void sb_append( string_buffer_t *b, const wchar_t * s)
 //	fwprintf( stderr, L"Append string \'%ls\'\n", s );
 
 	if( !s )
+	{
 		return;
-
+	}
+	
+	if( !b )
+	{
+		return;
+	}
 
 	b_append( b, s, sizeof(wchar_t)*(wcslen(s)+1) );
 	b->used -= sizeof(wchar_t);
@@ -868,7 +982,14 @@ void sb_append_substring( string_buffer_t *b, const wchar_t *s, size_t l )
     wchar_t tmp=0;
 
 	if( !s )
+	{
 		return;
+	}
+	
+	if( !b )
+	{
+		return;
+	}
 
     b_append( b, s, sizeof(wchar_t)*l );
     b_append( b, &tmp, sizeof(wchar_t) );
@@ -878,13 +999,16 @@ void sb_append_substring( string_buffer_t *b, const wchar_t *s, size_t l )
 
 void sb_append_char( string_buffer_t *b, wchar_t c )
 {
-	wchar_t buff[2]=
-		{
-			c, 0
-		}
-	;
-	sb_append( b, buff );
+    wchar_t tmp=0;
 
+	if( !b )
+	{
+		return;
+	}
+
+	b_append( b, &c, sizeof(wchar_t) );
+    b_append( b, &tmp, sizeof(wchar_t) );
+    b->used -= sizeof(wchar_t);
 }
 
 void sb_append2( string_buffer_t *b, ... )
@@ -892,6 +1016,11 @@ void sb_append2( string_buffer_t *b, ... )
 	va_list va;
 	wchar_t *arg;
 
+	if( !b )
+	{
+		return;
+	}
+
 	va_start( va, b );
 	while( (arg=va_arg(va, wchar_t *) )!= 0 )
 	{
@@ -905,6 +1034,11 @@ int sb_printf( string_buffer_t *buffer, const wchar_t *format, ... )
 	va_list va;
 	int res;
 	
+	if( !buffer )
+	{
+		return -1;
+	}
+
 	va_start( va, format );
 	res = sb_vprintf( buffer, format, va );	
 	va_end( va );
@@ -916,12 +1050,17 @@ int sb_vprintf( string_buffer_t *buffer, const wchar_t *format, va_list va_orig
 {
 	int res;
 	
+	if( !buffer )
+	{
+		return -1;
+	}
+
 	if( !buffer->length )
 	{
 		buffer->length = MIN_SIZE;
 		buffer->buff = malloc( MIN_SIZE );
 		if( !buffer->buff )
-			die_mem();
+			DIE_MEM();
 	}
 	
 
@@ -961,7 +1100,7 @@ int sb_vprintf( string_buffer_t *buffer, const wchar_t *format, va_list va_orig
 
 		buffer->buff = realloc( buffer->buff, 2*buffer->length );
 		if( !buffer->buff )
-			die_mem();
+			DIE_MEM();
 		buffer->length *= 2;				
 	}
 	return res;	
@@ -972,6 +1111,11 @@ int sb_vprintf( string_buffer_t *buffer, const wchar_t *format, va_list va_orig
 
 void sb_destroy( string_buffer_t * b )
 {
+	if( !b )
+	{
+		return;
+	}
+
 	free( b->buff );
 }
 
@@ -1029,7 +1173,7 @@ void b_append( buffer_t *b, const void *d, ssize_t len )
 		void *d = realloc( b->buff, l );
 		if( !d )
 		{
-			die_mem();
+			DIE_MEM();
 			
 		}
 		b->buff=d;
