@@ -1,5 +1,11 @@
 /** \file util.h
 	Generic utilities library.
+
+	All containers in this library except strinb_buffer_t are written
+	so that they don't allocate any memory until the first element is
+	inserted into them. That way it is known to be very cheap to
+	initialize various containers at startup, supporting the fish
+	notion of doing as much lazy initalization as possible.
 */
 
 #ifndef FISH_UTIL_H
@@ -7,6 +13,18 @@
 
 #include <wchar.h>
 #include <stdarg.h>
+#include <unistd.h>
+
+typedef void (*func_ptr_t)();
+
+typedef union 
+{
+	long long_val;
+	void *ptr_val;
+	func_ptr_t func_val;
+}
+	anything_t;
+
 
 /**
    Data structure for an automatically resizing dynamically allocated queue,
@@ -22,7 +40,7 @@ typedef struct dyn_queue
 	/** Where to remove elements */
 	void **get_pos;
 }
-dyn_queue_t;
+	dyn_queue_t;
 
 /**
    Internal struct used by hash_table_t.
@@ -30,11 +48,11 @@ dyn_queue_t;
 typedef struct
 {
 	/** Hash key*/
-	const void *key;
+	void *key;
 	/** Value */
-	const void *data;
+	void *data;
 }
-hash_struct_t;
+	hash_struct_t;
 
 /**
    Data structure for the hash table implementaion. A hash table allows for
@@ -60,11 +78,11 @@ typedef struct hash_table
 	/** Length of array */
 	int size;
 	/** Hash function */
-	int (*hash_func)( const void *key );
+	int (*hash_func)( void *key );
 	/** Comparison function */
-	int (*compare_func)( const void *key1, const void *key2 );
+	int (*compare_func)( void *key1, void *key2 );
 }
-hash_table_t;
+	hash_table_t;
 
 /**
    Data structure for an automatically resizing dynamically allocated
@@ -92,7 +110,7 @@ priority_queue_t;
 typedef struct array_list
 {
 	/** Array containing the data */
-	const void **arr;
+	anything_t *arr;
 	/** Position to append elements at*/
 	int pos;
 	/** Length of array */
@@ -142,16 +160,6 @@ int maxi( int a, int b );
  */
 int mini( int a, int b );
 
-/**
-   Returns the larger of two floats
-*/
-float maxf( float a, float b );
-
-/**
-   Returns the smaller of two floats
-*/
-float minf( float a, float b );
-
 /*
   All the datastuctures below autoresize. The queue, stack and
   priority queue are all impemented using an array and are guaranteed
@@ -165,34 +173,44 @@ float minf( float a, float b );
 */
 void q_init( dyn_queue_t *q );
 
-/** Destroy the queue */
+/**
+   Destroy the queue 
+*/
 void q_destroy( dyn_queue_t *q );
 
-/** Insert element into queue */
+/**
+   Insert element into queue 
+*/
 int q_put( dyn_queue_t *q, void *e );
 
-/** Remove and return next element from queue */
+/**
+   Remove and return next element from queue 
+*/
 void *q_get( dyn_queue_t *q);
 
-/** Return next element from queue without removing it */
+/**
+   Return next element from queue without removing it 
+*/
 void *q_peek( dyn_queue_t *q);
 
-/** Returns 1 if the queue is empty, 0 otherwise */
+/**
+   Returns 1 if the queue is empty, 0 otherwise 
+*/
 int q_empty( dyn_queue_t *q );
 
 /**
    Initialize a hash table. The hash function must never return the value 0.
 */
 void hash_init( hash_table_t *h,
-				int (*hash_func)(const void *key),
-				int (*compare_func)(const void *key1, const void *key2) );
+				int (*hash_func)( void *key),
+				int (*compare_func)( void *key1, void *key2 ) );
 
 /**
    Initialize a hash table. The hash function must never return the value 0.
 */
 void hash_init2( hash_table_t *h,
-				int (*hash_func)(const void *key),
-				 int (*compare_func)(const void *key1, const void *key2),
+				int (*hash_func)( void *key ),
+				 int (*compare_func)( void *key1, void *key2 ),
 				 size_t capacity);
 
 /**
@@ -208,20 +226,20 @@ int hash_put( hash_table_t *h,
 /**
    Returns the data with the associated key, or 0 if no such key is in the hashtable
 */
-const void *hash_get( hash_table_t *h,
-					  const void *key );
+void *hash_get( hash_table_t *h,
+				const void *key );
 /**
    Returns the hash tables version of the specified key
 */
-const void *hash_get_key( hash_table_t *h, 
-						  const void *key );
+void *hash_get_key( hash_table_t *h, 
+					const void *key );
 
 /**
    Returns the number of key/data pairs in the table.
 */
 int hash_get_count( hash_table_t *h);
 /**
-   Remove the specified key from the hash table
+   Remove the specified key from the hash table if it exists. Do nothing if it does not exist.
 
    \param h The hashtable
    \param key The key
@@ -230,8 +248,8 @@ int hash_get_count( hash_table_t *h);
 */
 void hash_remove( hash_table_t *h, 
 				  const void *key, 
-				  const void **old_key,
-				  const void **old_data );
+				  void **old_key,
+				  void **old_data );
 
 /**
    Checks whether the specified key is in the hash table
@@ -251,39 +269,55 @@ void hash_get_keys( hash_table_t *h,
 void hash_get_data( hash_table_t *h,
 					array_list_t *arr );
 
-/** 
-	Call the function func for each key/data pair in the table
+/**
+   Call the function func for each key/data pair in the table
 */
 void hash_foreach( hash_table_t *h, 
-				   void (*func)( const void *, const void * ) );
+				   void (*func)( void *, void * ) );
 
-/** 
-	Same as hash_foreach, but the function func takes an additional
-	argument, which is provided by the caller in the variable aux 
+/**
+   Same as hash_foreach, but the function func takes an additional
+   argument, which is provided by the caller in the variable aux 
 */
-void hash_foreach2( hash_table_t *h, void (*func)( const void *, 
-												 const void *, 
-												 void *), 
+void hash_foreach2( hash_table_t *h, void (*func)( void *, 
+												   void *, 
+												   void *), 
 					void *aux );
 
 /**
    Hash function suitable for character strings. 
 */
-int hash_str_func( const void *data );
+int hash_str_func( void *data );
 /**
    Hash comparison function suitable for character strings
 */
-int hash_str_cmp( const void *a, const void *b );
+int hash_str_cmp( void *a,
+				  void *b );
 
 /**
    Hash function suitable for wide character strings. 
 */
-int hash_wcs_func( const void *data );
+int hash_wcs_func( void *data );
 
 /**
    Hash comparison function suitable for wide character strings
 */
-int hash_wcs_cmp( const void *a, const void *b );
+int hash_wcs_cmp( void *a, 
+				  void *b );
+
+/**
+   Hash function suitable for direct pointer comparison
+*/
+int hash_ptr_func( void *data );
+
+
+/**
+   Hash comparison function suitable for direct pointer comparison
+*/
+int hash_ptr_cmp( void *a,
+                  void *b );
+
+
 
 /** 
 	Initialize the priority queue
@@ -296,12 +330,12 @@ void pq_init( priority_queue_t *q,
 /**
    Add element to the queue
 
- \param q the queue
- \param e the new element
+   \param q the queue
+   \param e the new element
  
 */
 int pq_put( priority_queue_t *q,
-			 void *e );
+			void *e );
 /**
   Removes and returns the last entry in the priority queue
 */
@@ -328,7 +362,8 @@ int pq_get_count( priority_queue_t *q );
 void pq_destroy(  priority_queue_t *q );
 
 /**
-   Allocate heap memory for creating a new list and initialize it
+   Allocate heap memory for creating a new list and initialize
+   it. Equivalent to calling malloc and al_init.
 */
 array_list_t *al_new();
 
@@ -351,6 +386,24 @@ void al_destroy( array_list_t *l );
    \return 1 if succesfull, 0 otherwise
 */
 int al_push( array_list_t *l, const void *o );
+/**
+   Append element to list
+
+   \param l The list
+   \param o The element
+   \return
+   \return 1 if succesfull, 0 otherwise
+*/
+int al_push_long( array_list_t *l, long o );
+/**
+   Append element to list
+
+   \param l The list
+   \param o The element
+   \return
+   \return 1 if succesfull, 0 otherwise
+*/
+int al_push_func( array_list_t *l, void (*f)() );
 
 /**
    Append all elements of a list to another
@@ -369,6 +422,22 @@ int al_push_all( array_list_t *a, array_list_t *b );
    \param o The element 
 */
 int al_set( array_list_t *l, int pos, const void *o );
+/**
+   Sets the element at the specified index
+
+   \param l The array_list_t
+   \param pos The index 
+   \param o The element 
+*/
+int al_set_long( array_list_t *l, int pos, long v );
+/**
+   Sets the element at the specified index
+
+   \param l The array_list_t
+   \param pos The index 
+   \param o The element 
+*/
+int al_set_func( array_list_t *l, int pos, void (*f)() );
 
 /**
    Returns the element at the specified index
@@ -377,7 +446,23 @@ int al_set( array_list_t *l, int pos, const void *o );
    \param pos The index 
    \return The element 
 */
-const void *al_get( array_list_t *l, int pos );
+void *al_get( array_list_t *l, int pos );
+/**
+   Returns the element at the specified index
+
+   \param l The array_list_t
+   \param pos The index 
+   \return The element 
+*/
+long al_get_long( array_list_t *l, int pos );
+/**
+   Returns the element at the specified index
+
+   \param l The array_list_t
+   \param pos The index 
+   \return The element 
+*/
+func_ptr_t al_get_func( array_list_t *l, int pos );
 
 /**
   Truncates the list to new_sz items.
@@ -387,7 +472,15 @@ void al_truncate( array_list_t *l, int new_sz );
 /**
   Removes and returns the last entry in the list
 */
-const void *al_pop( array_list_t *l );
+void *al_pop( array_list_t *l );
+/**
+  Removes and returns the last entry in the list
+*/
+long al_pop_long( array_list_t *l );
+/**
+  Removes and returns the last entry in the list
+*/
+func_ptr_t al_pop_func( array_list_t *l );
 
 /**
    Returns the number of elements in the list
@@ -397,7 +490,15 @@ int al_get_count( array_list_t *l );
 /**
   Returns the last entry in the list witout removing it.
 */
-const void *al_peek( array_list_t *l );
+void *al_peek( array_list_t *l );
+/**
+  Returns the last entry in the list witout removing it.
+*/
+long al_peek_long( array_list_t *l );
+/**
+  Returns the last entry in the list witout removing it.
+*/
+func_ptr_t al_peek_func( array_list_t *l );
 
 /**
    Returns 1 if the list is empty, 0 otherwise
@@ -407,13 +508,13 @@ int al_empty( array_list_t *l);
 /** 
 	Call the function func for each entry in the list
 */
-void al_foreach( array_list_t *l, void (*func)(const void * ));
+void al_foreach( array_list_t *l, void (*func)( void * ));
 
 /** 
 	Same as al_foreach, but the function func takes an additional
 	argument, which is provided by the caller in the variable aux
 */
-void al_foreach2( array_list_t *l, void (*func)(const void *, void *), void *aux);
+void al_foreach2( array_list_t *l, void (*func)( void *, void *), void *aux);
 
 /**
    Compares two wide character strings without case but with  
@@ -445,11 +546,18 @@ void al_foreach2( array_list_t *l, void (*func)(const void *, void *), void *aux
 int wcsfilecmp( const wchar_t *a, const wchar_t *b );
 
 
+/*
+  String buffer functions
+*/
+
 /**
    Initialize the specified string_buffer
 */
 void sb_init( string_buffer_t * );
 
+/**
+   Allocate memory for storing a stringbuffer and init it
+*/
 string_buffer_t *sb_new();
 
 /**
@@ -496,10 +604,13 @@ int sb_vprintf( string_buffer_t *buffer, const wchar_t *format, va_list va_orig
 void sb_destroy( string_buffer_t * );
 
 /**
-   Truncate the buffer. This will not deallocate the memory used, it will only set the contents of the string to L"\0".
+   Truncate the buffer. This will not deallocate the memory used, it will only set the contents of the string to L"\\0".
 */
 void sb_clear( string_buffer_t * );
 
+/*
+  Buffer functions
+*/
 
 /**
    Initialize the specified buffer_t
