@@ -18,7 +18,10 @@ wildcards using **.
 #include <dirent.h>
 #include <errno.h>
 
+
+#include "fallback.h"
 #include "util.h"
+
 #include "wutil.h"
 #include "complete.h"
 #include "common.h"
@@ -26,7 +29,7 @@ wildcards using **.
 #include "complete.h"
 #include "reader.h"
 #include "expand.h"
-#include "translate.h"
+
 
 /**
    This flag is set in the flags parameter of wildcard_expand if the
@@ -41,9 +44,38 @@ wildcards using **.
 */
 #define MAX_FILE_LENGTH 1024
 
+/**
+   Push the specified argument to the list if an identical string is
+   not already in the list. This function iterates over the list,
+   which is quite slow if the list is large. It might make sense to
+   use a hashtable for this.
+*/
+static void al_push_check( array_list_t *l, const wchar_t *new )
+{
+	int i;
+
+	for( i = 0; i < al_get_count(l); i++ ) 
+	{
+		if( !wcscmp( al_get(l, i), new ) ) 
+		{
+			free( (void *)new );
+			return;
+		}
+	}
+
+	al_push( l, new );
+}
+
+
 int wildcard_has( const wchar_t *str, int internal )
 {
 	wchar_t prev=0;
+	if( !str )
+	{
+		debug( 2, L"Got null string on line %d of file %s", __LINE__, __FILE__ );
+		return 0;		
+	}
+
 	if( internal )
 	{
 		for( ; *str; str++ )
@@ -72,10 +104,7 @@ int wildcard_has( const wchar_t *str, int internal )
    \param str String to be matched.
    \param wc The wildcard.
    \param is_first Whether files beginning with dots should not be matched against wildcards. 
-   \param wc_unescaped Whether the unescaped special character ANY_CHAR abd ANY_STRING should be used instead of '?' and '*' for wildcard matching
 */
-
-
 static int wildcard_match2( const wchar_t *str, 
 							const wchar_t *wc, 
 							int is_first )
@@ -101,6 +130,14 @@ static int wildcard_match2( const wchar_t *str,
 		while( *(str++) != 0 );
 		return 0;
 	}
+	else if( *str == 0 )
+	{
+		/*
+		  End of string, but not end of wildcard, and the next wildcard
+		  element is not a '*', so this is not a match.
+		*/
+		return 0;
+	}
 
 	if( *wc == ANY_CHAR )
 	{
@@ -131,13 +168,20 @@ static int wildcard_complete_internal( const wchar_t *orig,
 									   const wchar_t *(*desc_func)(const wchar_t *),
 									   array_list_t *out )
 {
+	if( !wc || !str || !orig)
+	{
+		debug( 2, L"Got null string on line %d of file %s", __LINE__, __FILE__ );
+		return 0;		
+	}
+
 	if( *wc == 0 &&
 		( ( *str != L'.') || (!is_first)) )
 	{
-		if( !out )
-			return 1;
-		
 		wchar_t *new;
+		if( !out )
+		{
+			return 1;
+		}
 			
 		if( wcschr( str, PROG_COMPLETE_SEP ) )
 		{
@@ -145,7 +189,7 @@ static int wildcard_complete_internal( const wchar_t *orig,
 			  This completion has an embedded description, du not use the generic description
 			*/
 			wchar_t *sep;
-				
+			
 			new = wcsdup( str );
 			sep = wcschr(new, PROG_COMPLETE_SEP );
 			*sep = COMPLETE_SEP;			
@@ -246,14 +290,17 @@ static wchar_t *make_path( const wchar_t *base_dir, const wchar_t *name )
 	int base_len = wcslen( base_dir );
 	if( !(long_name= malloc( sizeof(wchar_t)*(base_len+wcslen(name)+1) )))
 	{
-		die_mem();
+		DIE_MEM();
 	}					
 	wcscpy( long_name, base_dir );
 	wcscpy(&long_name[base_len], name );
 	return long_name;
 }
 
-void get_desc( wchar_t *fn, string_buffer_t *sb, int is_cmd )
+/**
+   Get the description of the specified filename. If this is a regular file, append the filesize to the description.
+*/
+static void get_desc( wchar_t *fn, string_buffer_t *sb, int is_cmd )
 {
 	const wchar_t *desc;
 	
@@ -265,6 +312,11 @@ void get_desc( wchar_t *fn, string_buffer_t *sb, int is_cmd )
 		}
 	;
 	
+	if( !fn || !sb )
+	{
+		debug( 2, L"Got null string on line %d of file %s", __LINE__, __FILE__ );
+		return;		
+	}
 	
 	sb_clear( sb );
 	
@@ -278,17 +330,22 @@ void get_desc( wchar_t *fn, string_buffer_t *sb, int is_cmd )
 	}
 						
 	desc = complete_get_desc( fn );
-	
+
+	if( wcschr( desc, COMPLETE_SEP )==0 )
+	{
+		sb_append( sb, COMPLETE_SEP_STR );
+	}
+		
 	if( sz >= 0 && S_ISDIR(buf.st_mode) )
 	{
-		sb_append2( sb, desc, (void *)0 );							
+		sb_append( sb, desc );
 	}
 	else
 	{							
 		sb_append2( sb, desc, L", ", (void *)0 );
 		if( sz < 0 )
 		{
-			sb_append( sb, _(L"unknown") );
+			sb_append( sb, L"unknown" );
 		}
 		else if( sz < 1 )
 		{
@@ -296,7 +353,7 @@ void get_desc( wchar_t *fn, string_buffer_t *sb, int is_cmd )
 		}
 		else if( sz < 1024 )
 		{
-			sb_printf( sb, L"%dB", sz );
+			sb_printf( sb, L"%lldB", sz );
 		}
 		else
 		{
@@ -321,7 +378,7 @@ void get_desc( wchar_t *fn, string_buffer_t *sb, int is_cmd )
 	}
 }
 
-/*
+/**
   Test if the file specified by the given filename matches the
   expansion flags specified. flags can be a combination of
   EXECUTABLES_ONLY and DIRECTORIES_ONLY.
@@ -338,7 +395,7 @@ static int test_flags( wchar_t *filename,
 		return 0;
 	}
 		
-	if( S_IFDIR & buf.st_mode )
+	if( S_ISDIR( buf.st_mode ) )
 		return 1;
 	
 	if( flags & EXECUTABLES_ONLY )
@@ -358,7 +415,7 @@ int wildcard_expand( const wchar_t *wc,
 	wchar_t *wc_end;
 
 	/* Variables for traversing a directory */
-	struct dirent *next;
+	struct wdirent *next;
 	DIR *dir;
 	
 	/* The result returned */
@@ -379,6 +436,12 @@ int wildcard_expand( const wchar_t *wc,
 	
 	//	debug( 3, L"WILDCARD_EXPAND %ls in %ls", wc, base_dir );
 
+	if( !wc || !base_dir || !out)
+	{
+		debug( 2, L"Got null string on line %d of file %s", __LINE__, __FILE__ );
+		return 0;		
+	}
+
 	if( flags & ACCEPT_INCOMPLETE )
 	{	
 		/* 
@@ -449,15 +512,11 @@ int wildcard_expand( const wchar_t *wc,
 			*/
 			if( flags & ACCEPT_INCOMPLETE )
 			{
-				while( (next=readdir(dir))!=0 )
+				while( (next=wreaddir(dir))!=0 )
 				{
-					if( next->d_name[0] != '.' )
+					if( next->d_name[0] != L'.' )
 					{
-						wchar_t *name = str2wcs(next->d_name);
-						if( name == 0 )
-						{
-							continue;
-						}
+						wchar_t *name = next->d_name;
 						wchar_t *long_name = make_path( base_dir, name );
 						
 						if( test_flags( long_name, flags ) )
@@ -469,8 +528,6 @@ int wildcard_expand( const wchar_t *wc,
 									 wcsdupcat(name, (wchar_t *)sb_desc.buff) );
 						}
 						
-						free(name);
-
 						free( long_name );
 					}					
 				}
@@ -486,13 +543,9 @@ int wildcard_expand( const wchar_t *wc,
 			/*
 			  This is the last wildcard segment, and it is not empty. Match files/directories.
 			*/
-			while( (next=readdir(dir))!=0 )
+			while( (next=wreaddir(dir))!=0 )
 			{
-				wchar_t *name = str2wcs(next->d_name);
-				if( name == 0 )
-				{
-					continue;
-				}
+				wchar_t *name = next->d_name;
 				
 				if( flags & ACCEPT_INCOMPLETE )
 				{
@@ -535,7 +588,6 @@ int wildcard_expand( const wchar_t *wc,
 						res = 1;
 					}
 				}
-				free( name );
 			}
 		}
 	}
@@ -561,7 +613,7 @@ int wildcard_expand( const wchar_t *wc,
 		/*
 		  The maximum length of a file element
 		*/
-		static size_t ln=MAX_FILE_LENGTH;
+		long ln=MAX_FILE_LENGTH;
 		char * narrow_dir_string = wcs2str( dir_string );
 		
 		if( narrow_dir_string )
@@ -585,19 +637,15 @@ int wildcard_expand( const wchar_t *wc,
 
 		if( (!new_dir) || (!wc_str) )
 		{
-			die_mem();
+			DIE_MEM();
 		}
 
 		wcscpy( new_dir, base_dir );
 		
-		while( (next=readdir(dir))!=0 )
+		while( (next=wreaddir(dir))!=0 )
 		{
-			wchar_t *name = str2wcs(next->d_name);
-			if( name == 0 )
-			{
-				continue;
-			}			
-
+			wchar_t *name = next->d_name;
+			
 			/*
 			  Test if the file/directory name matches the whole
 			  wildcard element, i.e. regular matching.
@@ -636,7 +684,7 @@ int wildcard_expand( const wchar_t *wc,
 					
 					if( !stat_res )
 					{
-						if( buf.st_mode & S_IFDIR )
+						if( S_ISDIR(buf.st_mode) )
 						{
 							new_len = wcslen( new_dir );
 							new_dir[new_len] = L'/';
@@ -647,11 +695,23 @@ int wildcard_expand( const wchar_t *wc,
 							*/
 							if( whole_match )
 							{
-								res |= wildcard_expand( wc_end?wc_end + 1:L"", 
+								wchar_t *new_wc = L"";
+								if( wc_end )
+								{
+									new_wc=wc_end+1;
+									/*
+									  Accept multiple '/' as a single direcotry separator
+									*/
+									while(*new_wc==L'/')
+									{
+										new_wc++;
+									}
+								}
+								
+								res |= wildcard_expand( new_wc,
 														new_dir, 
 														flags, 
 														out );
-
 							}
 							
 							/*
@@ -668,7 +728,6 @@ int wildcard_expand( const wchar_t *wc,
 					}
 				}
 			}
-			free(name);
 		}
 		
 		free( wc_str );
@@ -677,23 +736,10 @@ int wildcard_expand( const wchar_t *wc,
 	closedir( dir );
 	
 	if( flags & ACCEPT_INCOMPLETE )
+	{
 		sb_destroy( &sb_desc );
-	
+	}
+
 	return res;
 }
 
-void al_push_check( array_list_t *l, const wchar_t *new )
-{
-	int i;
-
-	for( i = 0; i < al_get_count(l); i++ ) 
-	{
-		if( !wcscmp( al_get(l, i), new ) ) 
-		{
-			free( (void *)new );
-			return;
-		}
-	}
-
-	al_push( l, new );
-}
