@@ -2,16 +2,30 @@
 
   Prototypes for wide character equivalents of various standard unix
   functions. 
-
 */
 #ifndef FISH_WUTIL_H
 #define FISH_WUTIL_H
 
+#include <stdlib.h>
+#include <stdio.h>
 #include <wchar.h>
 #include <dirent.h>
 #include <unistd.h>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <stdarg.h>
+
+/**
+   Wide version of the dirent data structure
+*/
+struct wdirent
+{
+	/**
+	   The name of the current directory
+	*/
+	wchar_t *d_name;
+}
+	;
 
 
 /**
@@ -81,38 +95,43 @@ wchar_t *wgetcwd( wchar_t *buff, size_t sz );
 */
 int wchdir( const wchar_t * dir );
 
-
-#if !HAVE_WPRINTF
+/** 
+	Wide character version of realpath function. Just like the GNU
+	version of realpath, wrealpath will accept 0 as the value for the
+	second argument, in which case the result will be allocated using
+	malloc, and must be free'd by the user.
+*/
+wchar_t *wrealpath(const wchar_t *pathname, wchar_t *resolved_path);
 
 /**
-   Print formated string. Some operating systems (Like NetBSD) do not
-   have wide string formating functions.  Therefore we implement our
-   own. Not at all complete. Supports wide and narrow characters,
-   strings and decimal numbers, position (%n), field width and
-   precision.
+   Wide character version of readdir()
 */
-int fwprintf( FILE *f, const wchar_t *format, ... );
-
+struct wdirent *wreaddir(DIR *dir );
 
 /**
-   Print formated string. Some operating systems (Like NetBSD) do not
-   have wide string formating functions.  Therefore we define our
-   own. Not at all complete. Supports wide and narrow characters,
-   strings and decimal numbers, position (%n), field width and
-   precision.
+   Wide character version of dirname()
 */
-int swprintf( wchar_t *str, size_t l, const wchar_t *format, ... );
+wchar_t *wdirname( const wchar_t *path );
 
 /**
-   Print formated string. Some operating systems (Like NetBSD) do not
-   have wide string formating functions.  Therefore we define our
-   own. Not at all complete. Supports wide and narrow characters,
-   strings and decimal numbers, position (%n), field width and
-   precision.
+   Wide character version of basename()
 */
-int wprintf( const wchar_t *format, ... );
+wchar_t *wbasename( const wchar_t *path );
+
+/**
+   Wide character wrapper around the gettext function. For historic
+   reasons, unlike the real gettext function, wgettext takes care of
+   setting the correct domain, etc. using the textdomain and
+   bindtextdomain functions. This should probably be moved out of
+   wgettext, so that wgettext will be nothing more than a wrapper
+   around gettext, like all other functions in this file.
+*/
+const wchar_t *wgettext( const wchar_t *in );
+
+/**
+   Wide character version of getenv
+*/
+wchar_t *wgetenv( const wchar_t *name );
 
 
-#endif
-
 #endif
